/*****************************************************************************
The Dark Mod GPL Source Code

This file is part of the The Dark Mod Source Code, originally based
on the Doom 3 GPL Source Code as published in 2011.

The Dark Mod Source Code is free software: you can redistribute it
and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of the License,
or (at your option) any later version. For details, see LICENSE.TXT.

Project: The Dark Mod (http://www.thedarkmod.com/)

******************************************************************************/

#ifndef _TYPEINFO_GENHELPER_H
#define _TYPEINFO_GENHELPER_H

//stgatilov: make sure all game class definitions are available here for two reasons:
//  1. TypeInfo utility would parse this cpp file and it must include everything
//  2. If memory debugging is enabled, the classes are referenced in GameTypeInfo.h

#include "../game/Moveable.h"
#include "../game/Mover.h"
#include "../game/MultiStateMover.h"
#include "../game/MultiStateMoverButton.h"
#include "../game/MultiStateMoverPosition.h"
#include "../game/Objectives/BoolParseNode.h"
#include "../game/Objectives/CampaignStatistics.h"
#include "../game/Objectives/EMissionResult.h"
#include "../game/Objectives/MissionData.h"
#include "../game/Objectives/MissionStatistics.h"
#include "../game/Objectives/Objective.h"
#include "../game/Objectives/ObjectiveComponent.h"
#include "../game/Objectives/ObjectiveCondition.h"
#include "../game/Objectives/ObjectiveLocation.h"
#include "../game/OverlaySys.h"
#include "../game/physics/Clip.h"
#include "../game/physics/Force.h"
#include "../game/physics/Force_Constant.h"
#include "../game/physics/Force_Drag.h"
#include "../game/physics/Force_Field.h"
#include "../game/physics/Force_Push.h"
#include "../game/physics/Force_Spring.h"
#include "../game/physics/Physics.h"
#include "../game/physics/Physics_Actor.h"
#include "../game/physics/Physics_AF.h"
#include "../game/physics/Physics_Base.h"
#include "../game/physics/Physics_Liquid.h"
#include "../game/physics/Physics_Monster.h"
#include "../game/physics/Physics_Parametric.h"
#include "../game/physics/Physics_Player.h"
#include "../game/physics/Physics_RigidBody.h"
#include "../game/physics/Physics_Static.h"
#include "../game/physics/Physics_StaticMulti.h"
#include "../game/physics/Push.h"
#include "../game/PickableLock.h"
#include "../game/Player.h"
#include "../game/PlayerIcon.h"
#include "../game/PlayerView.h"
#include "../game/PositionWithinRangeFinder.h"
#include "../game/precompiled_game.h"
#include "../game/Projectile.h"
#include "../game/ProjectileResult.h"
#include "../game/Pvs.h"
#include "../game/PVSToAASMapping.h"
#include "../game/RawVector.h"
#include "../game/Relations.h"
#include "../game/script/Script_Compiler.h"
#include "../game/script/Script_Doc_Export.h"
#include "../game/script/Script_Interpreter.h"
#include "../game/script/Script_Program.h"
#include "../game/script/Script_Thread.h"
#include "../game/SearchManager.h"
#include "../game/SecurityCamera.h"
#include "../game/SEED.h"
#include "../game/Shop/LootRuleSet.h"
#include "../game/Shop/Shop.h"
#include "../game/Shop/ShopItem.h"
#include "../game/SmokeParticles.h"
#include "../game/SndProp.h"
#include "../game/SndPropLoader.h"
#include "../game/Sound.h"
#include "../game/StaticMulti.h"
#include "../game/StimResponse/Response.h"
#include "../game/StimResponse/ResponseEffect.h"
#include "../game/StimResponse/Stim.h"
#include "../game/StimResponse/StimResponse.h"
#include "../game/StimResponse/StimResponseCollection.h"
#include "../game/StimResponse/StimResponseTimer.h"
#include "../game/StimResponse/StimType.h"
#include "../game/Target.h"
#include "../game/TimerManager.h"
#include "../game/Trigger.h"
#include "../game/Turret.h"
#include "../game/UserManager.h"
#include "../game/Weapon.h"
#include "../game/WorldSpawn.h"
#include "../game/AbsenceMarker.h"
#include "../game/Actor.h"
#include "../game/AF.h"
#include "../game/AFEntity.h"
#include "../game/ai/AAS.h"
#include "../game/ai/AAS_local.h"
#include "../game/ai/AI.h"
#include "../game/ai/AreaManager.h"
#include "../game/ai/CommunicationSubsystem.h"
#include "../game/ai/Conversation/Conversation.h"
#include "../game/ai/Conversation/ConversationCommand.h"
#include "../game/ai/Conversation/ConversationSystem.h"
#include "../game/ai/DoorInfo.h"
#include "../game/ai/EAS/ClusterInfo.h"
#include "../game/ai/EAS/EAS.h"
#include "../game/ai/EAS/ElevatorStationInfo.h"
#include "../game/ai/EAS/RouteInfo.h"
#include "../game/ai/EAS/RouteNode.h"
#include "../game/ai/Library.h"
#include "../game/ai/Memory.h"
#include "../game/ai/Mind.h"
#include "../game/ai/MovementSubsystem.h"
#include "../game/ai/MoveState.h"
#include "../game/ai/Queue.h"
#include "../game/ai/States/AgitatedSearchingState.h"
#include "../game/ai/States/AgitatedSearchingStateLanternBot.h"
#include "../game/ai/States/AlertIdleState.h"
#include "../game/ai/States/BlindedState.h"
#include "../game/ai/States/CombatState.h"
#include "../game/ai/States/ConversationState.h"
#include "../game/ai/States/DeadState.h"
#include "../game/ai/States/EmergeFromCoverState.h"
#include "../game/ai/States/ExamineRopeState.h"
#include "../game/ai/States/FailedKnockoutState.h"
#include "../game/ai/States/FleeDoneState.h"
#include "../game/ai/States/FleeState.h"
#include "../game/ai/States/HitByMoveableState.h"
#include "../game/ai/States/IdleSleepState.h"
#include "../game/ai/States/IdleState.h"
#include "../game/ai/States/KnockedOutState.h"
#include "../game/ai/States/LostTrackOfEnemyState.h"
#include "../game/ai/States/ObservantState.h"
#include "../game/ai/States/PainState.h"
#include "../game/ai/States/PocketPickedState.h"
#include "../game/ai/States/SearchingState.h"
#include "../game/ai/States/State.h"
#include "../game/ai/States/StayInCoverState.h"
#include "../game/ai/States/SuspiciousState.h"
#include "../game/ai/States/SwitchOnLightState.h"
#include "../game/ai/States/TakeCoverState.h"
#include "../game/ai/States/UnreachableTargetState.h"
#include "../game/ai/Subsystem.h"
#include "../game/ai/Tasks/AnimalPatrolTask.h"
#include "../game/ai/Tasks/ChaseEnemyRangedTask.h"
#include "../game/ai/Tasks/ChaseEnemyTask.h"
#include "../game/ai/Tasks/CombatTask.h"
#include "../game/ai/Tasks/CommunicationTask.h"
#include "../game/ai/Tasks/CommWaitTask.h"
#include "../game/ai/Tasks/FleeTask.h"
#include "../game/ai/Tasks/FollowActorTask.h"
#include "../game/ai/Tasks/GreetingBarkTask.h"
#include "../game/ai/Tasks/GuardSpotTask.h"
#include "../game/ai/Tasks/HandleDoorTask.h"
#include "../game/ai/Tasks/HandleElevatorTask.h"
#include "../game/ai/Tasks/IdleAnimationTask.h"
#include "../game/ai/Tasks/InteractionTask.h"
#include "../game/ai/Tasks/InvestigateSpotTask.h"
#include "../game/ai/Tasks/MeleeCombatTask.h"
#include "../game/ai/Tasks/MoveToCoverTask.h"
#include "../game/ai/Tasks/MoveToPositionTask.h"
#include "../game/ai/Tasks/PathAnimTask.h"
#include "../game/ai/Tasks/PathCornerTask.h"
#include "../game/ai/Tasks/PathCycleAnimTask.h"
#include "../game/ai/Tasks/PathHideTask.h"
#include "../game/ai/Tasks/PathInteractTask.h"
#include "../game/ai/Tasks/PathLookatTask.h"
#include "../game/ai/Tasks/PathSetMovetypeTask.h"
#include "../game/ai/Tasks/PathShowTask.h"
#include "../game/ai/Tasks/PathSitTask.h"
#include "../game/ai/Tasks/PathSleepTask.h"
#include "../game/ai/Tasks/PathTask.h"
#include "../game/ai/Tasks/PathTurnTask.h"
#include "../game/ai/Tasks/PathWaitForTriggerTask.h"
#include "../game/ai/Tasks/PathWaitTask.h"
#include "../game/ai/Tasks/PlayAnimationTask.h"
#include "../game/ai/Tasks/RandomHeadturnTask.h"
#include "../game/ai/Tasks/RandomTurningTask.h"
#include "../game/ai/Tasks/RangedCombatTask.h"
#include "../game/ai/Tasks/RepeatedBarkTask.h"
#include "../game/ai/Tasks/ResolveMovementBlockTask.h"
#include "../game/ai/Tasks/ScriptTask.h"
#include "../game/ai/Tasks/SingleBarkTask.h"
#include "../game/ai/Tasks/Task.h"
#include "../game/ai/Tasks/ThrowObjectTask.h"
#include "../game/ai/Tasks/WaitTask.h"
#include "../game/ai/Tasks/WanderInLocationTask.h"
#include "../game/ai/tdmAASFindEscape.h"
#include "../game/AIComm_Message.h"
#include "../game/AIVehicle.h"
#include "../game/anim/Anim.h"
#include "../game/anim/Anim_Testmodel.h"
#include "../game/BinaryFrobMover.h"
#include "../game/BloodMarker.h"
#include "../game/BrittleFracture.h"
#include "../game/ButtonStateTracker.h"
#include "../game/Camera.h"
#include "../game/DarkmodAASHidingSpotFinder.h"
#include "../game/DarkModGlobals.h"
#include "../game/darkmodHidingSpotTree.h"
#include "../game/darkModLAS.h"
#include "../game/decltdm_matinfo.h"
#include "../game/declxdata.h"
#include "../game/DifficultyManager.h"
#include "../game/DifficultySettings.h"
#include "../game/DownloadMenu.h"
#include "../game/Emitter.h"
#include "../game/Entity.h"
#include "../game/EscapePointEvaluator.h"
#include "../game/EscapePointManager.h"
#include "../game/Force_Grab.h"
#include "../game/FrobButton.h"
#include "../game/FrobDoor.h"
#include "../game/FrobDoorHandle.h"
#include "../game/FrobHandle.h"
#include "../game/FrobLever.h"
#include "../game/FrobLock.h"
#include "../game/FrobLockHandle.h"
#include "../game/Func_Shooter.h"
#include "../game/FX.h"
#include "../game/Game.h"
#include "../game/GameEdit.h"
#include "../game/GamePlayTimer.h"
#include "../game/gamesys/Class.h"
#include "../game/gamesys/DebugGraph.h"
#include "../game/gamesys/Event.h"
#include "../game/gamesys/EventArgs.h"
#include "../game/gamesys/SaveGame.h"
#include "../game/gamesys/SysCmds.h"
#include "../game/gamesys/SysCvar.h"
#include "../game/gamesys/TypeInfo.h"
#include "../game/Game_local.h"
#include "../game/Grabber.h"
#include "../game/HidingSpotSearchCollection.h"
#include "../game/Http/HttpConnection.h"
#include "../game/Http/HttpRequest.h"
#include "../game/IK.h"
#include "../game/IniFile.h"
#include "../game/Intersection.h"
#include "../game/Inventory/Category.h"
#include "../game/Inventory/Cursor.h"
#include "../game/Inventory/Inventory.h"
#include "../game/Inventory/InventoryItem.h"
#include "../game/Inventory/LootType.h"
#include "../game/Inventory/WeaponItem.h"
#include "../game/Item.h"
#include "../game/Light.h"
#include "../game/LightController.h"
#include "../game/LightGem.h"
#include "../game/Liquid.h"
#include "../game/MatrixSq.h"
#include "../game/MeleeWeapon.h"
#include "../game/Misc.h"
#include "../game/Missions/Download.h"
#include "../game/Missions/DownloadManager.h"
#include "../game/Missions/MissionDB.h"
#include "../game/Missions/MissionManager.h"
#include "../game/Missions/ModInfo.h"
#include "../game/Missions/ModInfoDecl.h"
#include "../game/ModelGenerator.h"
#include "../game/ModMenu.h"

#endif
