/*
** 2000-05-29
**
** The author disclaims copyright to this source code.  In place of
** a legal notice, here is a blessing:
**
**    May you do good and not evil.
**    May you find forgiveness for yourself and forgive others.
**    May you share freely, never taking more than you give.
**
*************************************************************************
** Driver template for the LEMON parser generator.
**
** The "lemon" program processes an LALR(1) input grammar file, then uses
** this template to construct a parser.  The "lemon" program inserts text
** at each "%%" line.  Also, any "P-a-r-s-e" identifer prefix (without the
** interstitial "-" characters) contained in this template is changed into
** the value of the %name directive from the grammar.  Otherwise, the content
** of this template is copied straight through into the generate parser
** source file.
**
** The following is the concatenation of all %include directives from the
** input grammar file:
*/
#include <stdio.h>
#include <string.h>
#include <assert.h>

#ifdef _MSC_VER
#define CDECL __cdecl
#else
#define CDECL
#endif

/************ Begin %include sections from the grammar ************************/
#line 40 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"

// Allocates a new AST node off the parse state's arena.
#define NEW_AST_NODE(type,name,tok) \
	ZCC_##type *name = static_cast<ZCC_##type *>(stat->InitNode(sizeof(ZCC_##type), AST_##type)); \
	SetNodeLine(name, tok)

static void SetNodeLine(ZCC_TreeNode *name, ZCCToken &tok)
{
	name->SourceLoc = tok.SourceLoc;
}

static void SetNodeLine(ZCC_TreeNode *name, ZCC_TreeNode *node)
{
	if (name == nullptr || node == nullptr)
	{
		I_Error("Fatal parse error");
	}
	name->SourceLoc = node->SourceLoc;
}

static void SetNodeLine(ZCC_TreeNode *name, int line)
{
	name->SourceLoc = line;
}

// If a is non-null, appends b to a. Otherwise, sets a to b.
#define SAFE_APPEND(a,b) \
	if (a == NULL) a = b; else AppendTreeNodeSibling(a, b);

#define UNARY_EXPR(X,T) NEW_AST_NODE(ExprUnary, expr1, X); expr1->Operation = T; expr1->Operand = X; expr1->Type = NULL
#define BINARY_EXPR(X,Y,T) NEW_AST_NODE(ExprBinary, expr2, X); expr2->Operation = T; expr2->Type = NULL; expr2->Left = X; expr2->Right = Y

#define NEW_INTCONST_NODE(name,type,val,tok) \
	NEW_AST_NODE(ExprConstant, name, tok); \
	name->Operation = PEX_ConstValue; \
	name->Type = type; \
	name->IntVal = val

	struct ClassFlagsBlock {
		VM_UWORD Flags;
		ZCC_Identifier *Replaces;
		ZCC_Identifier *Sealed;
		VersionInfo Version;
	};

	struct StateOpts {
		ZCC_Expression *Offset;
		ZCC_ExprConstant *Lights;
		bool Bright;
		bool Fast;
		bool Slow;
		bool NoDelay;
		bool CanRaise;

		void Zero() {
			Offset = nullptr;
			Lights = nullptr;
			Bright = false;
			Fast = false;
			Slow = false;
			NoDelay = false;
			CanRaise = false;
		}
	};

	struct VarOrFun
	{
		ZCC_VarName *VarNames;
		ZCC_FuncParamDecl *FuncParams;
		ZCC_CompoundStmt *FuncBody;
		ENamedName FuncName;
		int FuncFlags;
		int SourceLoc;
	};
#line 112 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
/**************** End of %include directives **********************************/
/* These constants specify the various numeric values for terminal symbols
** in a format understandable to "makeheaders".  This section is blank unless
** "lemon" is run with the "-m" command-line option.
***************** Begin makeheaders token definitions *************************/
/**************** End makeheaders token definitions ***************************/

/* The next section is a series of control #defines.
** various aspects of the generated parser.
**    YYCODETYPE         is the data type used to store the integer codes
**                       that represent terminal and non-terminal symbols.
**                       "unsigned char" is used if there are fewer than
**                       256 symbols.  Larger types otherwise.
**    YYNOCODE           is a number of type YYCODETYPE that is not used for
**                       any terminal or nonterminal symbol.
**    YYFALLBACK         If defined, this indicates that one or more tokens
**                       (also known as: "terminal symbols") have fall-back
**                       values which should be used if the original symbol
**                       would not parse.  This permits keywords to sometimes
**                       be used as identifiers, for example.
**    YYACTIONTYPE       is the data type used for "action codes" - numbers
**                       that indicate what to do in response to the next
**                       token.
**    ZCCParseTOKENTYPE     is the data type used for minor type for terminal
**                       symbols.  Background: A "minor type" is a semantic
**                       value associated with a terminal or non-terminal
**                       symbols.  For example, for an "ID" terminal symbol,
**                       the minor type might be the name of the identifier.
**                       Each non-terminal can have a different minor type.
**                       Terminal symbols all have the same minor type, though.
**                       This macros defines the minor type for terminal 
**                       symbols.
**    YYMINORTYPE        is the data type used for all minor types.
**                       This is typically a union of many types, one of
**                       which is ZCCParseTOKENTYPE.  The entry in the union
**                       for terminal symbols is called "yy0".
**    YYSTACKDEPTH       is the maximum depth of the parser's stack.  If
**                       zero the stack is dynamically sized using realloc()
**    ZCCParseARG_SDECL     A static variable declaration for the %extra_argument
**    ZCCParseARG_PDECL     A parameter declaration for the %extra_argument
**    ZCCParseARG_STORE     Code to store %extra_argument into yypParser
**    ZCCParseARG_FETCH     Code to extract %extra_argument from yypParser
**    YYERRORSYMBOL      is the code number of the error symbol.  If not
**                       defined, then do no error processing.
**    YYNSTATE           the combined number of states.
**    YYNRULE            the number of rules in the grammar
**    YY_MAX_SHIFT       Maximum value for shift actions
**    YY_MIN_SHIFTREDUCE Minimum value for shift-reduce actions
**    YY_MAX_SHIFTREDUCE Maximum value for shift-reduce actions
**    YY_MIN_REDUCE      Maximum value for reduce actions
**    YY_ERROR_ACTION    The yy_action[] code for syntax error
**    YY_ACCEPT_ACTION   The yy_action[] code for accept
**    YY_NO_ACTION       The yy_action[] code for no-op
*/
#ifndef INTERFACE
# define INTERFACE 1
#endif
/************* Begin control #defines *****************************************/
#define YYCODETYPE unsigned short int
#define YYNOCODE 279
#define YYACTIONTYPE unsigned short int
#define ZCCParseTOKENTYPE  ZCCToken 
typedef union {
  int yyinit;
  ZCCParseTOKENTYPE yy0;
  ZCC_Statement * yy3;
  ZCC_AssignDeclStmt * yy7;
  ZCC_IfStmt * yy45;
  ZCC_VarInit * yy55;
  ZCC_StatePart * yy74;
  ZCC_States * yy82;
  ZCC_TreeNode * yy128;
  ZCC_ConstantDef * yy139;
  ZCC_Type * yy196;
  ZCC_Struct * yy211;
  ZCC_ExprFuncCall * yy221;
  ZCC_Enum * yy239;
  ZCC_Property * yy240;
  ZCC_Default * yy251;
  ZCC_CaseStmt * yy254;
  ZCC_Expression * yy282;
  ZCC_PropertyStmt * yy283;
  ZCC_AssignStmt * yy284;
  ZCC_StaticArrayStatement * yy286;
  ZCC_Declarator * yy327;
  ZCC_ExpressionStmt * yy330;
  ZCC_FlagDef * yy343;
  ZCC_MixinStmt * yy351;
  StateOpts yy360;
  VarOrFun yy364;
  ZCC_Class * yy392;
  ZCC_DeclFlags * yy415;
  ZCC_FuncPtrType * yy418;
  ZCC_ExprConstant * yy423;
  ZCC_BasicType * yy430;
  ZCC_FuncParamDecl * yy439;
  ZCC_LocalVarStmt * yy454;
  ZCC_CompoundStmt * yy475;
  ZCC_FuncParm * yy478;
  ZCC_VarName * yy492;
  ZCC_Identifier * yy493;
  ZCC_MixinDef * yy498;
  ZCC_FlagStmt * yy500;
  ZCC_FuncPtrParamDecl * yy513;
  ClassFlagsBlock yy515;
  int yy557;
} YYMINORTYPE;
#ifndef YYSTACKDEPTH
#define YYSTACKDEPTH 0
#endif
#define ZCCParseARG_SDECL  ZCCParseState *stat ;
#define ZCCParseARG_PDECL , ZCCParseState *stat 
#define ZCCParseARG_FETCH  ZCCParseState *stat  = yypParser->stat 
#define ZCCParseARG_STORE yypParser->stat  = stat 
#define YYERRORSYMBOL 161
#define YYERRSYMDT yy557
#define YYFALLBACK 1
#define YYNSTATE             446
#define YYNRULE              401
#define YY_MAX_SHIFT         445
#define YY_MIN_SHIFTREDUCE   736
#define YY_MAX_SHIFTREDUCE   1136
#define YY_MIN_REDUCE        1137
#define YY_MAX_REDUCE        1537
#define YY_ERROR_ACTION      1538
#define YY_ACCEPT_ACTION     1539
#define YY_NO_ACTION         1540
/************* End control #defines *******************************************/

/* Define the yytestcase() macro to be a no-op if is not already defined
** otherwise.
**
** Applications can choose to define yytestcase() in the %include section
** to a macro that can assist in verifying code coverage.  For production
** code the yytestcase() macro should be turned off.  But it is useful
** for testing.
*/
#ifndef yytestcase
# define yytestcase(X)
#endif


/* Next are the tables used to determine what action to take based on the
** current state and lookahead token.  These tables are used to implement
** functions that take a state number and lookahead value and return an
** action integer.  
**
** Suppose the action integer is N.  Then the action is determined as
** follows
**
**   0 <= N <= YY_MAX_SHIFT             Shift N.  That is, push the lookahead
**                                      token onto the stack and goto state N.
**
**   N between YY_MIN_SHIFTREDUCE       Shift to an arbitrary state then
**     and YY_MAX_SHIFTREDUCE           reduce by rule N-YY_MIN_SHIFTREDUCE.
**
**   N between YY_MIN_REDUCE            Reduce by rule N-YY_MIN_REDUCE
**     and YY_MAX_REDUCE
**
**   N == YY_ERROR_ACTION               A syntax error has occurred.
**
**   N == YY_ACCEPT_ACTION              The parser accepts its input.
**
**   N == YY_NO_ACTION                  No such action.  Denotes unused
**                                      slots in the yy_action[] table.
**
** The action table is constructed as a single large table named yy_action[].
** Given state S and lookahead X, the action is computed as either:
**
**    (A)   N = yy_action[ yy_shift_ofst[S] + X ]
**    (B)   N = yy_default[S]
**
** The (A) formula is preferred.  The B formula is used instead if:
**    (1)  The yy_shift_ofst[S]+X value is out of range, or
**    (2)  yy_lookahead[yy_shift_ofst[S]+X] is not equal to X, or
**    (3)  yy_shift_ofst[S] equal YY_SHIFT_USE_DFLT.
** (Implementation note: YY_SHIFT_USE_DFLT is chosen so that
** YY_SHIFT_USE_DFLT+X will be out of range for all possible lookaheads X.
** Hence only tests (1) and (2) need to be evaluated.)
**
** The formulas above are for computing the action when the lookahead is
** a terminal symbol.  If the lookahead is a non-terminal (as occurs after
** a reduce action) then the yy_reduce_ofst[] array is used in place of
** the yy_shift_ofst[] array and YY_REDUCE_USE_DFLT is used in place of
** YY_SHIFT_USE_DFLT.
**
** The following are the tables generated in this section:
**
**  yy_action[]        A single table containing all actions.
**  yy_lookahead[]     A table containing the lookahead for each entry in
**                     yy_action.  Used to detect hash collisions.
**  yy_shift_ofst[]    For each state, the offset into yy_action for
**                     shifting terminals.
**  yy_reduce_ofst[]   For each state, the offset into yy_action for
**                     shifting non-terminals after a reduce.
**  yy_default[]       Default action for each state.
**
*********** Begin parsing tables **********************************************/
#define YY_ACTTAB_COUNT (3645)
static const YYACTIONTYPE yy_action[] = {
 /*     0 */   127,  126,  399,  397,  397,  325,   39, 1361,  959,  124,
 /*    10 */   125,  162,   33,   45,  226,  404, 1048,  846,  193,  310,
 /*    20 */   186,  376,  768,  444,  442,  295,  312,    2,  424, 1039,
 /*    30 */   422,  385,  169,  349,  893,    2, 1063,  763,  159,  159,
 /*    40 */  1041,   95,  108,  107,  106,  116,  117,  114,  113,  112,
 /*    50 */   110,  109,  111,  973,  912,  337,   39,  270,  311,   78,
 /*    60 */   961,  878,  879,  880,  881,  882,  883,  884,  885,  886,
 /*    70 */   887,  888,  889,  890,  891,  892,  379,  419,  418,  136,
 /*    80 */   413,  410,  407,  825,  833,  834,  835,  192,  420,  264,
 /*    90 */   226,  974,  270,  270,  272,  837,  838,  839,  840,  148,
 /*   100 */  1311,   40,  404,  382,  895,  123,  122,  121,  120, 1042,
 /*   110 */  1043, 1044, 1045, 1046, 1047,   40,  361,  360,   42,   13,
 /*   120 */   356,  353, 1076, 1077,  351,   38,  350,   62,  918,  895,
 /*   130 */   919,  141,  395,  400,  395,  963,  771,  925,  904,  127,
 /*   140 */   126,  274,  274,  129,   49,  774,  773,  777,  124,  125,
 /*   150 */   162,   33,   45,  911,  962, 1048,  770,  764,  336,  186,
 /*   160 */   376,  404,  772,  778,  779,  775,  776,  780, 1039,  765,
 /*   170 */   766,  761,  349,  893,    2, 1064,  264,  993,  974, 1041,
 /*   180 */   108,  107,  106,  116,  117,  114,  113,  112,  110,  109,
 /*   190 */   111,  135,  973,  739,   49,  180,  180,   78, 1539,   23,
 /*   200 */   878,  879,  880,  881,  882,  883,  884,  885,  886,  887,
 /*   210 */   888,  889,  890,  891,  892,  379,  419,  418,   19,  413,
 /*   220 */   410,  407,  984,  985,  402,   26,   75,  420,  738,  740,
 /*   230 */   741,  742,  743,  744,  745,  264,  992,  974,  926,  926,
 /*   240 */   163,  404,  347,  347,  123,  122,  121,  120, 1042, 1043,
 /*   250 */  1044, 1045, 1046, 1047,   40,  361,  360,   42,   13,  356,
 /*   260 */   353, 1076, 1077,  351,  812,  350,   62,  116,  117,  114,
 /*   270 */   113,  112,  110,  109,  111,  439,  438,  134,  127,  126,
 /*   280 */    51,   78,  807,   65,  339,  393,  378,  124,  125,  162,
 /*   290 */    33,   45,  223,  404, 1048,  226,  404,  762,  186,  376,
 /*   300 */   185,  185,  199,  404, 1076, 1077,  767, 1039,  763,  340,
 /*   310 */   325,  349,  893,    2,  806,  264,  991,  974, 1041,  176,
 /*   320 */   182,  383,  182,  139,  310,   49,  876,   68,  444,  442,
 /*   330 */   295,  973,  926,  424,  832,  422,  385,  169,  286,  878,
 /*   340 */   879,  880,  881,  882,  883,  884,  885,  886,  887,  888,
 /*   350 */   889,  890,  891,  892,  379,  419,  418,  196,  413,  410,
 /*   360 */   407,  387,  222,  404,  346,  156,  420,  264,  223,  974,
 /*   370 */   264,  226,  974,  286,  286, 1082,  191,  264,  199,  974,
 /*   380 */   128,  927,  174,  123,  122,  121,  120, 1042, 1043, 1044,
 /*   390 */  1045, 1046, 1047,   40,  361,  360,   42,   13,  356,  353,
 /*   400 */  1076, 1077,  351,  162,  350,   62,   40,  166,  306,   49,
 /*   410 */   269,  836,  186,  896,  951,  944,  939,  953,  954,   38,
 /*   420 */   331,    1,  380,   22,  173,  404,  893,  871,  341,  341,
 /*   430 */   324,  345,   68,  948,   68,  955,  404,  264,  222,  974,
 /*   440 */   848,  849,  850,  851,  852,  320,  319,  867,  222,  404,
 /*   450 */   343, 1095,  299,  878,  879,  880,  881,  882,  883,  884,
 /*   460 */   885,  886,  887,  888,  889,  890,  891,  892,  894,  419,
 /*   470 */   381,  438,  413,  410,  407,  922,   68,   55, 1062,   68,
 /*   480 */   420,  440,  335,  940,  941,  942,  943,  945,  946,  949,
 /*   490 */   950,  952,  956,  127,  126, 1096,  277,  796,  798,  264,
 /*   500 */   990,  974,  124,  125,  162,   33,  348,  816,  799,  800,
 /*   510 */   264,  989,  974,  186,  376,   36,  818,  817,  821,  354,
 /*   520 */   240,  404, 1039,  264,  222,  974,  802,  893,  864,  801,
 /*   530 */   322,   37,   37, 1041,  822,  823,  819,  820,  824,  754,
 /*   540 */   755,  756,  757,  758,  151,  303,  973,  304,  300,  843,
 /*   550 */   307,  815,  895,  308,  878,  879,  880,  881,  882,  883,
 /*   560 */   884,  885,  886,  887,  888,  889,  890,  891,  892,  894,
 /*   570 */   419,  418,  318,  413,  410,  407,  918,  895,  919,  141,
 /*   580 */    19,  420,  921,  226,  404,  854,  904,  111,  154,   19,
 /*   590 */   401,  240,  404,  841,   78,  264,  240,  974,  123,  122,
 /*   600 */   121,  120, 1042, 1043, 1044, 1045, 1046, 1047,   90,   87,
 /*   610 */    86,   85,   89,   88,   84,   83,   81,   80,   79,   82,
 /*   620 */    77,  842,   91,   92,   98,   97,   96,  104,  103,  102,
 /*   630 */   101,  100,   99,  105,   93,   94,   95,  108,  107,  106,
 /*   640 */   116,  117,  114,  113,  112,  110,  109,  111,  157,   60,
 /*   650 */   924,  336,   43,   52,   78,  357,   36,   72,  264,  226,
 /*   660 */   974, 1035, 1035,  265,  827,   12,  264,  240,  974,  980,
 /*   670 */   976,  293,   90,   87,   86,   85,   89,   88,   84,   83,
 /*   680 */    81,   80,   79,   82,   77,  175,   91,   92,   98,   97,
 /*   690 */    96,  104,  103,  102,  101,  100,   99,  105,   93,   94,
 /*   700 */    95,  108,  107,  106,  116,  117,  114,  113,  112,  110,
 /*   710 */   109,  111,  114,  113,  112,  110,  109,  111,   78,   36,
 /*   720 */   414,   73,  748,  266,   78,  878,  879,  880,  881,  882,
 /*   730 */   883,  150,  979,  828,  977,  417,   90,   87,   86,   85,
 /*   740 */    89,   88,   84,   83,   81,   80,   79,   82,   77,  190,
 /*   750 */    91,   92,   98,   97,   96,  104,  103,  102,  101,  100,
 /*   760 */    99,  105,   93,   94,   95,  108,  107,  106,  116,  117,
 /*   770 */   114,  113,  112,  110,  109,  111,  793,  790,  791,  404,
 /*   780 */   328, 1196,   78,  404,  295,   74,  366,   21,  268,  422,
 /*   790 */   385,  416,   24,  862,  371,  792,  855, 1055,  978,  315,
 /*   800 */    90,   87,   86,   85,   89,   88,   84,   83,   81,   80,
 /*   810 */    79,   82,   77,   41,   91,   92,   98,   97,   96,  104,
 /*   820 */   103,  102,  101,  100,   99,  105,   93,   94,   95,  108,
 /*   830 */   107,  106,  116,  117,  114,  113,  112,  110,  109,  111,
 /*   840 */   917,  435,  271,  172, 1101,  194,   78,  788,  332,  119,
 /*   850 */   868,  872,  873,  264,  988,  974,  333,  264,  987,  974,
 /*   860 */    40,  430,  853,  265,   90,   87,   86,   85,   89,   88,
 /*   870 */    84,   83,   81,   80,   79,   82,   77,  273,   91,   92,
 /*   880 */    98,   97,   96,  104,  103,  102,  101,  100,   99,  105,
 /*   890 */    93,   94,   95,  108,  107,  106,  116,  117,  114,  113,
 /*   900 */   112,  110,  109,  111,  934,  139,  931,  344,  276,   68,
 /*   910 */    78, 1098,  149,   53,  958,   90,   87,   86,   85,   89,
 /*   920 */    88,   84,   83,   81,   80,   79,   82,   77,  362,   91,
 /*   930 */    92,   98,   97,   96,  104,  103,  102,  101,  100,   99,
 /*   940 */   105,   93,   94,   95,  108,  107,  106,  116,  117,  114,
 /*   950 */   113,  112,  110,  109,  111, 1506,  219,  404,  295,  133,
 /*   960 */   808,   78,  404,  422,  385,  181, 1108,  804,  146,  384,
 /*   970 */   161,  915,  787,   68, 1039,  445,   18,    5,  260,   90,
 /*   980 */    87,   86,   85,   89,   88,   84,   83,   81,   80,   79,
 /*   990 */    82,   77,  367,   91,   92,   98,   97,   96,  104,  103,
 /*  1000 */   102,  101,  100,   99,  105,   93,   94,   95,  108,  107,
 /*  1010 */   106,  116,  117,  114,  113,  112,  110,  109,  111,   16,
 /*  1020 */   313,  314,  321,  263,   17,   78,  189,  296,  297,  265,
 /*  1030 */   760,  264,  219,  974,   40,  164,  264,  986,  974,  301,
 /*  1040 */   302,    6,  165,   90,   87,   86,   85,   89,   88,   84,
 /*  1050 */    83,   81,   80,   79,   82,   77,  749,   91,   92,   98,
 /*  1060 */    97,   96,  104,  103,  102,  101,  100,   99,  105,   93,
 /*  1070 */    94,   95,  108,  107,  106,  116,  117,  114,  113,  112,
 /*  1080 */   110,  109,  111, 1055,  871,  155,   50,  158, 1109,   78,
 /*  1090 */   860,  859,  858,  857,  856,  168,   54,  143,  323,  170,
 /*  1100 */   324,   20,  327, 1065,  330,    7,  794,   90,   87,   86,
 /*  1110 */    85,   89,   88,   84,   83,   81,   80,   79,   82,   77,
 /*  1120 */  1087,   91,   92,   98,   97,   96,  104,  103,  102,  101,
 /*  1130 */   100,   99,  105,   93,   94,   95,  108,  107,  106,  116,
 /*  1140 */   117,  114,  113,  112,  110,  109,  111,  326,  334,  938,
 /*  1150 */   329,  829,  937,   78,  936,   90,   87,   86,   85,   89,
 /*  1160 */    88,   84,   83,   81,   80,   79,   82,   77,  928,   91,
 /*  1170 */    92,   98,   97,   96,  104,  103,  102,  101,  100,   99,
 /*  1180 */   105,   93,   94,   95,  108,  107,  106,  116,  117,  114,
 /*  1190 */   113,  112,  110,  109,  111,  216,  404,  171,   56,   57,
 /*  1200 */   338,   78,  137,   43,   58,   59, 1097,   44,  177,   61,
 /*  1210 */  1088,   63,   64,   10,  178,   15,   66,    8,   67,   90,
 /*  1220 */    87,   86,   85,   89,   88,   84,   83,   81,   80,   79,
 /*  1230 */    82,   77,   35,   91,   92,   98,   97,   96,  104,  103,
 /*  1240 */   102,  101,  100,   99,  105,   93,   94,   95,  108,  107,
 /*  1250 */   106,  116,  117,  114,  113,  112,  110,  109,  111,  200,
 /*  1260 */   404,   46, 1060, 1059, 1058,   78,  363,   47,   48,  365,
 /*  1270 */   264,  216,  974,  364, 1057,  368,  370,  369, 1054,   69,
 /*  1280 */  1053,    9, 1052,   90,   87,   86,   85,   89,   88,   84,
 /*  1290 */    83,   81,   80,   79,   82,   77, 1051,   91,   92,   98,
 /*  1300 */    97,   96,  104,  103,  102,  101,  100,   99,  105,   93,
 /*  1310 */    94,   95,  108,  107,  106,  116,  117,  114,  113,  112,
 /*  1320 */   110,  109,  111,  201,  404, 1297,  377,  140,  294,   78,
 /*  1330 */   179,   70,  183,  184,  264,  200,  974,  132,   71,  386,
 /*  1340 */    25,  389,  390,  388,  391,   11, 1121,   90,   87,   86,
 /*  1350 */    85,   89,   88,   84,   83,   81,   80,   79,   82,   77,
 /*  1360 */   909,   91,   92,   98,   97,   96,  104,  103,  102,  101,
 /*  1370 */   100,   99,  105,   93,   94,   95,  108,  107,  106,  116,
 /*  1380 */   117,  114,  113,  112,  110,  109,  111,  914,   27,  908,
 /*  1390 */   144,  398,  130,   78,  923,  983,   78, 1040,  264,  201,
 /*  1400 */   974,  115,  903,   28,   29,  902,  905,  906,   30, 1070,
 /*  1410 */   901,   90,   87,   86,   85,   89,   88,   84,   83,   81,
 /*  1420 */    80,   79,   82,   77,  907,   91,   92,   98,   97,   96,
 /*  1430 */   104,  103,  102,  101,  100,   99,  105,   93,   94,   95,
 /*  1440 */   108,  107,  106,  116,  117,  114,  113,  112,  110,  109,
 /*  1450 */   111,  145,  404,   31,   32,  415,  899,   78,  898,  897,
 /*  1460 */   147,  118,  421,  187,  394,  195,  784,  427,  425,  426,
 /*  1470 */   428,  783,  429,   14,  432,   90,   87,   86,   85,   89,
 /*  1480 */    88,   84,   83,   81,   80,   79,   82,   77,  431,   91,
 /*  1490 */    92,   98,   97,   96,  104,  103,  102,  101,  100,   99,
 /*  1500 */   105,   93,   94,   95,  108,  107,  106,  116,  117,  114,
 /*  1510 */   113,  112,  110,  109,  111,  217,  404,  433,  434, 1230,
 /*  1520 */   436,   78,  437, 1229,  782,  142,  264,  145,  974,  781,
 /*  1530 */   441,  443, 1139, 1139, 1139, 1139, 1139,  981, 1139,   90,
 /*  1540 */    87,   86,   85,   89,   88,   84,   83,   81,   80,   79,
 /*  1550 */    82,   77, 1139,   91,   92,   98,   97,   96,  104,  103,
 /*  1560 */   102,  101,  100,   99,  105,   93,   94,   95,  108,  107,
 /*  1570 */   106,  116,  117,  114,  113,  112,  110,  109,  111, 1139,
 /*  1580 */  1139, 1139, 1139, 1139, 1139,   78, 1139, 1139, 1139, 1139,
 /*  1590 */   264,  217,  974, 1139, 1139, 1139, 1139, 1139, 1139, 1139,
 /*  1600 */  1139,  975, 1139,   90,   87,   86,   85,   89,   88,   84,
 /*  1610 */    83,   81,   80,   79,   82,   77, 1139,   91,   92,   98,
 /*  1620 */    97,   96,  104,  103,  102,  101,  100,   99,  105,   93,
 /*  1630 */    94,   95,  108,  107,  106,  116,  117,  114,  113,  112,
 /*  1640 */   110,  109,  111, 1139, 1139, 1139, 1139, 1139, 1139,   78,
 /*  1650 */  1139,   90,   87,   86,   85,   89,   88,   84,   83,   81,
 /*  1660 */    80,   79,   82,   77,   76,   91,   92,   98,   97,   96,
 /*  1670 */   104,  103,  102,  101,  100,   99,  105,   93,   94,   95,
 /*  1680 */   108,  107,  106,  116,  117,  114,  113,  112,  110,  109,
 /*  1690 */   111,  222,  404, 1139,  218,  404, 1139,   78,  104,  103,
 /*  1700 */   102,  101,  100,   99,  105,   93,   94,   95,  108,  107,
 /*  1710 */   106,  116,  117,  114,  113,  112,  110,  109,  111, 1139,
 /*  1720 */  1139, 1139, 1139, 1139,  982,   78,   90,   87,   86,   85,
 /*  1730 */    89,   88,   84,   83,   81,   80,   79,   82,   77,  281,
 /*  1740 */    91,   92,   98,   97,   96,  104,  103,  102,  101,  100,
 /*  1750 */    99,  105,   93,   94,   95,  108,  107,  106,  116,  117,
 /*  1760 */   114,  113,  112,  110,  109,  111,  264,  222,  974,  264,
 /*  1770 */   218,  974,   78, 1139,  803, 1139,   90,   87,   86,   85,
 /*  1780 */    89,   88,   84,   83,   81,   80,   79,   82,   77, 1139,
 /*  1790 */    91,   92,   98,   97,   96,  104,  103,  102,  101,  100,
 /*  1800 */    99,  105,   93,   94,   95,  108,  107,  106,  116,  117,
 /*  1810 */   114,  113,  112,  110,  109,  111,  167, 1139, 1139, 1139,
 /*  1820 */  1139, 1139,   78, 1139,  309, 1139, 1139, 1139, 1139, 1139,
 /*  1830 */   222,  404, 1139,  223,  404,   94,   95,  108,  107,  106,
 /*  1840 */   116,  117,  114,  113,  112,  110,  109,  111, 1056,   41,
 /*  1850 */   127,  126, 1139, 1139,   78, 1139, 1139, 1139,  895,  124,
 /*  1860 */   125, 1139,   33, 1139, 1139,  796,  798, 1139, 1139, 1139,
 /*  1870 */   392,  972, 1139,  316, 1139,    3,  799,  800,  278, 1039,
 /*  1880 */  1139, 1139,  918,  895,  919,  138,  869,  872,  873, 1139,
 /*  1890 */  1041, 1139,  904, 1139,  802, 1139, 1139,  801,  423,   37,
 /*  1900 */    37, 1139, 1139,  973, 1139,  264,  222,  974,  264,  223,
 /*  1910 */   974,  285, 1139, 1050, 1066, 1049,  375, 1066, 1066, 1066,
 /*  1920 */  1066, 1066, 1066, 1066,  374,  373,  372,  358,  223,  404,
 /*  1930 */   352, 1139, 1139, 1139, 1139, 1139, 1139, 1139,  420, 1139,
 /*  1940 */  1139,  895, 1139, 1056, 1139, 1139, 1139, 1139,   19, 1139,
 /*  1950 */  1139,  222,  404,  895, 1139,  123,  122,  121,  120, 1042,
 /*  1960 */  1043, 1044, 1045, 1046, 1047,  918,  895,  919,  141,  325,
 /*  1970 */     4,  406, 1139, 1139, 1139,  904, 1139,  918,  895,  919,
 /*  1980 */   138, 1137, 1139,  310, 1139, 1139,  813,  904,  442,  295,
 /*  1990 */  1139, 1139,  424, 1139,  422,  385,  169, 1139, 1139,  279,
 /*  2000 */  1139, 1139, 1139,  264,  223,  974, 1139, 1139, 1050, 1066,
 /*  2010 */  1049,  375, 1066, 1066, 1066, 1066, 1066, 1066, 1066,  374,
 /*  2020 */   373,  372,  358, 1139, 1139,  352,  264,  222,  974, 1099,
 /*  2030 */  1139, 1139,  188,  267,  305,  359,  105,   93,   94,   95,
 /*  2040 */   108,  107,  106,  116,  117,  114,  113,  112,  110,  109,
 /*  2050 */   111,  298,  223,  404, 1139,  895,  424,   78,  422,  385,
 /*  2060 */  1139, 1139, 1139, 1139, 1139,   40, 1139, 1056, 1139, 1139,
 /*  2070 */  1139, 1139, 1139, 1139, 1139, 1139, 1139,  895, 1139,  918,
 /*  2080 */   895,  919,  141,  895, 1139,  408, 1139,  798,  359,  904,
 /*  2090 */  1139, 1139,  317, 1139, 1139, 1139, 1139,  799,  800,  222,
 /*  2100 */   404,  918,  895,  919,  138,  223,  404,  918,  895,  919,
 /*  2110 */   141,  904, 1139,  409, 1139,  802, 1139,  904,  801, 1139,
 /*  2120 */  1056,  797, 1139, 1139, 1139, 1139,  895,  264,  223,  974,
 /*  2130 */   895, 1139, 1050, 1067, 1049,  375, 1067, 1067, 1067, 1067,
 /*  2140 */  1067, 1067, 1067,  374,  373,  372,  358,  282, 1139,  352,
 /*  2150 */   918,  895,  919,  275,  918,  895,  919,  138, 1139, 1139,
 /*  2160 */   904, 1139, 1139, 1139,  904,  226,  404, 1139, 1139,   19,
 /*  2170 */  1139, 1139, 1139, 1139,  264,  222,  974, 1139, 1139,  895,
 /*  2180 */   264,  223,  974,  359, 1139, 1050, 1067, 1049,  375, 1067,
 /*  2190 */  1067, 1067, 1067, 1067, 1067, 1067,  374,  373,  372,  358,
 /*  2200 */   223,  404,  352,  918,  895,  919,  141, 1139,  287,  411,
 /*  2210 */  1139, 1139, 1139,  904, 1139, 1056, 1139, 1139, 1139, 1139,
 /*  2220 */  1139, 1139, 1139,  223,  404,  895,   93,   94,   95,  108,
 /*  2230 */   107,  106,  116,  117,  114,  113,  112,  110,  109,  111,
 /*  2240 */   264,  226,  974,  287,  287,  895,   78, 1139, 1139,  918,
 /*  2250 */   895,  919,  138, 1139, 1139, 1139, 1139, 1139, 1139,  904,
 /*  2260 */  1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139,  918,
 /*  2270 */   895,  919,  396, 1139, 1139,  264,  223,  974, 1139,  904,
 /*  2280 */  1050, 1075, 1049,  375, 1075, 1075, 1075, 1075, 1075, 1075,
 /*  2290 */  1075,  374,  373,  372,  358,  359, 1139,  352,  264,  223,
 /*  2300 */   974, 1139, 1139, 1139, 1139, 1139, 1081,  222,  404, 1139,
 /*  2310 */  1139, 1139,  223,  404, 1139, 1139, 1139, 1139, 1139,  280,
 /*  2320 */   127,  126, 1139, 1139, 1139, 1139, 1139, 1056, 1139,  124,
 /*  2330 */   125,  265,   33, 1139, 1139, 1139,  877,  895, 1139, 1139,
 /*  2340 */  1139,  972, 1139,  895,  202,  404, 1139, 1139,  359, 1039,
 /*  2350 */  1139, 1139, 1139, 1139, 1139,  283, 1139, 1139, 1139, 1139,
 /*  2360 */  1041,  918,  895,  919,  138,  223,  404,  918,  895,  919,
 /*  2370 */   284,  904, 1139,  973, 1139, 1139, 1139,  904, 1139, 1139,
 /*  2380 */  1056, 1139,  264,  222,  974, 1139, 1139,  264,  223,  974,
 /*  2390 */   895, 1139, 1050, 1074, 1049,  375, 1074, 1074, 1074, 1074,
 /*  2400 */  1074, 1074, 1074,  374,  373,  372,  358,  203,  404,  352,
 /*  2410 */  1139, 1139, 1139, 1139,  918,  895,  919,  138, 1139,  264,
 /*  2420 */   202,  974, 1139, 1139,  904,  123,  122,  121,  120, 1042,
 /*  2430 */  1043, 1044, 1045, 1046, 1047, 1139, 1139, 1139,  223,  404,
 /*  2440 */   264,  223,  974,  359,  895, 1050, 1073, 1049,  375, 1073,
 /*  2450 */  1073, 1073, 1073, 1073, 1073, 1073,  374,  373,  372,  358,
 /*  2460 */   223,  404,  352,  895, 1139, 1139, 1139, 1139,  918,  895,
 /*  2470 */   919,  141, 1139, 1139,  412, 1056, 1139, 1139,  904, 1139,
 /*  2480 */  1139, 1139,  264,  203,  974,  895, 1139,  918,  895,  919,
 /*  2490 */   138, 1139, 1139, 1139, 1139, 1139, 1139,  904, 1139,  204,
 /*  2500 */   404, 1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139,  918,
 /*  2510 */   895,  919,  138,  264,  223,  974, 1139, 1139, 1139,  904,
 /*  2520 */  1139, 1081, 1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139,
 /*  2530 */  1139, 1078, 1139,  355,  342,  264,  223,  974, 1139, 1139,
 /*  2540 */  1050, 1086, 1049,  375, 1086, 1086, 1086, 1086, 1086, 1086,
 /*  2550 */  1086,  374,  373,  372,  358,  359, 1139,  352, 1139,  220,
 /*  2560 */   404, 1139,  205,  404, 1139, 1139, 1139,  152,  153, 1139,
 /*  2570 */   152,  153,  223,  404,  264,  204,  974, 1139, 1139, 1139,
 /*  2580 */   127,  126, 1139,  870, 1139, 1139,  870, 1056,  762,  124,
 /*  2590 */   125,  762,   33, 1139, 1139, 1139, 1061,  895, 1139,  763,
 /*  2600 */  1139,  972,  763,  865,  206,  404,  866, 1139,  359, 1039,
 /*  2610 */  1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139,
 /*  2620 */  1041,  918,  895,  919,  138,  223,  404, 1139, 1139, 1139,
 /*  2630 */  1139,  904, 1139,  973,  264,  220,  974,  264,  205,  974,
 /*  2640 */  1056, 1139, 1139, 1139, 1139, 1139, 1139,  264,  223,  974,
 /*  2650 */   895, 1139, 1050, 1085, 1049,  375, 1085, 1085, 1085, 1085,
 /*  2660 */  1085, 1085, 1085,  374,  373,  372,  358,  207,  404,  352,
 /*  2670 */  1139, 1139, 1139, 1139,  918,  895,  919,  138, 1139,  264,
 /*  2680 */   206,  974, 1139, 1139,  904,  123,  122,  121,  120, 1042,
 /*  2690 */  1043, 1044, 1045, 1046, 1047, 1139, 1139, 1139,  208,  404,
 /*  2700 */   264,  223,  974,  359, 1139, 1050, 1084, 1049,  375, 1084,
 /*  2710 */  1084, 1084, 1084, 1084, 1084, 1084,  374,  373,  372,  358,
 /*  2720 */   223,  404,  352, 1139, 1139, 1139, 1139, 1139, 1139, 1139,
 /*  2730 */  1139, 1139, 1139, 1139, 1139, 1056, 1139, 1139, 1139, 1139,
 /*  2740 */  1139, 1139,  264,  207,  974,  895,  209,  404, 1139,  210,
 /*  2750 */   404, 1139,  221,  404, 1139, 1139, 1139, 1139, 1139,  224,
 /*  2760 */   404, 1139,  225,  404, 1139, 1139, 1139,  211,  404,  918,
 /*  2770 */   895,  919,  138,  264,  208,  974, 1139, 1139, 1139,  904,
 /*  2780 */  1139, 1139, 1139,  212,  404, 1139, 1139,  197,  404, 1139,
 /*  2790 */  1139, 1139, 1139, 1139, 1139,  264,  223,  974, 1139, 1139,
 /*  2800 */  1050, 1072, 1049,  375, 1072, 1072, 1072, 1072, 1072, 1072,
 /*  2810 */  1072,  374,  373,  372,  358,  359, 1139,  352, 1139,  198,
 /*  2820 */   404,  264,  209,  974,  264,  210,  974,  264,  221,  974,
 /*  2830 */  1139, 1139,  223,  404,  264,  224,  974,  264,  225,  974,
 /*  2840 */   127,  126,  264,  211,  974, 1139, 1139, 1056, 1139,  124,
 /*  2850 */   125, 1139,   33, 1139, 1139,  213,  404,  895,  264,  212,
 /*  2860 */   974,  972,  264,  197,  974,  241,  404, 1139,  359, 1039,
 /*  2870 */  1139, 1139, 1139, 1139, 1139,   34, 1139, 1139, 1139, 1139,
 /*  2880 */  1041,  918,  895,  919,  138,  223,  404, 1139, 1139, 1139,
 /*  2890 */  1139,  904, 1139,  973,  264,  198,  974, 1139, 1139, 1139,
 /*  2900 */  1056, 1139, 1139, 1139, 1139, 1139, 1139,  264,  223,  974,
 /*  2910 */   895, 1139, 1050, 1071, 1049,  375, 1071, 1071, 1071, 1071,
 /*  2920 */  1071, 1071, 1071,  374,  373,  372,  358, 1139, 1139,  352,
 /*  2930 */   264,  213,  974, 1139,  918,  895,  919,  138,  214,  404,
 /*  2940 */   264,  241,  974, 1139,  904,  123,  122,  121,  120, 1042,
 /*  2950 */  1043, 1044, 1045, 1046, 1047, 1139, 1139, 1139, 1032,  404,
 /*  2960 */   264,  223,  974,  359, 1139, 1050,  160, 1049,  375,  160,
 /*  2970 */   160,  160,  160,  160,  160,  160,  374,  373,  372,  358,
 /*  2980 */   223,  404,  352, 1139, 1139, 1139, 1139, 1139, 1139, 1139,
 /*  2990 */  1139, 1139, 1139, 1139, 1139, 1056, 1139,  127,  126, 1139,
 /*  3000 */  1139, 1139, 1139,  227,  404,  895,  124,  125, 1139,   33,
 /*  3010 */  1139,  228,  404,  264,  214,  974,  229,  404,  405,  230,
 /*  3020 */   404, 1139, 1139, 1139, 1139, 1139, 1039, 1139, 1139,  918,
 /*  3030 */   895,  919,  138,  264, 1032,  974, 1139, 1041, 1139,  904,
 /*  3040 */   231,  404, 1139,  232,  404, 1139, 1139,  233,  404, 1139,
 /*  3050 */   973, 1139, 1139, 1139, 1139,  264,  223,  974, 1139, 1139,
 /*  3060 */  1050, 1069, 1049,  375, 1069, 1069, 1069, 1069, 1069, 1069,
 /*  3070 */  1069,  374,  373,  372,  358,  162, 1139,  352,  264,  227,
 /*  3080 */   974, 1139, 1139, 1139,  186,  896,  264,  228,  974, 1139,
 /*  3090 */  1139,  264,  229,  974,  264,  230,  974, 1139,  893,  234,
 /*  3100 */   404, 1139,  123,  122,  121,  120, 1042, 1043, 1044, 1045,
 /*  3110 */  1046, 1047, 1139, 1139, 1139,  264,  231,  974,  264,  232,
 /*  3120 */   974, 1139,  264,  233,  974,  878,  879,  880,  881,  882,
 /*  3130 */   883,  884,  885,  886,  887,  888,  889,  890,  891,  892,
 /*  3140 */   894,  419,  418, 1139,  413,  410,  407, 1139,  235,  404,
 /*  3150 */  1139, 1139,  420,  236,  404, 1139,  237,  404, 1139, 1139,
 /*  3160 */  1139, 1139, 1139, 1139, 1139, 1139,  969,  970,  971, 1139,
 /*  3170 */  1139, 1139, 1139, 1139,  264,  234,  974,  238,  404, 1139,
 /*  3180 */    77, 1139,   91,   92,   98,   97,   96,  104,  103,  102,
 /*  3190 */   101,  100,   99,  105,   93,   94,   95,  108,  107,  106,
 /*  3200 */   116,  117,  114,  113,  112,  110,  109,  111, 1139, 1139,
 /*  3210 */   242,  404,  162,  895,   78, 1139,  243,  404, 1139,  254,
 /*  3220 */   404,  186,  896,  264,  235,  974,  255,  404,  264,  236,
 /*  3230 */   974,  264,  237,  974, 1139,  893, 1139,  918,  895,  919,
 /*  3240 */   141,  395,  131,  395,  256,  404, 1139,  904,  244,  404,
 /*  3250 */  1139, 1139,  264,  238,  974, 1139,  245,  404, 1139,  933,
 /*  3260 */  1139, 1139,  878,  879,  880,  881,  882,  883,  884,  885,
 /*  3270 */   886,  887,  888,  889,  890,  891,  892,  894,  419,  418,
 /*  3280 */  1139,  413,  410,  407,  922,  264,  242,  974, 1139,  420,
 /*  3290 */  1139,  264,  243,  974,  264,  254,  974, 1139, 1139, 1139,
 /*  3300 */  1139,  264,  255,  974,  246,  404, 1139,  162, 1139, 1139,
 /*  3310 */  1139, 1139, 1139, 1139, 1139, 1139,  186,  896, 1139,  264,
 /*  3320 */   256,  974, 1139,  264,  244,  974, 1139, 1139, 1139, 1139,
 /*  3330 */   893,  264,  245,  974, 1139,  247,  404, 1139, 1139, 1139,
 /*  3340 */  1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139,
 /*  3350 */  1139, 1139, 1139, 1139, 1139, 1139, 1139,  878,  879,  880,
 /*  3360 */   881,  882,  883,  884,  885,  886,  887,  888,  889,  890,
 /*  3370 */   891,  892,  894,  419,  418, 1139,  413,  410,  407,  264,
 /*  3380 */   246,  974,  248,  404,  420,   92,   98,   97,   96,  104,
 /*  3390 */   103,  102,  101,  100,   99,  105,   93,   94,   95,  108,
 /*  3400 */   107,  106,  116,  117,  114,  113,  112,  110,  109,  111,
 /*  3410 */   264,  247,  974, 1139,  249,  404,   78,   98,   97,   96,
 /*  3420 */   104,  103,  102,  101,  100,   99,  105,   93,   94,   95,
 /*  3430 */   108,  107,  106,  116,  117,  114,  113,  112,  110,  109,
 /*  3440 */   111,  127,  126,  250,  404, 1139, 1139,   78,  251,  404,
 /*  3450 */   124,  125, 1139,   33,  252,  404, 1139,  264,  248,  974,
 /*  3460 */   253,  404,  972,  257,  404, 1139,  258,  404, 1139, 1139,
 /*  3470 */  1039, 1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139,
 /*  3480 */  1139, 1041,  259,  404, 1139, 1139,  288,  404, 1139,  264,
 /*  3490 */   249,  974, 1139, 1139,  973,  289,  404, 1139,  403,  404,
 /*  3500 */  1139, 1139,  290,  404, 1139, 1139, 1139, 1139, 1139,  291,
 /*  3510 */   404, 1139,  292,  404, 1139,  239,  404, 1139,  264,  250,
 /*  3520 */   974, 1139, 1139,  264,  251,  974, 1139, 1139, 1139,  264,
 /*  3530 */   252,  974, 1139, 1139, 1139,  264,  253,  974,  264,  257,
 /*  3540 */   974,  264,  258,  974,  262,  404,  123,  122,  121,  120,
 /*  3550 */  1042, 1043, 1044, 1045, 1046, 1047, 1139,  264,  259,  974,
 /*  3560 */  1139,  264,  288,  974,  261,  404, 1139,  215,  404, 1139,
 /*  3570 */   264,  289,  974,  264,  403,  974, 1139,  264,  290,  974,
 /*  3580 */  1139, 1139, 1139, 1139,  264,  291,  974,  264,  292,  974,
 /*  3590 */   264,  239,  974, 1139, 1139, 1139, 1139, 1139, 1139, 1139,
 /*  3600 */  1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139,
 /*  3610 */  1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139,  264,
 /*  3620 */   262,  974, 1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139,
 /*  3630 */  1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139, 1139,  264,
 /*  3640 */   261,  974,  264,  215,  974,
};
static const YYCODETYPE yy_lookahead[] = {
 /*     0 */    33,   34,  238,  239,  240,   52,  242,   63,   49,   42,
 /*    10 */    43,   44,   45,   46,  178,  179,   49,  213,  214,   66,
 /*    20 */    53,   54,   69,   70,   71,   72,   54,   68,   75,   62,
 /*    30 */    77,   78,   79,   66,   67,   68,   69,   65,  235,  236,
 /*    40 */    73,   29,   30,   31,   32,   33,   34,   35,   36,   37,
 /*    50 */    38,   39,   40,   86,  240,  161,  242,  221,   86,   47,
 /*    60 */   116,   94,   95,   96,   97,   98,   99,  100,  101,  102,
 /*    70 */   103,  104,  105,  106,  107,  108,  109,  110,  111,   54,
 /*    80 */   113,  114,  115,   69,  208,  209,  210,  211,  121,  253,
 /*    90 */   254,  255,  256,  257,   80,   81,   82,   83,   84,   85,
 /*   100 */    63,  148,  179,   54,  203,  138,  139,  140,  141,  142,
 /*   110 */   143,  144,  145,  146,  147,  148,  149,  150,  151,  152,
 /*   120 */   153,  154,  155,  156,  157,  242,  159,  160,  227,  228,
 /*   130 */   229,  230,  231,  232,  233,  252,  162,  243,  237,   33,
 /*   140 */    34,  247,  248,   45,   46,  171,  172,  173,   42,   43,
 /*   150 */    44,   45,   46,  116,  134,   49,  182,   54,   50,   53,
 /*   160 */    54,  179,  188,  189,  190,  191,  192,  193,   62,   66,
 /*   170 */    67,   63,   66,   67,   68,   69,  253,  254,  255,   73,
 /*   180 */    30,   31,   32,   33,   34,   35,   36,   37,   38,   39,
 /*   190 */    40,   54,   86,  161,   46,  235,  236,   47,  166,  167,
 /*   200 */    94,   95,   96,   97,   98,   99,  100,  101,  102,  103,
 /*   210 */   104,  105,  106,  107,  108,  109,  110,  111,  244,  113,
 /*   220 */   114,  115,   42,   43,   44,   45,   46,  121,  168,  169,
 /*   230 */   170,  171,  172,  173,  174,  253,  254,  255,  164,  165,
 /*   240 */   180,  179,  276,  277,  138,  139,  140,  141,  142,  143,
 /*   250 */   144,  145,  146,  147,  148,  149,  150,  151,  152,  153,
 /*   260 */   154,  155,  156,  157,  204,  159,  160,   33,   34,   35,
 /*   270 */    36,   37,   38,   39,   40,   54,  187,  216,   33,   34,
 /*   280 */     1,   47,  161,   14,   29,  161,  197,   42,   43,   44,
 /*   290 */    45,   46,  178,  179,   49,  178,  179,   54,   53,   54,
 /*   300 */   235,  236,  178,  179,  155,  156,   48,   62,   65,   54,
 /*   310 */    52,   66,   67,   68,  203,  253,  254,  255,   73,   50,
 /*   320 */   199,  200,  201,   54,   66,   46,   49,   50,   70,   71,
 /*   330 */    72,   86,  258,   75,  161,   77,   78,   79,  221,   94,
 /*   340 */    95,   96,   97,   98,   99,  100,  101,  102,  103,  104,
 /*   350 */   105,  106,  107,  108,  109,  110,  111,   49,  113,  114,
 /*   360 */   115,  237,  178,  179,    1,  186,  121,  253,  254,  255,
 /*   370 */   253,  254,  255,  256,  257,  261,   68,  253,  254,  255,
 /*   380 */   207,   49,   50,  138,  139,  140,  141,  142,  143,  144,
 /*   390 */   145,  146,  147,  148,  149,  150,  151,  152,  153,  154,
 /*   400 */   155,  156,  157,   44,  159,  160,  148,   54,   53,   46,
 /*   410 */   226,   14,   53,   54,   55,   56,   57,   58,   59,  242,
 /*   420 */    61,   68,  245,  241,   65,  179,   67,   49,  251,  252,
 /*   430 */    75,  161,   50,   74,   50,   76,  179,  253,  254,  255,
 /*   440 */    87,   88,   89,   90,   91,   92,   93,   69,  178,  179,
 /*   450 */   248,   69,  186,   94,   95,   96,   97,   98,   99,  100,
 /*   460 */   101,  102,  103,  104,  105,  106,  107,  108,  109,  110,
 /*   470 */   111,  187,  113,  114,  115,  116,   50,   80,   49,   50,
 /*   480 */   121,  197,  123,  124,  125,  126,  127,  128,  129,  130,
 /*   490 */   131,  132,  133,   33,   34,   69,  226,  161,  162,  253,
 /*   500 */   254,  255,   42,   43,   44,   45,  122,  162,  172,  173,
 /*   510 */   253,  254,  255,   53,   54,   50,  171,  172,  173,  177,
 /*   520 */   178,  179,   62,  253,  254,  255,  190,   67,   63,  193,
 /*   530 */   194,  195,  196,   73,  189,  190,  191,  192,  193,   55,
 /*   540 */    56,   57,   58,   59,   60,   61,   86,  186,   64,  212,
 /*   550 */   186,  206,  203,  186,   94,   95,   96,   97,   98,   99,
 /*   560 */   100,  101,  102,  103,  104,  105,  106,  107,  108,  109,
 /*   570 */   110,  111,   50,  113,  114,  115,  227,  228,  229,  230,
 /*   580 */   244,  121,  233,  178,  179,   63,  237,   40,  186,  244,
 /*   590 */   177,  178,  179,  212,   47,  253,  254,  255,  138,  139,
 /*   600 */   140,  141,  142,  143,  144,  145,  146,  147,    1,    2,
 /*   610 */     3,    4,    5,    6,    7,    8,    9,   10,   11,   12,
 /*   620 */    13,  212,   15,   16,   17,   18,   19,   20,   21,   22,
 /*   630 */    23,   24,   25,   26,   27,   28,   29,   30,   31,   32,
 /*   640 */    33,   34,   35,   36,   37,   38,   39,   40,  186,   14,
 /*   650 */   236,   50,   50,   34,   47,  272,   50,   50,  253,  254,
 /*   660 */   255,  256,  257,   44,   63,   63,  253,  254,  255,   63,
 /*   670 */    63,  186,    1,    2,    3,    4,    5,    6,    7,    8,
 /*   680 */     9,   10,   11,   12,   13,   50,   15,   16,   17,   18,
 /*   690 */    19,   20,   21,   22,   23,   24,   25,   26,   27,   28,
 /*   700 */    29,   30,   31,   32,   33,   34,   35,   36,   37,   38,
 /*   710 */    39,   40,   35,   36,   37,   38,   39,   40,   47,   50,
 /*   720 */   186,   50,  183,  187,   47,   94,   95,   96,   97,   98,
 /*   730 */    99,   14,   63,   54,   63,   54,    1,    2,    3,    4,
 /*   740 */     5,    6,    7,    8,    9,   10,   11,   12,   13,  184,
 /*   750 */    15,   16,   17,   18,   19,   20,   21,   22,   23,   24,
 /*   760 */    25,   26,   27,   28,   29,   30,   31,   32,   33,   34,
 /*   770 */    35,   36,   37,   38,   39,   40,   57,   58,   59,  179,
 /*   780 */    61,   69,   47,  179,   72,   50,   46,   68,  161,   77,
 /*   790 */    78,  110,   45,  215,   54,   76,   62,   49,   63,  218,
 /*   800 */     1,    2,    3,    4,    5,    6,    7,    8,    9,   10,
 /*   810 */    11,   12,   13,  186,   15,   16,   17,   18,   19,   20,
 /*   820 */    21,   22,   23,   24,   25,   26,   27,   28,   29,   30,
 /*   830 */    31,   32,   33,   34,   35,   36,   37,   38,   39,   40,
 /*   840 */    21,   54,  219,   45,   49,  217,   47,  175,  250,  222,
 /*   850 */   223,  224,  225,  253,  254,  255,   50,  253,  254,  255,
 /*   860 */   148,   74,   63,   44,    1,    2,    3,    4,    5,    6,
 /*   870 */     7,    8,    9,   10,   11,   12,   13,  187,   15,   16,
 /*   880 */    17,   18,   19,   20,   21,   22,   23,   24,   25,   26,
 /*   890 */    27,   28,   29,   30,   31,   32,   33,   34,   35,   36,
 /*   900 */    37,   38,   39,   40,  217,   54,  248,  248,  248,   50,
 /*   910 */    47,  277,  248,   50,   77,    1,    2,    3,    4,    5,
 /*   920 */     6,    7,    8,    9,   10,   11,   12,   13,   69,   15,
 /*   930 */    16,   17,   18,   19,   20,   21,   22,   23,   24,   25,
 /*   940 */    26,   27,   28,   29,   30,   31,   32,   33,   34,   35,
 /*   950 */    36,   37,   38,   39,   40,   69,  178,  179,   72,  246,
 /*   960 */   201,   47,  179,   77,   78,   50,  176,  175,   14,  202,
 /*   970 */    20,  234,  175,   50,   62,  179,  205,   63,  185,    1,
 /*   980 */     2,    3,    4,    5,    6,    7,    8,    9,   10,   11,
 /*   990 */    12,   13,   69,   15,   16,   17,   18,   19,   20,   21,
 /*  1000 */    22,   23,   24,   25,   26,   27,   28,   29,   30,   31,
 /*  1010 */    32,   33,   34,   35,   36,   37,   38,   39,   40,  181,
 /*  1020 */   218,  218,  218,  198,  181,   47,   68,   54,   53,   44,
 /*  1030 */    63,  253,  254,  255,  148,   45,  253,  254,  255,   62,
 /*  1040 */    45,   63,   54,    1,    2,    3,    4,    5,    6,    7,
 /*  1050 */     8,    9,   10,   11,   12,   13,   54,   15,   16,   17,
 /*  1060 */    18,   19,   20,   21,   22,   23,   24,   25,   26,   27,
 /*  1070 */    28,   29,   30,   31,   32,   33,   34,   35,   36,   37,
 /*  1080 */    38,   39,   40,   49,   49,   47,   68,   47,   49,   47,
 /*  1090 */    49,   69,   69,   69,   62,   45,   45,   68,   54,   69,
 /*  1100 */    75,   68,   45,   69,   45,   63,   63,    1,    2,    3,
 /*  1110 */     4,    5,    6,    7,    8,    9,   10,   11,   12,   13,
 /*  1120 */    14,   15,   16,   17,   18,   19,   20,   21,   22,   23,
 /*  1130 */    24,   25,   26,   27,   28,   29,   30,   31,   32,   33,
 /*  1140 */    34,   35,   36,   37,   38,   39,   40,   62,   45,   63,
 /*  1150 */    62,   54,   63,   47,   62,    1,    2,    3,    4,    5,
 /*  1160 */     6,    7,    8,    9,   10,   11,   12,   13,   49,   15,
 /*  1170 */    16,   17,   18,   19,   20,   21,   22,   23,   24,   25,
 /*  1180 */    26,   27,   28,   29,   30,   31,   32,   33,   34,   35,
 /*  1190 */    36,   37,   38,   39,   40,  178,  179,   62,    1,    1,
 /*  1200 */    54,   47,   50,   50,   14,   14,   69,   68,   50,    1,
 /*  1210 */    14,   45,   45,  158,   45,   45,   45,   63,   45,    1,
 /*  1220 */     2,    3,    4,    5,    6,    7,    8,    9,   10,   11,
 /*  1230 */    12,   13,   49,   15,   16,   17,   18,   19,   20,   21,
 /*  1240 */    22,   23,   24,   25,   26,   27,   28,   29,   30,   31,
 /*  1250 */    32,   33,   34,   35,   36,   37,   38,   39,   40,  178,
 /*  1260 */   179,   49,   49,   49,   49,   47,    1,   68,   68,  122,
 /*  1270 */   253,  254,  255,   54,   49,    1,   46,  122,   49,    1,
 /*  1280 */    49,   63,   49,    1,    2,    3,    4,    5,    6,    7,
 /*  1290 */     8,    9,   10,   11,   12,   13,   49,   15,   16,   17,
 /*  1300 */    18,   19,   20,   21,   22,   23,   24,   25,   26,   27,
 /*  1310 */    28,   29,   30,   31,   32,   33,   34,   35,   36,   37,
 /*  1320 */    38,   39,   40,  178,  179,   54,  122,   46,   20,   47,
 /*  1330 */    63,    1,   69,   54,  253,  254,  255,   68,   45,   63,
 /*  1340 */    45,   21,   54,   63,   20,   63,   63,    1,    2,    3,
 /*  1350 */     4,    5,    6,    7,    8,    9,   10,   11,   12,   13,
 /*  1360 */    21,   15,   16,   17,   18,   19,   20,   21,   22,   23,
 /*  1370 */    24,   25,   26,   27,   28,   29,   30,   31,   32,   33,
 /*  1380 */    34,   35,   36,   37,   38,   39,   40,   29,   50,   21,
 /*  1390 */    50,   63,   45,   47,  122,   54,   47,   62,  253,  254,
 /*  1400 */   255,   14,   21,   20,   50,   21,   58,   59,   20,   63,
 /*  1410 */    21,    1,    2,    3,    4,    5,    6,    7,    8,    9,
 /*  1420 */    10,   11,   12,   13,   76,   15,   16,   17,   18,   19,
 /*  1430 */    20,   21,   22,   23,   24,   25,   26,   27,   28,   29,
 /*  1440 */    30,   31,   32,   33,   34,   35,   36,   37,   38,   39,
 /*  1450 */    40,  178,  179,   50,   20,   54,   21,   47,   21,   54,
 /*  1460 */    20,    1,   54,   69,  116,   54,   49,   54,   73,   50,
 /*  1470 */    14,   49,   54,   63,   50,    1,    2,    3,    4,    5,
 /*  1480 */     6,    7,    8,    9,   10,   11,   12,   13,   73,   15,
 /*  1490 */    16,   17,   18,   19,   20,   21,   22,   23,   24,   25,
 /*  1500 */    26,   27,   28,   29,   30,   31,   32,   33,   34,   35,
 /*  1510 */    36,   37,   38,   39,   40,  178,  179,   54,   14,   50,
 /*  1520 */    54,   47,   50,   50,   49,   14,  253,  254,  255,   49,
 /*  1530 */    54,   54,  278,  278,  278,  278,  278,   63,  278,    1,
 /*  1540 */     2,    3,    4,    5,    6,    7,    8,    9,   10,   11,
 /*  1550 */    12,   13,  278,   15,   16,   17,   18,   19,   20,   21,
 /*  1560 */    22,   23,   24,   25,   26,   27,   28,   29,   30,   31,
 /*  1570 */    32,   33,   34,   35,   36,   37,   38,   39,   40,  278,
 /*  1580 */   278,  278,  278,  278,  278,   47,  278,  278,  278,  278,
 /*  1590 */   253,  254,  255,  278,  278,  278,  278,  278,  278,  278,
 /*  1600 */   278,   63,  278,    1,    2,    3,    4,    5,    6,    7,
 /*  1610 */     8,    9,   10,   11,   12,   13,  278,   15,   16,   17,
 /*  1620 */    18,   19,   20,   21,   22,   23,   24,   25,   26,   27,
 /*  1630 */    28,   29,   30,   31,   32,   33,   34,   35,   36,   37,
 /*  1640 */    38,   39,   40,  278,  278,  278,  278,  278,  278,   47,
 /*  1650 */   278,    1,    2,    3,    4,    5,    6,    7,    8,    9,
 /*  1660 */    10,   11,   12,   13,   14,   15,   16,   17,   18,   19,
 /*  1670 */    20,   21,   22,   23,   24,   25,   26,   27,   28,   29,
 /*  1680 */    30,   31,   32,   33,   34,   35,   36,   37,   38,   39,
 /*  1690 */    40,  178,  179,  278,  178,  179,  278,   47,   20,   21,
 /*  1700 */    22,   23,   24,   25,   26,   27,   28,   29,   30,   31,
 /*  1710 */    32,   33,   34,   35,   36,   37,   38,   39,   40,  278,
 /*  1720 */   278,  278,  278,  278,  122,   47,    1,    2,    3,    4,
 /*  1730 */     5,    6,    7,    8,    9,   10,   11,   12,   13,  226,
 /*  1740 */    15,   16,   17,   18,   19,   20,   21,   22,   23,   24,
 /*  1750 */    25,   26,   27,   28,   29,   30,   31,   32,   33,   34,
 /*  1760 */    35,   36,   37,   38,   39,   40,  253,  254,  255,  253,
 /*  1770 */   254,  255,   47,  278,   49,  278,    1,    2,    3,    4,
 /*  1780 */     5,    6,    7,    8,    9,   10,   11,   12,   13,  278,
 /*  1790 */    15,   16,   17,   18,   19,   20,   21,   22,   23,   24,
 /*  1800 */    25,   26,   27,   28,   29,   30,   31,   32,   33,   34,
 /*  1810 */    35,   36,   37,   38,   39,   40,  161,  278,  278,  278,
 /*  1820 */   278,  278,   47,  278,  161,  278,  278,  278,  278,  278,
 /*  1830 */   178,  179,  278,  178,  179,   28,   29,   30,   31,   32,
 /*  1840 */    33,   34,   35,   36,   37,   38,   39,   40,  193,  186,
 /*  1850 */    33,   34,  278,  278,   47,  278,  278,  278,  203,   42,
 /*  1860 */    43,  278,   45,  278,  278,  161,  162,  278,  278,  278,
 /*  1870 */    53,   54,  278,  218,  278,  220,  172,  173,  226,   62,
 /*  1880 */   278,  278,  227,  228,  229,  230,  223,  224,  225,  278,
 /*  1890 */    73,  278,  237,  278,  190,  278,  278,  193,  194,  195,
 /*  1900 */   196,  278,  278,   86,  278,  253,  254,  255,  253,  254,
 /*  1910 */   255,  161,  278,  258,  259,  260,  261,  262,  263,  264,
 /*  1920 */   265,  266,  267,  268,  269,  270,  271,  272,  178,  179,
 /*  1930 */   275,  278,  278,  278,  278,  278,  278,  278,  121,  278,
 /*  1940 */   278,  203,  278,  193,  278,  278,  278,  278,  244,  278,
 /*  1950 */   278,  178,  179,  203,  278,  138,  139,  140,  141,  142,
 /*  1960 */   143,  144,  145,  146,  147,  227,  228,  229,  230,   52,
 /*  1970 */   220,  233,  278,  278,  278,  237,  278,  227,  228,  229,
 /*  1980 */   230,    0,  278,   66,  278,  278,   69,  237,   71,   72,
 /*  1990 */   278,  278,   75,  278,   77,   78,   79,  278,  278,  226,
 /*  2000 */   278,  278,  278,  253,  254,  255,  278,  278,  258,  259,
 /*  2010 */   260,  261,  262,  263,  264,  265,  266,  267,  268,  269,
 /*  2020 */   270,  271,  272,  278,  278,  275,  253,  254,  255,   48,
 /*  2030 */   278,  278,   51,   52,   53,  161,   26,   27,   28,   29,
 /*  2040 */    30,   31,   32,   33,   34,   35,   36,   37,   38,   39,
 /*  2050 */    40,   70,  178,  179,  278,  203,   75,   47,   77,   78,
 /*  2060 */   278,  278,  278,  278,  278,  148,  278,  193,  278,  278,
 /*  2070 */   278,  278,  278,  278,  278,  278,  278,  203,  278,  227,
 /*  2080 */   228,  229,  230,  203,  278,  233,  278,  162,  161,  237,
 /*  2090 */   278,  278,  218,  278,  278,  278,  278,  172,  173,  178,
 /*  2100 */   179,  227,  228,  229,  230,  178,  179,  227,  228,  229,
 /*  2110 */   230,  237,  278,  233,  278,  190,  278,  237,  193,  278,
 /*  2120 */   193,  196,  278,  278,  278,  278,  203,  253,  254,  255,
 /*  2130 */   203,  278,  258,  259,  260,  261,  262,  263,  264,  265,
 /*  2140 */   266,  267,  268,  269,  270,  271,  272,  226,  278,  275,
 /*  2150 */   227,  228,  229,  230,  227,  228,  229,  230,  278,  278,
 /*  2160 */   237,  278,  278,  278,  237,  178,  179,  278,  278,  244,
 /*  2170 */   278,  278,  278,  278,  253,  254,  255,  278,  278,  203,
 /*  2180 */   253,  254,  255,  161,  278,  258,  259,  260,  261,  262,
 /*  2190 */   263,  264,  265,  266,  267,  268,  269,  270,  271,  272,
 /*  2200 */   178,  179,  275,  227,  228,  229,  230,  278,  221,  233,
 /*  2210 */   278,  278,  278,  237,  278,  193,  278,  278,  278,  278,
 /*  2220 */   278,  278,  278,  178,  179,  203,   27,   28,   29,   30,
 /*  2230 */    31,   32,   33,   34,   35,   36,   37,   38,   39,   40,
 /*  2240 */   253,  254,  255,  256,  257,  203,   47,  278,  278,  227,
 /*  2250 */   228,  229,  230,  278,  278,  278,  278,  278,  278,  237,
 /*  2260 */   278,  278,  278,  278,  278,  278,  278,  278,  278,  227,
 /*  2270 */   228,  229,  230,  278,  278,  253,  254,  255,  278,  237,
 /*  2280 */   258,  259,  260,  261,  262,  263,  264,  265,  266,  267,
 /*  2290 */   268,  269,  270,  271,  272,  161,  278,  275,  253,  254,
 /*  2300 */   255,  278,  278,  278,  278,  278,  261,  178,  179,  278,
 /*  2310 */   278,  278,  178,  179,  278,  278,  278,  278,  278,  274,
 /*  2320 */    33,   34,  278,  278,  278,  278,  278,  193,  278,   42,
 /*  2330 */    43,   44,   45,  278,  278,  278,   49,  203,  278,  278,
 /*  2340 */   278,   54,  278,  203,  178,  179,  278,  278,  161,   62,
 /*  2350 */   278,  278,  278,  278,  278,  226,  278,  278,  278,  278,
 /*  2360 */    73,  227,  228,  229,  230,  178,  179,  227,  228,  229,
 /*  2370 */   230,  237,  278,   86,  278,  278,  278,  237,  278,  278,
 /*  2380 */   193,  278,  253,  254,  255,  278,  278,  253,  254,  255,
 /*  2390 */   203,  278,  258,  259,  260,  261,  262,  263,  264,  265,
 /*  2400 */   266,  267,  268,  269,  270,  271,  272,  178,  179,  275,
 /*  2410 */   278,  278,  278,  278,  227,  228,  229,  230,  278,  253,
 /*  2420 */   254,  255,  278,  278,  237,  138,  139,  140,  141,  142,
 /*  2430 */   143,  144,  145,  146,  147,  278,  278,  278,  178,  179,
 /*  2440 */   253,  254,  255,  161,  203,  258,  259,  260,  261,  262,
 /*  2450 */   263,  264,  265,  266,  267,  268,  269,  270,  271,  272,
 /*  2460 */   178,  179,  275,  203,  278,  278,  278,  278,  227,  228,
 /*  2470 */   229,  230,  278,  278,  233,  193,  278,  278,  237,  278,
 /*  2480 */   278,  278,  253,  254,  255,  203,  278,  227,  228,  229,
 /*  2490 */   230,  278,  278,  278,  278,  278,  278,  237,  278,  178,
 /*  2500 */   179,  278,  278,  278,  278,  278,  278,  278,  278,  227,
 /*  2510 */   228,  229,  230,  253,  254,  255,  278,  278,  278,  237,
 /*  2520 */   278,  261,  278,  278,  278,  278,  278,  278,  278,  278,
 /*  2530 */   278,  271,  278,  273,  274,  253,  254,  255,  278,  278,
 /*  2540 */   258,  259,  260,  261,  262,  263,  264,  265,  266,  267,
 /*  2550 */   268,  269,  270,  271,  272,  161,  278,  275,  278,  178,
 /*  2560 */   179,  278,  178,  179,  278,  278,  278,   33,   34,  278,
 /*  2570 */    33,   34,  178,  179,  253,  254,  255,  278,  278,  278,
 /*  2580 */    33,   34,  278,   49,  278,  278,   49,  193,   54,   42,
 /*  2590 */    43,   54,   45,  278,  278,  278,   49,  203,  278,   65,
 /*  2600 */   278,   54,   65,   69,  178,  179,   69,  278,  161,   62,
 /*  2610 */   278,  278,  278,  278,  278,  278,  278,  278,  278,  278,
 /*  2620 */    73,  227,  228,  229,  230,  178,  179,  278,  278,  278,
 /*  2630 */   278,  237,  278,   86,  253,  254,  255,  253,  254,  255,
 /*  2640 */   193,  278,  278,  278,  278,  278,  278,  253,  254,  255,
 /*  2650 */   203,  278,  258,  259,  260,  261,  262,  263,  264,  265,
 /*  2660 */   266,  267,  268,  269,  270,  271,  272,  178,  179,  275,
 /*  2670 */   278,  278,  278,  278,  227,  228,  229,  230,  278,  253,
 /*  2680 */   254,  255,  278,  278,  237,  138,  139,  140,  141,  142,
 /*  2690 */   143,  144,  145,  146,  147,  278,  278,  278,  178,  179,
 /*  2700 */   253,  254,  255,  161,  278,  258,  259,  260,  261,  262,
 /*  2710 */   263,  264,  265,  266,  267,  268,  269,  270,  271,  272,
 /*  2720 */   178,  179,  275,  278,  278,  278,  278,  278,  278,  278,
 /*  2730 */   278,  278,  278,  278,  278,  193,  278,  278,  278,  278,
 /*  2740 */   278,  278,  253,  254,  255,  203,  178,  179,  278,  178,
 /*  2750 */   179,  278,  178,  179,  278,  278,  278,  278,  278,  178,
 /*  2760 */   179,  278,  178,  179,  278,  278,  278,  178,  179,  227,
 /*  2770 */   228,  229,  230,  253,  254,  255,  278,  278,  278,  237,
 /*  2780 */   278,  278,  278,  178,  179,  278,  278,  178,  179,  278,
 /*  2790 */   278,  278,  278,  278,  278,  253,  254,  255,  278,  278,
 /*  2800 */   258,  259,  260,  261,  262,  263,  264,  265,  266,  267,
 /*  2810 */   268,  269,  270,  271,  272,  161,  278,  275,  278,  178,
 /*  2820 */   179,  253,  254,  255,  253,  254,  255,  253,  254,  255,
 /*  2830 */   278,  278,  178,  179,  253,  254,  255,  253,  254,  255,
 /*  2840 */    33,   34,  253,  254,  255,  278,  278,  193,  278,   42,
 /*  2850 */    43,  278,   45,  278,  278,  178,  179,  203,  253,  254,
 /*  2860 */   255,   54,  253,  254,  255,  178,  179,  278,  161,   62,
 /*  2870 */   278,  278,  278,  278,  278,   68,  278,  278,  278,  278,
 /*  2880 */    73,  227,  228,  229,  230,  178,  179,  278,  278,  278,
 /*  2890 */   278,  237,  278,   86,  253,  254,  255,  278,  278,  278,
 /*  2900 */   193,  278,  278,  278,  278,  278,  278,  253,  254,  255,
 /*  2910 */   203,  278,  258,  259,  260,  261,  262,  263,  264,  265,
 /*  2920 */   266,  267,  268,  269,  270,  271,  272,  278,  278,  275,
 /*  2930 */   253,  254,  255,  278,  227,  228,  229,  230,  178,  179,
 /*  2940 */   253,  254,  255,  278,  237,  138,  139,  140,  141,  142,
 /*  2950 */   143,  144,  145,  146,  147,  278,  278,  278,  178,  179,
 /*  2960 */   253,  254,  255,  161,  278,  258,  259,  260,  261,  262,
 /*  2970 */   263,  264,  265,  266,  267,  268,  269,  270,  271,  272,
 /*  2980 */   178,  179,  275,  278,  278,  278,  278,  278,  278,  278,
 /*  2990 */   278,  278,  278,  278,  278,  193,  278,   33,   34,  278,
 /*  3000 */   278,  278,  278,  178,  179,  203,   42,   43,  278,   45,
 /*  3010 */   278,  178,  179,  253,  254,  255,  178,  179,   54,  178,
 /*  3020 */   179,  278,  278,  278,  278,  278,   62,  278,  278,  227,
 /*  3030 */   228,  229,  230,  253,  254,  255,  278,   73,  278,  237,
 /*  3040 */   178,  179,  278,  178,  179,  278,  278,  178,  179,  278,
 /*  3050 */    86,  278,  278,  278,  278,  253,  254,  255,  278,  278,
 /*  3060 */   258,  259,  260,  261,  262,  263,  264,  265,  266,  267,
 /*  3070 */   268,  269,  270,  271,  272,   44,  278,  275,  253,  254,
 /*  3080 */   255,  278,  278,  278,   53,   54,  253,  254,  255,  278,
 /*  3090 */   278,  253,  254,  255,  253,  254,  255,  278,   67,  178,
 /*  3100 */   179,  278,  138,  139,  140,  141,  142,  143,  144,  145,
 /*  3110 */   146,  147,  278,  278,  278,  253,  254,  255,  253,  254,
 /*  3120 */   255,  278,  253,  254,  255,   94,   95,   96,   97,   98,
 /*  3130 */    99,  100,  101,  102,  103,  104,  105,  106,  107,  108,
 /*  3140 */   109,  110,  111,  278,  113,  114,  115,  278,  178,  179,
 /*  3150 */   278,  278,  121,  178,  179,  278,  178,  179,  278,  278,
 /*  3160 */   278,  278,  278,  278,  278,  278,  135,  136,  137,  278,
 /*  3170 */   278,  278,  278,  278,  253,  254,  255,  178,  179,  278,
 /*  3180 */    13,  278,   15,   16,   17,   18,   19,   20,   21,   22,
 /*  3190 */    23,   24,   25,   26,   27,   28,   29,   30,   31,   32,
 /*  3200 */    33,   34,   35,   36,   37,   38,   39,   40,  278,  278,
 /*  3210 */   178,  179,   44,  203,   47,  278,  178,  179,  278,  178,
 /*  3220 */   179,   53,   54,  253,  254,  255,  178,  179,  253,  254,
 /*  3230 */   255,  253,  254,  255,  278,   67,  278,  227,  228,  229,
 /*  3240 */   230,  231,  232,  233,  178,  179,  278,  237,  178,  179,
 /*  3250 */   278,  278,  253,  254,  255,  278,  178,  179,  278,  249,
 /*  3260 */   278,  278,   94,   95,   96,   97,   98,   99,  100,  101,
 /*  3270 */   102,  103,  104,  105,  106,  107,  108,  109,  110,  111,
 /*  3280 */   278,  113,  114,  115,  116,  253,  254,  255,  278,  121,
 /*  3290 */   278,  253,  254,  255,  253,  254,  255,  278,  278,  278,
 /*  3300 */   278,  253,  254,  255,  178,  179,  278,   44,  278,  278,
 /*  3310 */   278,  278,  278,  278,  278,  278,   53,   54,  278,  253,
 /*  3320 */   254,  255,  278,  253,  254,  255,  278,  278,  278,  278,
 /*  3330 */    67,  253,  254,  255,  278,  178,  179,  278,  278,  278,
 /*  3340 */   278,  278,  278,  278,  278,  278,  278,  278,  278,  278,
 /*  3350 */   278,  278,  278,  278,  278,  278,  278,   94,   95,   96,
 /*  3360 */    97,   98,   99,  100,  101,  102,  103,  104,  105,  106,
 /*  3370 */   107,  108,  109,  110,  111,  278,  113,  114,  115,  253,
 /*  3380 */   254,  255,  178,  179,  121,   16,   17,   18,   19,   20,
 /*  3390 */    21,   22,   23,   24,   25,   26,   27,   28,   29,   30,
 /*  3400 */    31,   32,   33,   34,   35,   36,   37,   38,   39,   40,
 /*  3410 */   253,  254,  255,  278,  178,  179,   47,   17,   18,   19,
 /*  3420 */    20,   21,   22,   23,   24,   25,   26,   27,   28,   29,
 /*  3430 */    30,   31,   32,   33,   34,   35,   36,   37,   38,   39,
 /*  3440 */    40,   33,   34,  178,  179,  278,  278,   47,  178,  179,
 /*  3450 */    42,   43,  278,   45,  178,  179,  278,  253,  254,  255,
 /*  3460 */   178,  179,   54,  178,  179,  278,  178,  179,  278,  278,
 /*  3470 */    62,  278,  278,  278,  278,  278,  278,  278,  278,  278,
 /*  3480 */   278,   73,  178,  179,  278,  278,  178,  179,  278,  253,
 /*  3490 */   254,  255,  278,  278,   86,  178,  179,  278,  178,  179,
 /*  3500 */   278,  278,  178,  179,  278,  278,  278,  278,  278,  178,
 /*  3510 */   179,  278,  178,  179,  278,  178,  179,  278,  253,  254,
 /*  3520 */   255,  278,  278,  253,  254,  255,  278,  278,  278,  253,
 /*  3530 */   254,  255,  278,  278,  278,  253,  254,  255,  253,  254,
 /*  3540 */   255,  253,  254,  255,  178,  179,  138,  139,  140,  141,
 /*  3550 */   142,  143,  144,  145,  146,  147,  278,  253,  254,  255,
 /*  3560 */   278,  253,  254,  255,  178,  179,  278,  178,  179,  278,
 /*  3570 */   253,  254,  255,  253,  254,  255,  278,  253,  254,  255,
 /*  3580 */   278,  278,  278,  278,  253,  254,  255,  253,  254,  255,
 /*  3590 */   253,  254,  255,  278,  278,  278,  278,  278,  278,  278,
 /*  3600 */   278,  278,  278,  278,  278,  278,  278,  278,  278,  278,
 /*  3610 */   278,  278,  278,  278,  278,  278,  278,  278,  278,  253,
 /*  3620 */   254,  255,  278,  278,  278,  278,  278,  278,  278,  278,
 /*  3630 */   278,  278,  278,  278,  278,  278,  278,  278,  278,  253,
 /*  3640 */   254,  255,  253,  254,  255,
};
#define YY_SHIFT_USE_DFLT (3645)
#define YY_SHIFT_COUNT    (445)
#define YY_SHIFT_MIN      (-56)
#define YY_SHIFT_MAX      (3408)
static const short yy_shift_ofst[] = {
 /*     0 */  3645,  245,  -33,  245,  106,  245,  245,  245,  245,  245,
 /*    10 */   245,  245,  245,  245,  245,  460,  -47,  258, 1917,  359,
 /*    20 */   712,  712, 3168, 1981, 2964, 2964, 2964, 3263, 3263, 3263,
 /*    30 */  3263, 3263, 3263, 1817, 3408, 3408, 2964,  886, 3031, 3031,
 /*    40 */  3263, 2287, 2547, 3408, 3408, 3408, 3408, 3408, 3408, 3408,
 /*    50 */  2534, 2807, 3408, 3408, 3408, 3408, 3408, 3408, 3408, 3408,
 /*    60 */  3408, 3408, 3408, 3408, 3408, 3408, 3408, 3408, 3408, 3408,
 /*    70 */  3408, 3408, 3408, 3408, 3408, 3408, 3408, 3408, 3408, 3408,
 /*    80 */  3408, 3408, 3408, 3408, 3408, 3408, 3408, 3408, 3408, 3408,
 /*    90 */  3408, 3408, 3408, 3408, 3408, 3408, 3408, 3408, 3408, 3408,
 /*   100 */  3408, 3408, 3408, 3408, 3408, 3408, 3408, 3408, 3408, 3408,
 /*   110 */  3408, 3408, 3408, 3408, 3408, 3408, 3408, 3408, 3408, 2537,
 /*   120 */  3408, 3408, 3408, 3408, 3408, 3408, 3408, 3408,   14,  -56,
 /*   130 */    37,   25,   49,  -41,  353,  279,   98,   20,  137,  148,
 /*   140 */   221,  148,  221, 3645, 3645, 1775,  631, 1348,  -28,  269,
 /*   150 */   243,  243,  243,  243,  619,  243,  619,  619,  243,  363,
 /*   160 */   149,  243,  243,  308,  679,  717,  747,  748,  734,  798,
 /*   170 */   795,  806,  679,  798,  851,  851,  851,  137,  851,  837,
 /*   180 */   148,   49,  915,  795,  954,  148,  950,  795,  912, 3645,
 /*   190 */  3645, 3645, 3645, 3645, 3645, 3645, 3645,  607,  671,  735,
 /*   200 */   799,  863,  914,  978, 1042, 1106, 1154, 1218, 1282, 1346,
 /*   210 */  1410, 1474, 1538, 1602, 1650, 1725, 1775, 1775, 1775, 1775,
 /*   220 */  1775, 1775, 1775, 1775, 1775, 1775, 1775, 1775, 1775, 1775,
 /*   230 */  1775, 1775, 1775, 1775, 1775, 1775, 1775, 1775, 1775, 1775,
 /*   240 */  1775, 3167, 3369, 3400, 1678, 1678, 1678, 2010, 2010, 2010,
 /*   250 */  2010, 2010, 2010, 2199, 1807,   12,  150,  234,  234,  234,
 /*   260 */   484,  677,  677,  719,  180,  103,  108,  355,  378,  277,
 /*   270 */   465,  522,  397,  601,  332,  255,  635,  382,  426,  384,
 /*   280 */   602,  429,  859,  923,  740, 1034,  606,  669,  547,  547,
 /*   290 */   547,  547,  547,  819,  681,  787,  958,  973,  975,  985,
 /*   300 */   990,  967,  977,  995,  985,  988, 1002,  985,  985, 1035,
 /*   310 */  1018, 1038, 1040, 1039, 1041, 1022, 1023, 1024, 1032, 1050,
 /*   320 */  1051, 1029, 1030, 1033, 1044, 1025, 1043, 1085, 1057, 1086,
 /*   330 */  1088, 1059, 1089, 1092, 1135, 1103, 1097, 1119, 1197, 1146,
 /*   340 */  1198, 1152, 1153, 1190, 1191, 1137, 1139, 1158, 1208, 1196,
 /*   350 */  1166, 1167, 1055, 1169, 1183, 1212, 1170, 1171, 1173,  748,
 /*   360 */  1213, 1214, 1215, 1199, 1265, 1219, 1147, 1225, 1200, 1274,
 /*   370 */  1155, 1230, 1229, 1231, 1233, 1247, 1271, 1278, 1204, 1281,
 /*   380 */  1267, 1308, 1330, 1263, 1269, 1279, 1293, 1276, 1295, 1280,
 /*   390 */  1320, 1288, 1324, 1283, 1339, 1338, 1358, 1340, 1368, 1328,
 /*   400 */  1347, 1272, 1341, 1349, 1335, 1387, 1381, 1383, 1384, 1354,
 /*   410 */  1388, 1389, 1403, 1434,  985, 1435, 1401, 1437, 1308, 1405,
 /*   420 */  1440, 1460, 1408, 1394, 1411, 1417, 1395, 1419, 1413, 1456,
 /*   430 */  1418, 1422, 1415, 1424, 1463, 1504, 1469, 1466, 1472, 1473,
 /*   440 */  1475, 1511, 1476, 1480, 1477, 1335,
};
#define YY_REDUCE_USE_DFLT (-237)
#define YY_REDUCE_COUNT (196)
#define YY_REDUCE_MIN   (-236)
#define YY_REDUCE_MAX   (3389)
static const short yy_reduce_ofst[] = {
 /*     0 */    32, 1655, 1750, 1874, 1927, 2022, 2134, 2187, 2282, 2394,
 /*    10 */  2447, 2542, 2654, 2707, 2802, 2260,  -26,  -26,  345, 3010,
 /*    20 */   336, 1704,  -99,   60, -164,  117, 1987,  349, 1738, 1852,
 /*    30 */  1880, 1976, 2241,  124,  270, 2045,  405, 1925, 1923, 2042,
 /*    40 */  2140,  184, 1513,  114, 1652, 1773,  342, 1921, 2129,  413,
 /*    50 */   627,  778, 1017, 1081, 1145, 1273, 1337, 1516, 2166, 2229,
 /*    60 */  2321, 2381, 2384, 2426, 2489, 2520, 2568, 2571, 2574, 2581,
 /*    70 */  2584, 2589, 2605, 2609, 2641, 2677, 2687, 2760, 2780, 2825,
 /*    80 */  2833, 2838, 2841, 2862, 2865, 2869, 2921, 2970, 2975, 2978,
 /*    90 */  2999, 3032, 3038, 3041, 3048, 3066, 3070, 3078, 3126, 3157,
 /*   100 */  3204, 3236, 3265, 3270, 3276, 3282, 3285, 3288, 3304, 3308,
 /*   110 */  3317, 3320, 3324, 3331, 3334, 3337, 3366, 3386, 3389, 1663,
 /*   120 */   -77,  -18,   62,  246,  257,  600,  604,  783, -124,  177,
 /*   130 */  -236, -106,  121,   74, -196, -197,  -40, -117,  -34,  -40,
 /*   140 */    89,   65,  284,  173, -186,   61,  111,  182,  179,  202,
 /*   150 */   266,  361,  364,  367,  337,  402,  381,  409,  462,  414,
 /*   160 */   383,  485,  534,  539,  536,  565,  578,  581,  623,  628,
 /*   170 */   672,  598,  690,  687,  658,  659,  660,  634,  664,  713,
 /*   180 */   414,  759,  790,  792,  767,  414,  737,  797,  796,  771,
 /*   190 */   793,  838,  802,  803,  804,  825,  843,
};
static const YYACTIONTYPE yy_default[] = {
 /*     0 */  1138, 1231, 1538, 1231, 1538, 1538, 1538, 1538, 1538, 1538,
 /*    10 */  1538, 1538, 1538, 1538, 1538, 1481, 1333, 1333, 1333, 1538,
 /*    20 */  1333, 1333, 1538, 1538, 1437, 1437, 1437, 1538, 1538, 1538,
 /*    30 */  1538, 1538, 1538, 1538, 1538, 1481, 1437, 1333, 1538, 1538,
 /*    40 */  1538, 1538, 1538, 1538, 1538, 1538, 1147, 1538, 1538, 1147,
 /*    50 */  1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538,
 /*    60 */  1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538,
 /*    70 */  1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538,
 /*    80 */  1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538,
 /*    90 */  1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538,
 /*   100 */  1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538,
 /*   110 */  1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538,
 /*   120 */  1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1369,
 /*   130 */  1369, 1538, 1210, 1538, 1262, 1493, 1330, 1369, 1538, 1330,
 /*   140 */  1538, 1514, 1538, 1232, 1369, 1248, 1538, 1538, 1538, 1538,
 /*   150 */  1538, 1538, 1538, 1538, 1245, 1538, 1245, 1245, 1538, 1494,
 /*   160 */  1538, 1538, 1538, 1538, 1538, 1152, 1264, 1231, 1538, 1227,
 /*   170 */  1501, 1336, 1538, 1227, 1538, 1538, 1538, 1538, 1538, 1358,
 /*   180 */  1331, 1504, 1503, 1501, 1206, 1321, 1317, 1501, 1538, 1215,
 /*   190 */  1154, 1170, 1231, 1231, 1231, 1190, 1170, 1538, 1538, 1538,
 /*   200 */  1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538,
 /*   210 */  1538, 1538, 1538, 1538, 1538, 1538, 1246, 1368, 1366, 1495,
 /*   220 */  1490, 1435, 1525, 1469, 1491, 1212, 1439, 1432, 1431, 1430,
 /*   230 */  1429, 1428, 1427, 1426, 1425, 1424, 1423, 1422, 1421, 1438,
 /*   240 */  1505, 1434, 1420, 1419, 1415, 1414, 1413, 1412, 1411, 1410,
 /*   250 */  1409, 1408, 1407, 1406, 1418, 1417, 1416, 1405, 1404, 1403,
 /*   260 */  1151, 1395, 1396, 1538, 1523, 1538, 1538, 1538, 1538, 1538,
 /*   270 */  1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538,
 /*   280 */  1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1402, 1401,
 /*   290 */  1399, 1398, 1397, 1538, 1538, 1538, 1538, 1538, 1538, 1153,
 /*   300 */  1538, 1538, 1538, 1538, 1160, 1538, 1538, 1276, 1275, 1538,
 /*   310 */  1538, 1538, 1163, 1538, 1538, 1538, 1538, 1538, 1538, 1538,
 /*   320 */  1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538,
 /*   330 */  1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1367, 1538,
 /*   340 */  1365, 1520, 1480, 1538, 1538, 1538, 1538, 1492, 1538, 1538,
 /*   350 */  1538, 1538, 1484, 1538, 1538, 1538, 1538, 1538, 1538, 1538,
 /*   360 */  1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538,
 /*   370 */  1538, 1538, 1538, 1538, 1538, 1538, 1373, 1538, 1538, 1295,
 /*   380 */  1538, 1348, 1211, 1538, 1538, 1538, 1538, 1538, 1538, 1538,
 /*   390 */  1538, 1538, 1538, 1538, 1538, 1516, 1314, 1512, 1538, 1538,
 /*   400 */  1538, 1538, 1538, 1400, 1528, 1373, 1538, 1538, 1538, 1538,
 /*   410 */  1538, 1538, 1538, 1538, 1301, 1538, 1538, 1538, 1538, 1538,
 /*   420 */  1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538, 1538,
 /*   430 */  1538, 1538, 1538, 1538, 1538, 1538, 1187, 1538, 1538, 1186,
 /*   440 */  1538, 1538, 1538, 1538, 1538, 1148,
};
/********** End of lemon-generated parsing tables *****************************/

/* The next table maps tokens (terminal symbols) into fallback tokens.  
** If a construct like the following:
** 
**      %fallback ID X Y Z.
**
** appears in the grammar, then ID becomes a fallback token for X, Y,
** and Z.  Whenever one of the tokens X, Y, or Z is input to the parser
** but it does not parse, the type of the token is changed to ID and
** the parse is retried before an error is thrown.
**
** This feature can be used, for example, to cause some keywords in a language
** to revert to identifiers if they keyword does not apply in the context where
** it appears.
*/
#ifdef YYFALLBACK
static const YYCODETYPE yyFallback[] = {
    0,  /*          $ => nothing */
    0,  /*         EQ => nothing */
    0,  /*      MULEQ => nothing */
    0,  /*      DIVEQ => nothing */
    0,  /*      MODEQ => nothing */
    0,  /*      ADDEQ => nothing */
    0,  /*      SUBEQ => nothing */
    0,  /*      LSHEQ => nothing */
    0,  /*      RSHEQ => nothing */
    0,  /*      ANDEQ => nothing */
    0,  /*       OREQ => nothing */
    0,  /*      XOREQ => nothing */
    0,  /*     URSHEQ => nothing */
    0,  /*   QUESTION => nothing */
    0,  /*      COLON => nothing */
    0,  /*       OROR => nothing */
    0,  /*     ANDAND => nothing */
    0,  /*       EQEQ => nothing */
    0,  /*        NEQ => nothing */
    0,  /*   APPROXEQ => nothing */
    0,  /*         LT => nothing */
    0,  /*         GT => nothing */
    0,  /*       LTEQ => nothing */
    0,  /*       GTEQ => nothing */
    0,  /*     LTGTEQ => nothing */
    0,  /*         IS => nothing */
    0,  /*     DOTDOT => nothing */
    0,  /*         OR => nothing */
    0,  /*        XOR => nothing */
    0,  /*        AND => nothing */
    0,  /*        LSH => nothing */
    0,  /*        RSH => nothing */
    0,  /*       URSH => nothing */
    0,  /*        SUB => nothing */
    0,  /*        ADD => nothing */
    0,  /*        MUL => nothing */
    0,  /*        DIV => nothing */
    0,  /*        MOD => nothing */
    0,  /*  CROSSPROD => nothing */
    0,  /*    DOTPROD => nothing */
    0,  /*        POW => nothing */
    0,  /*      UNARY => nothing */
    0,  /*     ADDADD => nothing */
    0,  /*     SUBSUB => nothing */
    0,  /*        DOT => nothing */
    0,  /*     LPAREN => nothing */
    0,  /*   LBRACKET => nothing */
    0,  /*      SCOPE => nothing */
    0,  /*        EOF => nothing */
    0,  /*  SEMICOLON => nothing */
    0,  /*      COMMA => nothing */
    0,  /*    INCLUDE => nothing */
    0,  /*     EXTEND => nothing */
    0,  /*      CLASS => nothing */
    0,  /* IDENTIFIER => nothing */
    0,  /*   ABSTRACT => nothing */
    0,  /*      FINAL => nothing */
    0,  /*     NATIVE => nothing */
    0,  /*         UI => nothing */
    0,  /*       PLAY => nothing */
    0,  /*   REPLACES => nothing */
    0,  /*    VERSION => nothing */
    0,  /*   STRCONST => nothing */
    0,  /*     RPAREN => nothing */
    0,  /*     SEALED => nothing */
    0,  /*     ACTION => nothing */
   54,  /*    DEFAULT => IDENTIFIER */
   54,  /*      COLOR => IDENTIFIER */
    0,  /*     LBRACE => nothing */
    0,  /*     RBRACE => nothing */
    0,  /*      MIXIN => nothing */
   54,  /*   PROPERTY => IDENTIFIER */
    0,  /*    FLAGDEF => nothing */
    0,  /*   INTCONST => nothing */
    0,  /*   INTERNAL => nothing */
    0,  /*     STRUCT => nothing */
    0,  /* CLEARSCOPE => nothing */
    0,  /*      CONST => nothing */
    0,  /*       ENUM => nothing */
    0,  /*     STATES => nothing */
    0,  /*        NWS => nothing */
    0,  /*       STOP => nothing */
    0,  /*       WAIT => nothing */
    0,  /*       FAIL => nothing */
    0,  /*       LOOP => nothing */
    0,  /*       GOTO => nothing */
    0,  /*      SUPER => nothing */
    0,  /*     BRIGHT => nothing */
    0,  /*       FAST => nothing */
    0,  /*       SLOW => nothing */
    0,  /*    NODELAY => nothing */
    0,  /*   CANRAISE => nothing */
    0,  /*     OFFSET => nothing */
    0,  /*      LIGHT => nothing */
   54,  /*      SBYTE => IDENTIFIER */
   54,  /*       BYTE => IDENTIFIER */
   54,  /*      SHORT => IDENTIFIER */
   54,  /*     USHORT => IDENTIFIER */
   54,  /*        INT => IDENTIFIER */
   54,  /*       UINT => IDENTIFIER */
   54,  /*       BOOL => IDENTIFIER */
   54,  /*      FLOAT => IDENTIFIER */
   54,  /*     DOUBLE => IDENTIFIER */
   54,  /*    VECTOR2 => IDENTIFIER */
   54,  /*    VECTOR3 => IDENTIFIER */
   54,  /*    VECTOR4 => IDENTIFIER */
   54,  /*       NAME => IDENTIFIER */
   54,  /*      SOUND => IDENTIFIER */
   54,  /*      STATE => IDENTIFIER */
    0,  /*        LET => nothing */
    0,  /*     ATSIGN => nothing */
    0,  /*   READONLY => nothing */
   54,  /*     STRING => IDENTIFIER */
   54,  /*        MAP => IDENTIFIER */
   54,  /* MAPITERATOR => IDENTIFIER */
   54,  /*      ARRAY => IDENTIFIER */
   54,  /*       VOID => IDENTIFIER */
   54,  /*      UINT8 => IDENTIFIER */
   54,  /*       INT8 => IDENTIFIER */
   54,  /*     UINT16 => IDENTIFIER */
   54,  /*      INT16 => IDENTIFIER */
};
#endif /* YYFALLBACK */

/* The following structure represents a single element of the
** parser's stack.  Information stored includes:
**
**   +  The state number for the parser at this level of the stack.
**
**   +  The value of the token stored at this level of the stack.
**      (In other words, the "major" token.)
**
**   +  The semantic value stored at this level of the stack.  This is
**      the information used by the action routines in the grammar.
**      It is sometimes called the "minor" token.
**
** After the "shift" half of a SHIFTREDUCE action, the stateno field
** actually contains the reduce action for the second half of the
** SHIFTREDUCE.
*/
struct yyStackEntry {
  YYACTIONTYPE stateno;  /* The state-number, or reduce action in SHIFTREDUCE */
  YYCODETYPE major;      /* The major token value.  This is the code
                         ** number for the token at this stack level */
  YYMINORTYPE minor;     /* The user-supplied minor token value.  This
                         ** is the value of the token  */
};
typedef struct yyStackEntry yyStackEntry;

/* The state of the parser is completely contained in an instance of
** the following structure */
struct yyParser {
  yyStackEntry *yytos;          /* Pointer to top element of the stack */
#ifdef YYTRACKMAXSTACKDEPTH
  int yyhwm;                    /* High-water mark of the stack */
#endif
#ifndef YYNOERRORRECOVERY
  int yyerrcnt;                 /* Shifts left before out of the error */
#endif
  ZCCParseARG_SDECL                /* A place to hold %extra_argument */
#if YYSTACKDEPTH<=0
  int yystksz;                  /* Current side of the stack */
  yyStackEntry *yystack;        /* The parser's stack */
  yyStackEntry yystk0;          /* First stack entry */
#else
  yyStackEntry yystack[YYSTACKDEPTH];  /* The parser's stack */
#endif
};
typedef struct yyParser yyParser;

#ifndef NDEBUG
#include <stdio.h>
static FILE *yyTraceFILE = 0;
static char *yyTracePrompt = 0;
#endif /* NDEBUG */

#ifndef NDEBUG
/* 
** Turn parser tracing on by giving a stream to which to write the trace
** and a prompt to preface each trace message.  Tracing is turned off
** by making either argument NULL 
**
** Inputs:
** <ul>
** <li> A FILE* to which trace output should be written.
**      If NULL, then tracing is turned off.
** <li> A prefix string written at the beginning of every
**      line of trace output.  If NULL, then tracing is
**      turned off.
** </ul>
**
** Outputs:
** None.
*/
void ZCCParseTrace(FILE *TraceFILE, char *zTracePrompt){
  yyTraceFILE = TraceFILE;
  yyTracePrompt = zTracePrompt;
  if( yyTraceFILE==0 ) yyTracePrompt = 0;
  else if( yyTracePrompt==0 ) yyTraceFILE = 0;
}
#endif /* NDEBUG */

#ifndef NDEBUG
/* For tracing shifts, the names of all terminals and nonterminals
** are required.  The following table supplies these names */
static const char *const yyTokenName[] = { 
  "$",             "EQ",            "MULEQ",         "DIVEQ",       
  "MODEQ",         "ADDEQ",         "SUBEQ",         "LSHEQ",       
  "RSHEQ",         "ANDEQ",         "OREQ",          "XOREQ",       
  "URSHEQ",        "QUESTION",      "COLON",         "OROR",        
  "ANDAND",        "EQEQ",          "NEQ",           "APPROXEQ",    
  "LT",            "GT",            "LTEQ",          "GTEQ",        
  "LTGTEQ",        "IS",            "DOTDOT",        "OR",          
  "XOR",           "AND",           "LSH",           "RSH",         
  "URSH",          "SUB",           "ADD",           "MUL",         
  "DIV",           "MOD",           "CROSSPROD",     "DOTPROD",     
  "POW",           "UNARY",         "ADDADD",        "SUBSUB",      
  "DOT",           "LPAREN",        "LBRACKET",      "SCOPE",       
  "EOF",           "SEMICOLON",     "COMMA",         "INCLUDE",     
  "EXTEND",        "CLASS",         "IDENTIFIER",    "ABSTRACT",    
  "FINAL",         "NATIVE",        "UI",            "PLAY",        
  "REPLACES",      "VERSION",       "STRCONST",      "RPAREN",      
  "SEALED",        "ACTION",        "DEFAULT",       "COLOR",       
  "LBRACE",        "RBRACE",        "MIXIN",         "PROPERTY",    
  "FLAGDEF",       "INTCONST",      "INTERNAL",      "STRUCT",      
  "CLEARSCOPE",    "CONST",         "ENUM",          "STATES",      
  "NWS",           "STOP",          "WAIT",          "FAIL",        
  "LOOP",          "GOTO",          "SUPER",         "BRIGHT",      
  "FAST",          "SLOW",          "NODELAY",       "CANRAISE",    
  "OFFSET",        "LIGHT",         "SBYTE",         "BYTE",        
  "SHORT",         "USHORT",        "INT",           "UINT",        
  "BOOL",          "FLOAT",         "DOUBLE",        "VECTOR2",     
  "VECTOR3",       "VECTOR4",       "NAME",          "SOUND",       
  "STATE",         "LET",           "ATSIGN",        "READONLY",    
  "STRING",        "MAP",           "MAPITERATOR",   "ARRAY",       
  "VOID",          "UINT8",         "INT8",          "UINT16",      
  "INT16",         "FNTYPE",        "RBRACKET",      "DEPRECATED",  
  "STATIC",        "PRIVATE",       "PROTECTED",     "LATENT",      
  "META",          "TRANSIENT",     "VIRTUAL",       "OVERRIDE",    
  "VARARG",        "VIRTUALSCOPE",  "ELLIPSIS",      "IN",          
  "OUT",           "OPTIONAL",      "TILDE",         "BANG",        
  "SIZEOF",        "ALIGNOF",       "UINTCONST",     "FLOATCONST",  
  "NAMECONST",     "FALSE",         "TRUE",          "NULLPTR",     
  "STATICCONST",   "CONTINUE",      "BREAK",         "RETURN",      
  "DO",            "FOR",           "FOREACH",       "WHILE",       
  "UNTIL",         "IF",            "ELSE",          "SWITCH",      
  "CASE",          "error",         "declarator",    "declarator_no_fun",
  "opt_func_body",  "function_body",  "main",          "translation_unit",
  "external_declaration",  "mixin_definition",  "class_definition",  "struct_def",  
  "enum_def",      "const_def",     "include_def",   "opt_semicolon",
  "opt_comma",     "opt_expr",      "expr",          "string_constant",
  "class_head",    "class_innards",  "class_member",  "class_body",  
  "class_ancestry",  "class_flags",   "dottable_id",   "states_opt",  
  "mixin_statement",  "property_def",  "flag_def",      "states_def",  
  "default_def",   "staticarray_statement",  "opt_struct_body",  "struct_body", 
  "struct_member",  "identifier_list",  "struct_flags",  "enum_list",   
  "opt_enum_list",  "enumerator",    "enum_type",     "int_type",    
  "mixin_class_definition",  "mixin_class_body",  "mixin_class_member",  "states_body", 
  "state_line",    "state_label",   "state_flow",    "state_flow_type",
  "state_goto_offset",  "state_action",  "state_call",    "state_call_params",
  "state_opts",    "states_opts",   "scanner_mode",  "light_list",  
  "statement_list",  "func_expr_list",  "default_statement_list",  "default_statement",
  "property_statement",  "flag_statement",  "expr_list",     "type_name",   
  "type_name1",    "aggregate_type",  "type",          "type_list",   
  "type_list_or_void",  "type_or_array",  "class_restrictor",  "array_size",  
  "array_size_expr",  "func_ptr_type",  "func_ptr_params",  "func_ptr_param_list",
  "func_ptr_param",  "fn_ptr_flag",   "func_param_flags",  "variables_or_function",
  "decl_flags",    "func_params",   "func_const",    "variable_list",
  "variable_name",  "decl_flag",     "opt_deprecation_message",  "func_param_list",
  "func_param",    "primary",       "unary_expr",    "constant",    
  "func_expr_item",  "named_expr",    "compound_statement",  "statement",   
  "labeled_statement",  "expression_statement",  "selection_statement",  "iteration_statement",
  "array_iteration_statement",  "two_arg_iteration_statement",  "three_arg_iteration_statement",  "typed_iteration_statement",
  "jump_statement",  "assign_statement",  "assign_decl_statement",  "local_var",   
  "while_or_until",  "for_init",      "for_bump",      "if_front",    
  "variable_list_with_init",  "var_init",    
};
#endif /* NDEBUG */

#ifndef NDEBUG
/* For tracing reduce actions, the names of all rules are required.
*/
static const char *const yyRuleName[] = {
 /*   0 */ "main ::= translation_unit",
 /*   1 */ "translation_unit ::=",
 /*   2 */ "translation_unit ::= translation_unit external_declaration",
 /*   3 */ "translation_unit ::= error",
 /*   4 */ "external_declaration ::= mixin_definition",
 /*   5 */ "external_declaration ::= class_definition",
 /*   6 */ "external_declaration ::= struct_def",
 /*   7 */ "external_declaration ::= enum_def",
 /*   8 */ "external_declaration ::= const_def",
 /*   9 */ "external_declaration ::= include_def",
 /*  10 */ "opt_expr ::=",
 /*  11 */ "include_def ::= INCLUDE string_constant",
 /*  12 */ "class_definition ::= class_head class_body",
 /*  13 */ "class_head ::= EXTEND CLASS IDENTIFIER",
 /*  14 */ "class_head ::= CLASS IDENTIFIER class_ancestry class_flags",
 /*  15 */ "class_ancestry ::=",
 /*  16 */ "class_ancestry ::= COLON dottable_id",
 /*  17 */ "class_flags ::=",
 /*  18 */ "class_flags ::= class_flags ABSTRACT",
 /*  19 */ "class_flags ::= class_flags FINAL",
 /*  20 */ "class_flags ::= class_flags NATIVE",
 /*  21 */ "class_flags ::= class_flags UI",
 /*  22 */ "class_flags ::= class_flags PLAY",
 /*  23 */ "class_flags ::= class_flags REPLACES dottable_id",
 /*  24 */ "class_flags ::= class_flags VERSION LPAREN STRCONST RPAREN",
 /*  25 */ "class_flags ::= class_flags SEALED LPAREN states_opt RPAREN",
 /*  26 */ "dottable_id ::= IDENTIFIER",
 /*  27 */ "dottable_id ::= ACTION",
 /*  28 */ "dottable_id ::= dottable_id DOT IDENTIFIER",
 /*  29 */ "dottable_id ::= dottable_id DOT DEFAULT",
 /*  30 */ "dottable_id ::= dottable_id DOT COLOR",
 /*  31 */ "class_body ::= SEMICOLON class_innards EOF",
 /*  32 */ "class_body ::= LBRACE class_innards RBRACE",
 /*  33 */ "class_innards ::=",
 /*  34 */ "class_innards ::= class_innards class_member",
 /*  35 */ "class_member ::= declarator",
 /*  36 */ "class_member ::= mixin_statement",
 /*  37 */ "class_member ::= enum_def",
 /*  38 */ "class_member ::= struct_def",
 /*  39 */ "class_member ::= states_def",
 /*  40 */ "class_member ::= default_def",
 /*  41 */ "class_member ::= const_def",
 /*  42 */ "class_member ::= property_def",
 /*  43 */ "class_member ::= flag_def",
 /*  44 */ "class_member ::= staticarray_statement",
 /*  45 */ "mixin_statement ::= MIXIN IDENTIFIER SEMICOLON",
 /*  46 */ "property_def ::= PROPERTY IDENTIFIER COLON identifier_list SEMICOLON",
 /*  47 */ "flag_def ::= FLAGDEF IDENTIFIER COLON IDENTIFIER COMMA INTCONST SEMICOLON",
 /*  48 */ "flag_def ::= FLAGDEF INTERNAL IDENTIFIER COLON IDENTIFIER COMMA INTCONST SEMICOLON",
 /*  49 */ "identifier_list ::= IDENTIFIER",
 /*  50 */ "identifier_list ::= states_opt COMMA IDENTIFIER",
 /*  51 */ "struct_def ::= STRUCT IDENTIFIER struct_flags LBRACE opt_struct_body RBRACE opt_semicolon",
 /*  52 */ "struct_def ::= EXTEND STRUCT IDENTIFIER LBRACE opt_struct_body RBRACE opt_semicolon",
 /*  53 */ "struct_flags ::=",
 /*  54 */ "struct_flags ::= struct_flags UI",
 /*  55 */ "struct_flags ::= struct_flags PLAY",
 /*  56 */ "struct_flags ::= struct_flags CLEARSCOPE",
 /*  57 */ "struct_flags ::= struct_flags NATIVE",
 /*  58 */ "struct_flags ::= struct_flags VERSION LPAREN STRCONST RPAREN",
 /*  59 */ "opt_struct_body ::=",
 /*  60 */ "opt_struct_body ::= error",
 /*  61 */ "struct_body ::= struct_body struct_member",
 /*  62 */ "struct_member ::= declarator",
 /*  63 */ "struct_member ::= enum_def",
 /*  64 */ "struct_member ::= const_def",
 /*  65 */ "struct_member ::= staticarray_statement",
 /*  66 */ "struct_member ::= flag_def",
 /*  67 */ "const_def ::= CONST IDENTIFIER EQ expr SEMICOLON",
 /*  68 */ "enum_def ::= ENUM IDENTIFIER enum_type LBRACE opt_enum_list RBRACE opt_semicolon",
 /*  69 */ "enum_type ::=",
 /*  70 */ "enum_type ::= COLON int_type",
 /*  71 */ "enum_list ::= error",
 /*  72 */ "enum_list ::= enum_list COMMA enumerator",
 /*  73 */ "opt_enum_list ::=",
 /*  74 */ "enumerator ::= IDENTIFIER",
 /*  75 */ "enumerator ::= IDENTIFIER EQ expr",
 /*  76 */ "mixin_definition ::= mixin_class_definition",
 /*  77 */ "mixin_class_definition ::= MIXIN CLASS IDENTIFIER LBRACE mixin_class_body RBRACE",
 /*  78 */ "mixin_class_body ::=",
 /*  79 */ "mixin_class_body ::= mixin_class_body mixin_class_member",
 /*  80 */ "mixin_class_member ::= declarator",
 /*  81 */ "mixin_class_member ::= enum_def",
 /*  82 */ "mixin_class_member ::= struct_def",
 /*  83 */ "mixin_class_member ::= states_def",
 /*  84 */ "mixin_class_member ::= default_def",
 /*  85 */ "mixin_class_member ::= const_def",
 /*  86 */ "mixin_class_member ::= property_def",
 /*  87 */ "mixin_class_member ::= flag_def",
 /*  88 */ "mixin_class_member ::= staticarray_statement",
 /*  89 */ "states_def ::= STATES states_opts scanner_mode LBRACE states_body RBRACE",
 /*  90 */ "states_opts ::=",
 /*  91 */ "states_opts ::= LPAREN states_opt RPAREN",
 /*  92 */ "states_opt ::= IDENTIFIER",
 /*  93 */ "states_opt ::= states_opt COMMA IDENTIFIER",
 /*  94 */ "scanner_mode ::=",
 /*  95 */ "states_body ::=",
 /*  96 */ "states_body ::= error",
 /*  97 */ "states_body ::= states_body state_line",
 /*  98 */ "states_body ::= states_body state_label",
 /*  99 */ "states_body ::= states_body state_flow",
 /* 100 */ "state_label ::= NWS COLON",
 /* 101 */ "state_flow_type ::= STOP",
 /* 102 */ "state_flow_type ::= WAIT",
 /* 103 */ "state_flow_type ::= FAIL",
 /* 104 */ "state_flow_type ::= LOOP",
 /* 105 */ "state_flow_type ::= GOTO dottable_id state_goto_offset",
 /* 106 */ "state_flow_type ::= GOTO IDENTIFIER SCOPE dottable_id state_goto_offset",
 /* 107 */ "state_flow_type ::= GOTO SUPER SCOPE dottable_id state_goto_offset",
 /* 108 */ "state_goto_offset ::=",
 /* 109 */ "state_goto_offset ::= ADD expr",
 /* 110 */ "state_line ::= NWS NWS expr state_opts state_action",
 /* 111 */ "state_opts ::=",
 /* 112 */ "state_opts ::= state_opts BRIGHT",
 /* 113 */ "state_opts ::= state_opts FAST",
 /* 114 */ "state_opts ::= state_opts SLOW",
 /* 115 */ "state_opts ::= state_opts NODELAY",
 /* 116 */ "state_opts ::= state_opts CANRAISE",
 /* 117 */ "state_opts ::= state_opts OFFSET LPAREN expr COMMA expr RPAREN",
 /* 118 */ "state_opts ::= state_opts LIGHT LPAREN light_list RPAREN",
 /* 119 */ "light_list ::= STRCONST",
 /* 120 */ "light_list ::= light_list COMMA STRCONST",
 /* 121 */ "state_action ::= LBRACE statement_list scanner_mode RBRACE",
 /* 122 */ "state_action ::= LBRACE scanner_mode RBRACE",
 /* 123 */ "state_action ::= LBRACE error scanner_mode RBRACE",
 /* 124 */ "state_action ::= state_call scanner_mode SEMICOLON",
 /* 125 */ "state_call ::=",
 /* 126 */ "state_call ::= IDENTIFIER state_call_params",
 /* 127 */ "state_call_params ::=",
 /* 128 */ "state_call_params ::= LPAREN func_expr_list RPAREN",
 /* 129 */ "default_def ::= DEFAULT LBRACE RBRACE",
 /* 130 */ "default_def ::= DEFAULT LBRACE default_statement_list RBRACE",
 /* 131 */ "default_def ::= DEFAULT LBRACE error RBRACE",
 /* 132 */ "default_statement_list ::= default_statement",
 /* 133 */ "default_statement_list ::= default_statement_list default_statement",
 /* 134 */ "default_statement ::= SEMICOLON",
 /* 135 */ "default_statement ::= error SEMICOLON",
 /* 136 */ "default_statement ::= property_statement",
 /* 137 */ "default_statement ::= flag_statement",
 /* 138 */ "flag_statement ::= ADD dottable_id",
 /* 139 */ "flag_statement ::= SUB dottable_id",
 /* 140 */ "property_statement ::= dottable_id expr_list SEMICOLON",
 /* 141 */ "property_statement ::= dottable_id SEMICOLON",
 /* 142 */ "int_type ::= SBYTE",
 /* 143 */ "int_type ::= BYTE",
 /* 144 */ "int_type ::= SHORT",
 /* 145 */ "int_type ::= USHORT",
 /* 146 */ "int_type ::= INT",
 /* 147 */ "int_type ::= UINT",
 /* 148 */ "type_name1 ::= BOOL",
 /* 149 */ "type_name1 ::= FLOAT",
 /* 150 */ "type_name1 ::= DOUBLE",
 /* 151 */ "type_name1 ::= VECTOR2",
 /* 152 */ "type_name1 ::= VECTOR3",
 /* 153 */ "type_name1 ::= VECTOR4",
 /* 154 */ "type_name1 ::= NAME",
 /* 155 */ "type_name1 ::= SOUND",
 /* 156 */ "type_name1 ::= STATE",
 /* 157 */ "type_name1 ::= COLOR",
 /* 158 */ "type_name1 ::= LET",
 /* 159 */ "type_name ::= type_name1",
 /* 160 */ "type_name ::= IDENTIFIER",
 /* 161 */ "type_name ::= ATSIGN IDENTIFIER",
 /* 162 */ "type_name ::= READONLY LT IDENTIFIER GT",
 /* 163 */ "type_name ::= READONLY LT ATSIGN IDENTIFIER GT",
 /* 164 */ "type_name ::= DOT dottable_id",
 /* 165 */ "aggregate_type ::= MAP LT type_or_array COMMA type_or_array GT",
 /* 166 */ "aggregate_type ::= MAPITERATOR LT type_or_array COMMA type_or_array GT",
 /* 167 */ "aggregate_type ::= ARRAY LT type_or_array GT",
 /* 168 */ "aggregate_type ::= func_ptr_type",
 /* 169 */ "fn_ptr_flag ::= UI",
 /* 170 */ "fn_ptr_flag ::= PLAY",
 /* 171 */ "fn_ptr_flag ::= CLEARSCOPE",
 /* 172 */ "func_ptr_type ::= FNTYPE LT fn_ptr_flag type_list_or_void LPAREN func_ptr_params RPAREN GT",
 /* 173 */ "func_ptr_type ::= FNTYPE LT VOID GT",
 /* 174 */ "func_ptr_params ::=",
 /* 175 */ "func_ptr_params ::= VOID",
 /* 176 */ "func_ptr_param_list ::= func_ptr_param_list COMMA func_ptr_param",
 /* 177 */ "func_ptr_param ::= func_param_flags type",
 /* 178 */ "func_ptr_param ::= func_param_flags type AND",
 /* 179 */ "aggregate_type ::= CLASS class_restrictor",
 /* 180 */ "class_restrictor ::=",
 /* 181 */ "class_restrictor ::= LT dottable_id GT",
 /* 182 */ "type ::= type_name",
 /* 183 */ "type ::= aggregate_type",
 /* 184 */ "type_or_array ::= type array_size",
 /* 185 */ "type_list ::= type_list COMMA type_or_array",
 /* 186 */ "type_list_or_void ::= VOID",
 /* 187 */ "array_size_expr ::= LBRACKET opt_expr RBRACKET",
 /* 188 */ "array_size ::= array_size array_size_expr",
 /* 189 */ "declarator ::= decl_flags type_list_or_void variables_or_function",
 /* 190 */ "variables_or_function ::= IDENTIFIER LPAREN func_params RPAREN func_const opt_func_body",
 /* 191 */ "variables_or_function ::= variable_list SEMICOLON",
 /* 192 */ "variables_or_function ::= error SEMICOLON",
 /* 193 */ "variable_name ::= IDENTIFIER",
 /* 194 */ "variable_name ::= IDENTIFIER array_size",
 /* 195 */ "variable_list ::= variable_list COMMA variable_name",
 /* 196 */ "decl_flags ::=",
 /* 197 */ "decl_flags ::= decl_flags decl_flag",
 /* 198 */ "decl_flags ::= decl_flags ACTION states_opts",
 /* 199 */ "opt_deprecation_message ::=",
 /* 200 */ "opt_deprecation_message ::= COMMA STRCONST",
 /* 201 */ "decl_flags ::= decl_flags DEPRECATED LPAREN STRCONST opt_deprecation_message RPAREN",
 /* 202 */ "decl_flags ::= decl_flags VERSION LPAREN STRCONST RPAREN",
 /* 203 */ "decl_flag ::= NATIVE",
 /* 204 */ "decl_flag ::= STATIC",
 /* 205 */ "decl_flag ::= PRIVATE",
 /* 206 */ "decl_flag ::= PROTECTED",
 /* 207 */ "decl_flag ::= LATENT",
 /* 208 */ "decl_flag ::= FINAL",
 /* 209 */ "decl_flag ::= META",
 /* 210 */ "decl_flag ::= TRANSIENT",
 /* 211 */ "decl_flag ::= READONLY",
 /* 212 */ "decl_flag ::= INTERNAL",
 /* 213 */ "decl_flag ::= VIRTUAL",
 /* 214 */ "decl_flag ::= OVERRIDE",
 /* 215 */ "decl_flag ::= ABSTRACT",
 /* 216 */ "decl_flag ::= VARARG",
 /* 217 */ "decl_flag ::= UI",
 /* 218 */ "decl_flag ::= PLAY",
 /* 219 */ "decl_flag ::= CLEARSCOPE",
 /* 220 */ "decl_flag ::= VIRTUALSCOPE",
 /* 221 */ "func_const ::=",
 /* 222 */ "func_const ::= CONST",
 /* 223 */ "opt_func_body ::= SEMICOLON",
 /* 224 */ "func_params ::=",
 /* 225 */ "func_params ::= VOID",
 /* 226 */ "func_params ::= func_param_list COMMA ELLIPSIS",
 /* 227 */ "func_param_list ::= func_param_list COMMA func_param",
 /* 228 */ "func_param ::= func_param_flags type IDENTIFIER",
 /* 229 */ "func_param ::= func_param_flags type IDENTIFIER EQ expr",
 /* 230 */ "func_param ::= func_param_flags type AND IDENTIFIER",
 /* 231 */ "func_param ::= func_param_flags type AND IDENTIFIER EQ expr",
 /* 232 */ "func_param_flags ::=",
 /* 233 */ "func_param_flags ::= func_param_flags IN",
 /* 234 */ "func_param_flags ::= func_param_flags OUT",
 /* 235 */ "func_param_flags ::= func_param_flags OPTIONAL",
 /* 236 */ "primary ::= IDENTIFIER",
 /* 237 */ "primary ::= SUPER",
 /* 238 */ "primary ::= constant",
 /* 239 */ "primary ::= LPAREN expr COMMA expr COMMA expr COMMA expr RPAREN",
 /* 240 */ "primary ::= LPAREN expr COMMA expr COMMA expr RPAREN",
 /* 241 */ "primary ::= LPAREN expr COMMA expr RPAREN",
 /* 242 */ "primary ::= LPAREN expr RPAREN",
 /* 243 */ "primary ::= primary LPAREN func_expr_list RPAREN",
 /* 244 */ "primary ::= LPAREN CLASS LT IDENTIFIER GT RPAREN LPAREN func_expr_list RPAREN",
 /* 245 */ "primary ::= LPAREN func_ptr_type RPAREN LPAREN expr RPAREN",
 /* 246 */ "primary ::= primary LBRACKET expr RBRACKET",
 /* 247 */ "primary ::= primary DOT IDENTIFIER",
 /* 248 */ "primary ::= primary ADDADD",
 /* 249 */ "primary ::= primary SUBSUB",
 /* 250 */ "unary_expr ::= SUB unary_expr",
 /* 251 */ "unary_expr ::= ADD unary_expr",
 /* 252 */ "unary_expr ::= SUBSUB unary_expr",
 /* 253 */ "unary_expr ::= ADDADD unary_expr",
 /* 254 */ "unary_expr ::= TILDE unary_expr",
 /* 255 */ "unary_expr ::= BANG unary_expr",
 /* 256 */ "unary_expr ::= SIZEOF unary_expr",
 /* 257 */ "unary_expr ::= ALIGNOF unary_expr",
 /* 258 */ "expr ::= expr ADD expr",
 /* 259 */ "expr ::= expr SUB expr",
 /* 260 */ "expr ::= expr MUL expr",
 /* 261 */ "expr ::= expr DIV expr",
 /* 262 */ "expr ::= expr MOD expr",
 /* 263 */ "expr ::= expr POW expr",
 /* 264 */ "expr ::= expr CROSSPROD expr",
 /* 265 */ "expr ::= expr DOTPROD expr",
 /* 266 */ "expr ::= expr LSH expr",
 /* 267 */ "expr ::= expr RSH expr",
 /* 268 */ "expr ::= expr URSH expr",
 /* 269 */ "expr ::= expr DOTDOT expr",
 /* 270 */ "expr ::= expr LT expr",
 /* 271 */ "expr ::= expr GT expr",
 /* 272 */ "expr ::= expr LTEQ expr",
 /* 273 */ "expr ::= expr GTEQ expr",
 /* 274 */ "expr ::= expr LTGTEQ expr",
 /* 275 */ "expr ::= expr IS expr",
 /* 276 */ "expr ::= expr EQEQ expr",
 /* 277 */ "expr ::= expr NEQ expr",
 /* 278 */ "expr ::= expr APPROXEQ expr",
 /* 279 */ "expr ::= expr AND expr",
 /* 280 */ "expr ::= expr XOR expr",
 /* 281 */ "expr ::= expr OR expr",
 /* 282 */ "expr ::= expr ANDAND expr",
 /* 283 */ "expr ::= expr OROR expr",
 /* 284 */ "expr ::= expr EQ expr",
 /* 285 */ "expr ::= expr ADDEQ expr",
 /* 286 */ "expr ::= expr SUBEQ expr",
 /* 287 */ "expr ::= expr MULEQ expr",
 /* 288 */ "expr ::= expr DIVEQ expr",
 /* 289 */ "expr ::= expr MODEQ expr",
 /* 290 */ "expr ::= expr LSHEQ expr",
 /* 291 */ "expr ::= expr RSHEQ expr",
 /* 292 */ "expr ::= expr URSHEQ expr",
 /* 293 */ "expr ::= expr ANDEQ expr",
 /* 294 */ "expr ::= expr OREQ expr",
 /* 295 */ "expr ::= expr XOREQ expr",
 /* 296 */ "expr ::= expr SCOPE expr",
 /* 297 */ "expr ::= expr QUESTION expr COLON expr",
 /* 298 */ "expr_list ::= expr_list COMMA expr",
 /* 299 */ "func_expr_list ::= func_expr_list COMMA func_expr_item",
 /* 300 */ "func_expr_item ::=",
 /* 301 */ "named_expr ::= IDENTIFIER COLON expr",
 /* 302 */ "named_expr ::= expr",
 /* 303 */ "string_constant ::= STRCONST",
 /* 304 */ "string_constant ::= string_constant STRCONST",
 /* 305 */ "constant ::= INTCONST",
 /* 306 */ "constant ::= UINTCONST",
 /* 307 */ "constant ::= FLOATCONST",
 /* 308 */ "constant ::= NAMECONST",
 /* 309 */ "constant ::= FALSE",
 /* 310 */ "constant ::= TRUE",
 /* 311 */ "constant ::= NULLPTR",
 /* 312 */ "statement ::= SEMICOLON",
 /* 313 */ "statement ::= labeled_statement",
 /* 314 */ "statement ::= compound_statement",
 /* 315 */ "statement ::= expression_statement SEMICOLON",
 /* 316 */ "statement ::= assign_statement SEMICOLON",
 /* 317 */ "statement ::= assign_decl_statement SEMICOLON",
 /* 318 */ "statement ::= local_var SEMICOLON",
 /* 319 */ "statement ::= error SEMICOLON",
 /* 320 */ "statement ::= staticarray_statement",
 /* 321 */ "staticarray_statement ::= STATICCONST type IDENTIFIER LBRACKET RBRACKET EQ LBRACE expr_list RBRACE SEMICOLON",
 /* 322 */ "staticarray_statement ::= STATICCONST type LBRACKET RBRACKET IDENTIFIER EQ LBRACE expr_list RBRACE SEMICOLON",
 /* 323 */ "jump_statement ::= CONTINUE SEMICOLON",
 /* 324 */ "jump_statement ::= BREAK SEMICOLON",
 /* 325 */ "jump_statement ::= RETURN SEMICOLON",
 /* 326 */ "jump_statement ::= RETURN expr_list SEMICOLON",
 /* 327 */ "compound_statement ::= LBRACE RBRACE",
 /* 328 */ "compound_statement ::= LBRACE statement_list RBRACE",
 /* 329 */ "compound_statement ::= LBRACE error RBRACE",
 /* 330 */ "statement_list ::= statement",
 /* 331 */ "statement_list ::= statement_list statement",
 /* 332 */ "expression_statement ::= expr",
 /* 333 */ "iteration_statement ::= while_or_until LPAREN expr RPAREN statement",
 /* 334 */ "iteration_statement ::= DO statement while_or_until LPAREN expr RPAREN",
 /* 335 */ "iteration_statement ::= FOR LPAREN for_init SEMICOLON opt_expr SEMICOLON for_bump RPAREN statement",
 /* 336 */ "array_iteration_statement ::= FOREACH LPAREN variable_name COLON expr RPAREN statement",
 /* 337 */ "two_arg_iteration_statement ::= FOREACH LPAREN variable_name COMMA variable_name COLON expr RPAREN statement",
 /* 338 */ "three_arg_iteration_statement ::= FOREACH LPAREN variable_name COMMA variable_name COMMA variable_name COLON expr RPAREN statement",
 /* 339 */ "typed_iteration_statement ::= FOREACH LPAREN variable_name variable_name COLON expr RPAREN statement",
 /* 340 */ "while_or_until ::= WHILE",
 /* 341 */ "while_or_until ::= UNTIL",
 /* 342 */ "for_init ::= local_var",
 /* 343 */ "for_init ::= for_bump",
 /* 344 */ "for_bump ::=",
 /* 345 */ "for_bump ::= expression_statement",
 /* 346 */ "for_bump ::= for_bump COMMA expression_statement",
 /* 347 */ "selection_statement ::= if_front",
 /* 348 */ "selection_statement ::= if_front ELSE statement",
 /* 349 */ "if_front ::= IF LPAREN expr RPAREN statement",
 /* 350 */ "selection_statement ::= SWITCH LPAREN expr RPAREN statement",
 /* 351 */ "labeled_statement ::= CASE expr COLON",
 /* 352 */ "labeled_statement ::= DEFAULT COLON",
 /* 353 */ "assign_statement ::= LBRACKET expr_list RBRACKET EQ expr",
 /* 354 */ "assign_decl_statement ::= LET LBRACKET identifier_list RBRACKET EQ expr",
 /* 355 */ "local_var ::= type variable_list_with_init",
 /* 356 */ "var_init ::= IDENTIFIER",
 /* 357 */ "var_init ::= IDENTIFIER array_size",
 /* 358 */ "var_init ::= IDENTIFIER EQ expr",
 /* 359 */ "var_init ::= IDENTIFIER EQ LBRACE expr_list RBRACE",
 /* 360 */ "var_init ::= IDENTIFIER array_size EQ LBRACE expr_list RBRACE",
 /* 361 */ "var_init ::= IDENTIFIER EQ LBRACE error RBRACE",
 /* 362 */ "variable_list_with_init ::= variable_list_with_init COMMA var_init",
 /* 363 */ "translation_unit ::= translation_unit EOF",
 /* 364 */ "opt_semicolon ::=",
 /* 365 */ "opt_semicolon ::= SEMICOLON",
 /* 366 */ "opt_comma ::=",
 /* 367 */ "opt_comma ::= COMMA",
 /* 368 */ "opt_expr ::= expr",
 /* 369 */ "opt_struct_body ::= struct_body",
 /* 370 */ "struct_body ::= struct_member",
 /* 371 */ "enum_list ::= enumerator",
 /* 372 */ "opt_enum_list ::= enum_list opt_comma",
 /* 373 */ "state_flow ::= state_flow_type scanner_mode SEMICOLON",
 /* 374 */ "type_name1 ::= int_type",
 /* 375 */ "func_ptr_params ::= func_ptr_param_list",
 /* 376 */ "func_ptr_param_list ::= func_ptr_param",
 /* 377 */ "type_or_array ::= type",
 /* 378 */ "type_list ::= type_or_array",
 /* 379 */ "type_list_or_void ::= type_list",
 /* 380 */ "array_size ::= array_size_expr",
 /* 381 */ "variable_list ::= variable_name",
 /* 382 */ "opt_func_body ::= function_body",
 /* 383 */ "func_params ::= func_param_list",
 /* 384 */ "func_param_list ::= func_param",
 /* 385 */ "primary ::= LPAREN error RPAREN",
 /* 386 */ "unary_expr ::= primary",
 /* 387 */ "expr ::= unary_expr",
 /* 388 */ "expr_list ::= expr",
 /* 389 */ "func_expr_list ::= func_expr_item",
 /* 390 */ "func_expr_item ::= named_expr",
 /* 391 */ "constant ::= string_constant",
 /* 392 */ "function_body ::= compound_statement",
 /* 393 */ "statement ::= selection_statement",
 /* 394 */ "statement ::= iteration_statement",
 /* 395 */ "statement ::= array_iteration_statement",
 /* 396 */ "statement ::= two_arg_iteration_statement",
 /* 397 */ "statement ::= three_arg_iteration_statement",
 /* 398 */ "statement ::= typed_iteration_statement",
 /* 399 */ "statement ::= jump_statement",
 /* 400 */ "variable_list_with_init ::= var_init",
};
#endif /* NDEBUG */


#if YYSTACKDEPTH<=0
/*
** Try to increase the size of the parser stack.  Return the number
** of errors.  Return 0 on success.
*/
static int yyGrowStack(yyParser *p){
  int newSize;
  int idx;
  yyStackEntry *pNew;

  newSize = p->yystksz*2 + 100;
  idx = p->yytos ? (int)(p->yytos - p->yystack) : 0;
  if( p->yystack==&p->yystk0 ){
    pNew = (yyStackEntry *)malloc(newSize*sizeof(pNew[0]));
    if( pNew ) pNew[0] = p->yystk0;
  }else{
    pNew = (yyStackEntry *)realloc(p->yystack, newSize*sizeof(pNew[0]));
  }
  if( pNew ){
    p->yystack = pNew;
    p->yytos = &p->yystack[idx];
#ifndef NDEBUG
    if( yyTraceFILE ){
      fprintf(yyTraceFILE,"%sStack grows from %d to %d entries.\n",
              yyTracePrompt, p->yystksz, newSize);
      fflush(yyTraceFILE);
    }
#endif
    p->yystksz = newSize;
  }
  return pNew==0; 
}
#endif

/* Datatype of the argument to the memory allocated passed as the
** second argument to ZCCParseAlloc() below.  This can be changed by
** putting an appropriate #define in the %include section of the input
** grammar.
*/
#ifndef YYMALLOCARGTYPE
# define YYMALLOCARGTYPE size_t
#endif

/* 
** This function allocates a new parser.
** The only argument is a pointer to a function which works like
** malloc.
**
** Inputs:
** A pointer to the function used to allocate memory.
**
** Outputs:
** A pointer to a parser.  This pointer is used in subsequent calls
** to ZCCParse and ZCCParseFree.
*/
void *ZCCParseAlloc(void *(CDECL *mallocProc)(YYMALLOCARGTYPE)){
  yyParser *pParser;
  pParser = (yyParser*)(*mallocProc)( (YYMALLOCARGTYPE)sizeof(yyParser) );
  if( pParser ){
#ifdef YYTRACKMAXSTACKDEPTH
    pParser->yyhwm = 0;
#endif
#if YYSTACKDEPTH<=0
    pParser->yytos = NULL;
    pParser->yystack = NULL;
    pParser->yystksz = 0;
    if( yyGrowStack(pParser) ){
      pParser->yystack = &pParser->yystk0;
      pParser->yystksz = 1;
    }
#endif
#ifndef YYNOERRORRECOVERY
    pParser->yyerrcnt = -1;
#endif
    pParser->yytos = pParser->yystack;
    pParser->yystack[0].stateno = 0;
    pParser->yystack[0].major = 0;
  }
  return pParser;
}

/* The following function deletes the "minor type" or semantic value
** associated with a symbol.  The symbol can be either a terminal
** or nonterminal. "yymajor" is the symbol code, and "yypminor" is
** a pointer to the value to be deleted.  The code used to do the 
** deletions is derived from the %destructor and/or %token_destructor
** directives of the input grammar.
*/
static void yy_destructor(
  yyParser *yypParser,    /* The parser */
  YYCODETYPE yymajor,     /* Type code for object to destroy */
  YYMINORTYPE *yypminor   /* The object to be destroyed */
){
  ZCCParseARG_FETCH;
  switch( yymajor ){
    /* Here is inserted the actions which take place when a
    ** terminal or non-terminal is destroyed.  This can happen
    ** when the symbol is popped from the stack during a
    ** reduce or during error processing or when a parser is 
    ** being destroyed before it is finished parsing.
    **
    ** Note: during a reduce, the only symbols destroyed are those
    ** which appear on the RHS of the rule, but which are *not* used
    ** inside the C code.
    */
/********* Begin destructor definitions ***************************************/
      /* TERMINAL Destructor */
    case 1: /* EQ */
    case 2: /* MULEQ */
    case 3: /* DIVEQ */
    case 4: /* MODEQ */
    case 5: /* ADDEQ */
    case 6: /* SUBEQ */
    case 7: /* LSHEQ */
    case 8: /* RSHEQ */
    case 9: /* ANDEQ */
    case 10: /* OREQ */
    case 11: /* XOREQ */
    case 12: /* URSHEQ */
    case 13: /* QUESTION */
    case 14: /* COLON */
    case 15: /* OROR */
    case 16: /* ANDAND */
    case 17: /* EQEQ */
    case 18: /* NEQ */
    case 19: /* APPROXEQ */
    case 20: /* LT */
    case 21: /* GT */
    case 22: /* LTEQ */
    case 23: /* GTEQ */
    case 24: /* LTGTEQ */
    case 25: /* IS */
    case 26: /* DOTDOT */
    case 27: /* OR */
    case 28: /* XOR */
    case 29: /* AND */
    case 30: /* LSH */
    case 31: /* RSH */
    case 32: /* URSH */
    case 33: /* SUB */
    case 34: /* ADD */
    case 35: /* MUL */
    case 36: /* DIV */
    case 37: /* MOD */
    case 38: /* CROSSPROD */
    case 39: /* DOTPROD */
    case 40: /* POW */
    case 41: /* UNARY */
    case 42: /* ADDADD */
    case 43: /* SUBSUB */
    case 44: /* DOT */
    case 45: /* LPAREN */
    case 46: /* LBRACKET */
    case 47: /* SCOPE */
    case 48: /* EOF */
    case 49: /* SEMICOLON */
    case 50: /* COMMA */
    case 51: /* INCLUDE */
    case 52: /* EXTEND */
    case 53: /* CLASS */
    case 54: /* IDENTIFIER */
    case 55: /* ABSTRACT */
    case 56: /* FINAL */
    case 57: /* NATIVE */
    case 58: /* UI */
    case 59: /* PLAY */
    case 60: /* REPLACES */
    case 61: /* VERSION */
    case 62: /* STRCONST */
    case 63: /* RPAREN */
    case 64: /* SEALED */
    case 65: /* ACTION */
    case 66: /* DEFAULT */
    case 67: /* COLOR */
    case 68: /* LBRACE */
    case 69: /* RBRACE */
    case 70: /* MIXIN */
    case 71: /* PROPERTY */
    case 72: /* FLAGDEF */
    case 73: /* INTCONST */
    case 74: /* INTERNAL */
    case 75: /* STRUCT */
    case 76: /* CLEARSCOPE */
    case 77: /* CONST */
    case 78: /* ENUM */
    case 79: /* STATES */
    case 80: /* NWS */
    case 81: /* STOP */
    case 82: /* WAIT */
    case 83: /* FAIL */
    case 84: /* LOOP */
    case 85: /* GOTO */
    case 86: /* SUPER */
    case 87: /* BRIGHT */
    case 88: /* FAST */
    case 89: /* SLOW */
    case 90: /* NODELAY */
    case 91: /* CANRAISE */
    case 92: /* OFFSET */
    case 93: /* LIGHT */
    case 94: /* SBYTE */
    case 95: /* BYTE */
    case 96: /* SHORT */
    case 97: /* USHORT */
    case 98: /* INT */
    case 99: /* UINT */
    case 100: /* BOOL */
    case 101: /* FLOAT */
    case 102: /* DOUBLE */
    case 103: /* VECTOR2 */
    case 104: /* VECTOR3 */
    case 105: /* VECTOR4 */
    case 106: /* NAME */
    case 107: /* SOUND */
    case 108: /* STATE */
    case 109: /* LET */
    case 110: /* ATSIGN */
    case 111: /* READONLY */
    case 112: /* STRING */
    case 113: /* MAP */
    case 114: /* MAPITERATOR */
    case 115: /* ARRAY */
    case 116: /* VOID */
    case 117: /* UINT8 */
    case 118: /* INT8 */
    case 119: /* UINT16 */
    case 120: /* INT16 */
    case 121: /* FNTYPE */
    case 122: /* RBRACKET */
    case 123: /* DEPRECATED */
    case 124: /* STATIC */
    case 125: /* PRIVATE */
    case 126: /* PROTECTED */
    case 127: /* LATENT */
    case 128: /* META */
    case 129: /* TRANSIENT */
    case 130: /* VIRTUAL */
    case 131: /* OVERRIDE */
    case 132: /* VARARG */
    case 133: /* VIRTUALSCOPE */
    case 134: /* ELLIPSIS */
    case 135: /* IN */
    case 136: /* OUT */
    case 137: /* OPTIONAL */
    case 138: /* TILDE */
    case 139: /* BANG */
    case 140: /* SIZEOF */
    case 141: /* ALIGNOF */
    case 142: /* UINTCONST */
    case 143: /* FLOATCONST */
    case 144: /* NAMECONST */
    case 145: /* FALSE */
    case 146: /* TRUE */
    case 147: /* NULLPTR */
    case 148: /* STATICCONST */
    case 149: /* CONTINUE */
    case 150: /* BREAK */
    case 151: /* RETURN */
    case 152: /* DO */
    case 153: /* FOR */
    case 154: /* FOREACH */
    case 155: /* WHILE */
    case 156: /* UNTIL */
    case 157: /* IF */
    case 158: /* ELSE */
    case 159: /* SWITCH */
    case 160: /* CASE */
{
#line 118 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"

#line 2146 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
      break;
/********* End destructor definitions *****************************************/
    default:  break;   /* If no destructor action specified: do nothing */
  }
}

/*
** Pop the parser's stack once.
**
** If there is a destructor routine associated with the token which
** is popped from the stack, then call it.
*/
static void yy_pop_parser_stack(yyParser *pParser){
  yyStackEntry *yytos;
  assert( pParser->yytos!=0 );
  assert( pParser->yytos > pParser->yystack );
  yytos = pParser->yytos--;
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sPopping %s\n",
      yyTracePrompt,
      yyTokenName[yytos->major]);
    fflush(yyTraceFILE);
  }
#endif
  yy_destructor(pParser, yytos->major, &yytos->minor);
}

/* 
** Deallocate and destroy a parser.  Destructors are called for
** all stack elements before shutting the parser down.
**
** If the YYPARSEFREENEVERNULL macro exists (for example because it
** is defined in a %include section of the input grammar) then it is
** assumed that the input pointer is never NULL.
*/
void ZCCParseFree(
  void *p,                    /* The parser to be deleted */
  void (CDECL *freeProc)(void*)     /* Function used to reclaim memory */
){
  yyParser *pParser = (yyParser*)p;
#ifndef YYPARSEFREENEVERNULL
  if( pParser==0 ) return;
#endif
  while( pParser->yytos>pParser->yystack ) yy_pop_parser_stack(pParser);
#if YYSTACKDEPTH<=0
  if( pParser->yystack!=&pParser->yystk0 ) free(pParser->yystack);
#endif
  (*freeProc)((void*)pParser);
}

/*
** Return the peak depth of the stack for a parser.
*/
#ifdef YYTRACKMAXSTACKDEPTH
int ZCCParseStackPeak(void *p){
  yyParser *pParser = (yyParser*)p;
  return pParser->yyhwm;
}
#endif

/*
** Find the appropriate action for a parser given the terminal
** look-ahead token iLookAhead.
*/
static unsigned int yy_find_shift_action(
  yyParser *pParser,        /* The parser */
  YYCODETYPE iLookAhead     /* The look-ahead token */
){
  int i;
  int stateno = pParser->yytos->stateno;
 
  if( stateno>=YY_MIN_REDUCE ) return stateno;
  assert( stateno <= YY_SHIFT_COUNT );
  do{
    i = yy_shift_ofst[stateno];
    assert( iLookAhead!=YYNOCODE );
    i += iLookAhead;
    if( i<0 || i>=YY_ACTTAB_COUNT || yy_lookahead[i]!=iLookAhead ){
#ifdef YYFALLBACK
      YYCODETYPE iFallback;            /* Fallback token */
      if( iLookAhead<sizeof(yyFallback)/sizeof(yyFallback[0])
             && (iFallback = yyFallback[iLookAhead])!=0 ){
#ifndef NDEBUG
        if( yyTraceFILE ){
          fprintf(yyTraceFILE, "%sFALLBACK %s => %s\n",
             yyTracePrompt, yyTokenName[iLookAhead], yyTokenName[iFallback]);
          fflush(yyTraceFILE);
        }
#endif
        assert( yyFallback[iFallback]==0 ); /* Fallback loop must terminate */
        iLookAhead = iFallback;
        continue;
      }
#endif
#ifdef YYWILDCARD
      {
        int j = i - iLookAhead + YYWILDCARD;
        if( 
#if YY_SHIFT_MIN+YYWILDCARD<0
          j>=0 &&
#endif
#if YY_SHIFT_MAX+YYWILDCARD>=YY_ACTTAB_COUNT
          j<YY_ACTTAB_COUNT &&
#endif
          yy_lookahead[j]==YYWILDCARD && iLookAhead>0
        ){
#ifndef NDEBUG
          if( yyTraceFILE ){
            fprintf(yyTraceFILE, "%sWILDCARD %s => %s\n",
               yyTracePrompt, yyTokenName[iLookAhead],
               yyTokenName[YYWILDCARD]);
            fflush(yyTraceFILE);
          }
#endif /* NDEBUG */
          return yy_action[j];
        }
      }
#endif /* YYWILDCARD */
      return yy_default[stateno];
    }else{
      return yy_action[i];
    }
  }while(1);
}

/*
** Find the appropriate action for a parser given the non-terminal
** look-ahead token iLookAhead.
*/
static int yy_find_reduce_action(
  int stateno,              /* Current state number */
  YYCODETYPE iLookAhead     /* The look-ahead token */
){
  int i;
#ifdef YYERRORSYMBOL
  if( stateno>YY_REDUCE_COUNT ){
    return yy_default[stateno];
  }
#else
  assert( stateno<=YY_REDUCE_COUNT );
#endif
  i = yy_reduce_ofst[stateno];
  assert( i!=YY_REDUCE_USE_DFLT );
  assert( iLookAhead!=YYNOCODE );
  i += iLookAhead;
#ifdef YYERRORSYMBOL
  if( i<0 || i>=YY_ACTTAB_COUNT || yy_lookahead[i]!=iLookAhead ){
    return yy_default[stateno];
  }
#else
  assert( i>=0 && i<YY_ACTTAB_COUNT );
  assert( yy_lookahead[i]==iLookAhead );
#endif
  return yy_action[i];
}

/*
** The following routine is called if the stack overflows.
*/
static void yyStackOverflow(yyParser *yypParser){
   ZCCParseARG_FETCH;
   yypParser->yytos--;
#ifndef NDEBUG
   if( yyTraceFILE ){
     fprintf(yyTraceFILE,"%sStack Overflow!\n",yyTracePrompt);
     fflush(yyTraceFILE);
   }
#endif
   while( yypParser->yytos>yypParser->yystack ) yy_pop_parser_stack(yypParser);
   /* Here code is inserted which will execute if the parser
   ** stack every overflows */
/******** Begin %stack_overflow code ******************************************/
/******** End %stack_overflow code ********************************************/
   ZCCParseARG_STORE; /* Suppress warning about unused %extra_argument var */
}

/*
** Print tracing information for a SHIFT action
*/
#ifndef NDEBUG
static void yyTraceShift(yyParser *yypParser, int yyNewState){
  if( yyTraceFILE ){
    if( yyNewState<YYNSTATE ){
      fprintf(yyTraceFILE,"%sShift '%s', go to state %d\n",
         yyTracePrompt,yyTokenName[yypParser->yytos->major],
         yyNewState);
    }else{
      fprintf(yyTraceFILE,"%sShift '%s'\n",
         yyTracePrompt,yyTokenName[yypParser->yytos->major]);
    }
    fflush(yyTraceFILE);
  }
}
#else
# define yyTraceShift(X,Y)
#endif

/*
** Perform a shift action.
*/
static void yy_shift(
  yyParser *yypParser,          /* The parser to be shifted */
  int yyNewState,               /* The new state to shift in */
  int yyMajor,                  /* The major token to shift in */
  ZCCParseTOKENTYPE yyMinor        /* The minor token to shift in */
){
  yyStackEntry *yytos;
  yypParser->yytos++;
#ifdef YYTRACKMAXSTACKDEPTH
  if( (int)(yypParser->yytos - yypParser->yystack)>yypParser->yyhwm ){
    yypParser->yyhwm++;
    assert( yypParser->yyhwm == (int)(yypParser->yytos - yypParser->yystack) );
  }
#endif
#if YYSTACKDEPTH>0 
  if( yypParser->yytos>=&yypParser->yystack[YYSTACKDEPTH] ){
    yyStackOverflow(yypParser);
    return;
  }
#else
  if( yypParser->yytos>=&yypParser->yystack[yypParser->yystksz] ){
    if( yyGrowStack(yypParser) ){
      yyStackOverflow(yypParser);
      return;
    }
  }
#endif
  if( yyNewState > YY_MAX_SHIFT ){
    yyNewState += YY_MIN_REDUCE - YY_MIN_SHIFTREDUCE;
  }
  yytos = yypParser->yytos;
  yytos->stateno = (YYACTIONTYPE)yyNewState;
  yytos->major = (YYCODETYPE)yyMajor;
  yytos->minor.yy0 = yyMinor;
  yyTraceShift(yypParser, yyNewState);
}

/* The following table contains information about every rule that
** is used during the reduce.
*/
static const struct {
  YYCODETYPE lhs;         /* Symbol on the left-hand side of the rule */
  unsigned char nrhs;     /* Number of right-hand side symbols in the rule */
} yyRuleInfo[] = {
  { 166, 1 },
  { 167, 0 },
  { 167, 2 },
  { 167, 1 },
  { 168, 1 },
  { 168, 1 },
  { 168, 1 },
  { 168, 1 },
  { 168, 1 },
  { 168, 1 },
  { 177, 0 },
  { 174, 2 },
  { 170, 2 },
  { 180, 3 },
  { 180, 4 },
  { 184, 0 },
  { 184, 2 },
  { 185, 0 },
  { 185, 2 },
  { 185, 2 },
  { 185, 2 },
  { 185, 2 },
  { 185, 2 },
  { 185, 3 },
  { 185, 5 },
  { 185, 5 },
  { 186, 1 },
  { 186, 1 },
  { 186, 3 },
  { 186, 3 },
  { 186, 3 },
  { 183, 3 },
  { 183, 3 },
  { 181, 0 },
  { 181, 2 },
  { 182, 1 },
  { 182, 1 },
  { 182, 1 },
  { 182, 1 },
  { 182, 1 },
  { 182, 1 },
  { 182, 1 },
  { 182, 1 },
  { 182, 1 },
  { 182, 1 },
  { 188, 3 },
  { 189, 5 },
  { 190, 7 },
  { 190, 8 },
  { 197, 1 },
  { 197, 3 },
  { 171, 7 },
  { 171, 7 },
  { 198, 0 },
  { 198, 2 },
  { 198, 2 },
  { 198, 2 },
  { 198, 2 },
  { 198, 5 },
  { 194, 0 },
  { 194, 1 },
  { 195, 2 },
  { 196, 1 },
  { 196, 1 },
  { 196, 1 },
  { 196, 1 },
  { 196, 1 },
  { 173, 5 },
  { 172, 7 },
  { 202, 0 },
  { 202, 2 },
  { 199, 1 },
  { 199, 3 },
  { 200, 0 },
  { 201, 1 },
  { 201, 3 },
  { 169, 1 },
  { 204, 6 },
  { 205, 0 },
  { 205, 2 },
  { 206, 1 },
  { 206, 1 },
  { 206, 1 },
  { 206, 1 },
  { 206, 1 },
  { 206, 1 },
  { 206, 1 },
  { 206, 1 },
  { 206, 1 },
  { 191, 6 },
  { 217, 0 },
  { 217, 3 },
  { 187, 1 },
  { 187, 3 },
  { 218, 0 },
  { 207, 0 },
  { 207, 1 },
  { 207, 2 },
  { 207, 2 },
  { 207, 2 },
  { 209, 2 },
  { 211, 1 },
  { 211, 1 },
  { 211, 1 },
  { 211, 1 },
  { 211, 3 },
  { 211, 5 },
  { 211, 5 },
  { 212, 0 },
  { 212, 2 },
  { 208, 5 },
  { 216, 0 },
  { 216, 2 },
  { 216, 2 },
  { 216, 2 },
  { 216, 2 },
  { 216, 2 },
  { 216, 7 },
  { 216, 5 },
  { 219, 1 },
  { 219, 3 },
  { 213, 4 },
  { 213, 3 },
  { 213, 4 },
  { 213, 3 },
  { 214, 0 },
  { 214, 2 },
  { 215, 0 },
  { 215, 3 },
  { 192, 3 },
  { 192, 4 },
  { 192, 4 },
  { 222, 1 },
  { 222, 2 },
  { 223, 1 },
  { 223, 2 },
  { 223, 1 },
  { 223, 1 },
  { 225, 2 },
  { 225, 2 },
  { 224, 3 },
  { 224, 2 },
  { 203, 1 },
  { 203, 1 },
  { 203, 1 },
  { 203, 1 },
  { 203, 1 },
  { 203, 1 },
  { 228, 1 },
  { 228, 1 },
  { 228, 1 },
  { 228, 1 },
  { 228, 1 },
  { 228, 1 },
  { 228, 1 },
  { 228, 1 },
  { 228, 1 },
  { 228, 1 },
  { 228, 1 },
  { 227, 1 },
  { 227, 1 },
  { 227, 2 },
  { 227, 4 },
  { 227, 5 },
  { 227, 2 },
  { 229, 6 },
  { 229, 6 },
  { 229, 4 },
  { 229, 1 },
  { 241, 1 },
  { 241, 1 },
  { 241, 1 },
  { 237, 8 },
  { 237, 4 },
  { 238, 0 },
  { 238, 1 },
  { 239, 3 },
  { 240, 2 },
  { 240, 3 },
  { 229, 2 },
  { 234, 0 },
  { 234, 3 },
  { 230, 1 },
  { 230, 1 },
  { 233, 2 },
  { 231, 3 },
  { 232, 1 },
  { 236, 3 },
  { 235, 2 },
  { 162, 3 },
  { 243, 6 },
  { 243, 2 },
  { 243, 2 },
  { 248, 1 },
  { 248, 2 },
  { 247, 3 },
  { 244, 0 },
  { 244, 2 },
  { 244, 3 },
  { 250, 0 },
  { 250, 2 },
  { 244, 6 },
  { 244, 5 },
  { 249, 1 },
  { 249, 1 },
  { 249, 1 },
  { 249, 1 },
  { 249, 1 },
  { 249, 1 },
  { 249, 1 },
  { 249, 1 },
  { 249, 1 },
  { 249, 1 },
  { 249, 1 },
  { 249, 1 },
  { 249, 1 },
  { 249, 1 },
  { 249, 1 },
  { 249, 1 },
  { 249, 1 },
  { 249, 1 },
  { 246, 0 },
  { 246, 1 },
  { 164, 1 },
  { 245, 0 },
  { 245, 1 },
  { 245, 3 },
  { 251, 3 },
  { 252, 3 },
  { 252, 5 },
  { 252, 4 },
  { 252, 6 },
  { 242, 0 },
  { 242, 2 },
  { 242, 2 },
  { 242, 2 },
  { 253, 1 },
  { 253, 1 },
  { 253, 1 },
  { 253, 9 },
  { 253, 7 },
  { 253, 5 },
  { 253, 3 },
  { 253, 4 },
  { 253, 9 },
  { 253, 6 },
  { 253, 4 },
  { 253, 3 },
  { 253, 2 },
  { 253, 2 },
  { 254, 2 },
  { 254, 2 },
  { 254, 2 },
  { 254, 2 },
  { 254, 2 },
  { 254, 2 },
  { 254, 2 },
  { 254, 2 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 3 },
  { 178, 5 },
  { 226, 3 },
  { 221, 3 },
  { 256, 0 },
  { 257, 3 },
  { 257, 1 },
  { 179, 1 },
  { 179, 2 },
  { 255, 1 },
  { 255, 1 },
  { 255, 1 },
  { 255, 1 },
  { 255, 1 },
  { 255, 1 },
  { 255, 1 },
  { 259, 1 },
  { 259, 1 },
  { 259, 1 },
  { 259, 2 },
  { 259, 2 },
  { 259, 2 },
  { 259, 2 },
  { 259, 2 },
  { 259, 1 },
  { 193, 10 },
  { 193, 10 },
  { 268, 2 },
  { 268, 2 },
  { 268, 2 },
  { 268, 3 },
  { 258, 2 },
  { 258, 3 },
  { 258, 3 },
  { 220, 1 },
  { 220, 2 },
  { 261, 1 },
  { 263, 5 },
  { 263, 6 },
  { 263, 9 },
  { 264, 7 },
  { 265, 9 },
  { 266, 11 },
  { 267, 8 },
  { 272, 1 },
  { 272, 1 },
  { 273, 1 },
  { 273, 1 },
  { 274, 0 },
  { 274, 1 },
  { 274, 3 },
  { 262, 1 },
  { 262, 3 },
  { 275, 5 },
  { 262, 5 },
  { 260, 3 },
  { 260, 2 },
  { 269, 5 },
  { 270, 6 },
  { 271, 2 },
  { 277, 1 },
  { 277, 2 },
  { 277, 3 },
  { 277, 5 },
  { 277, 6 },
  { 277, 5 },
  { 276, 3 },
  { 167, 2 },
  { 175, 0 },
  { 175, 1 },
  { 176, 0 },
  { 176, 1 },
  { 177, 1 },
  { 194, 1 },
  { 195, 1 },
  { 199, 1 },
  { 200, 2 },
  { 210, 3 },
  { 228, 1 },
  { 238, 1 },
  { 239, 1 },
  { 233, 1 },
  { 231, 1 },
  { 232, 1 },
  { 235, 1 },
  { 247, 1 },
  { 164, 1 },
  { 245, 1 },
  { 251, 1 },
  { 253, 3 },
  { 254, 1 },
  { 178, 1 },
  { 226, 1 },
  { 221, 1 },
  { 256, 1 },
  { 255, 1 },
  { 165, 1 },
  { 259, 1 },
  { 259, 1 },
  { 259, 1 },
  { 259, 1 },
  { 259, 1 },
  { 259, 1 },
  { 259, 1 },
  { 276, 1 },
};

static void yy_accept(yyParser*);  /* Forward Declaration */

/*
** Perform a reduce action and the shift that must immediately
** follow the reduce.
*/
static void yy_reduce(
  yyParser *yypParser,         /* The parser */
  unsigned int yyruleno        /* Number of the rule by which to reduce */
){
  int yygoto;                     /* The next state */
  int yyact;                      /* The next action */
  yyStackEntry *yymsp;            /* The top of the parser's stack */
  int yysize;                     /* Amount to pop the stack */
  ZCCParseARG_FETCH;
  yymsp = yypParser->yytos;
#ifndef NDEBUG
  if( yyTraceFILE && yyruleno<(int)(sizeof(yyRuleName)/sizeof(yyRuleName[0])) ){
    yysize = yyRuleInfo[yyruleno].nrhs;
    fprintf(yyTraceFILE, "%sReduce [%s], go to state %d.\n", yyTracePrompt,
      yyRuleName[yyruleno], yymsp[-yysize].stateno);
    fflush(yyTraceFILE);
  }
#endif /* NDEBUG */

  /* Check that the stack is large enough to grow by a single entry
  ** if the RHS of the rule is empty.  This ensures that there is room
  ** enough on the stack to push the LHS value */
  if( yyRuleInfo[yyruleno].nrhs==0 ){
#ifdef YYTRACKMAXSTACKDEPTH
    if( (int)(yypParser->yytos - yypParser->yystack)>yypParser->yyhwm ){
      yypParser->yyhwm++;
      assert( yypParser->yyhwm == (int)(yypParser->yytos - yypParser->yystack));
    }
#endif
#if YYSTACKDEPTH>0 
    if( yypParser->yytos>=&yypParser->yystack[YYSTACKDEPTH-1] ){
      yyStackOverflow(yypParser);
      return;
    }
#else
    if( yypParser->yytos>=&yypParser->yystack[yypParser->yystksz-1] ){
      if( yyGrowStack(yypParser) ){
        yyStackOverflow(yypParser);
        return;
      }
      yymsp = yypParser->yytos;
    }
#endif
  }

  switch( yyruleno ){
  /* Beginning here are the reduction cases.  A typical example
  ** follows:
  **   case 0:
  **  #line <lineno> <grammarfile>
  **     { ... }           // User supplied code
  **  #line <lineno> <thisfile>
  **     break;
  */
/********** Begin reduce actions **********************************************/
        YYMINORTYPE yylhsminor;
      case 0: /* main ::= translation_unit */
#line 174 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ stat->TopNode = yymsp[0].minor.yy128; DPrintf(DMSG_SPAMMY, "Parse complete\n"); }
#line 2861 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 1: /* translation_unit ::= */
      case 33: /*class_innards ::= */ yytestcase(yyruleno==33);
      case 59: /*opt_struct_body ::= */ yytestcase(yyruleno==59);
      case 78: /*mixin_class_body ::= */ yytestcase(yyruleno==78);
#line 177 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[1].minor.yy128 = NULL; }
#line 2869 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 2: /* translation_unit ::= translation_unit external_declaration */
      case 34: /*class_innards ::= class_innards class_member */ yytestcase(yyruleno==34);
      case 79: /*mixin_class_body ::= mixin_class_body mixin_class_member */ yytestcase(yyruleno==79);
#line 178 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ SAFE_APPEND(yymsp[-1].minor.yy128,yymsp[0].minor.yy128); }
#line 2876 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 3: /* translation_unit ::= error */
      case 60: /*opt_struct_body ::= error */ yytestcase(yyruleno==60);
#line 180 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy128 = NULL; }
#line 2882 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 4: /* external_declaration ::= mixin_definition */
#line 183 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy128 = yymsp[0].minor.yy498; /*X-overwrites-A*/ }
#line 2887 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 5: /* external_declaration ::= class_definition */
#line 184 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy128 = yymsp[0].minor.yy392; /*X-overwrites-A*/ }
#line 2892 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 6: /* external_declaration ::= struct_def */
      case 38: /*class_member ::= struct_def */ yytestcase(yyruleno==38);
      case 82: /*mixin_class_member ::= struct_def */ yytestcase(yyruleno==82);
#line 185 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy128 = yymsp[0].minor.yy211; /*X-overwrites-A*/ }
#line 2899 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 7: /* external_declaration ::= enum_def */
      case 37: /*class_member ::= enum_def */ yytestcase(yyruleno==37);
      case 63: /*struct_member ::= enum_def */ yytestcase(yyruleno==63);
      case 81: /*mixin_class_member ::= enum_def */ yytestcase(yyruleno==81);
#line 186 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy128 = yymsp[0].minor.yy239; /*X-overwrites-A*/ }
#line 2907 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 8: /* external_declaration ::= const_def */
      case 41: /*class_member ::= const_def */ yytestcase(yyruleno==41);
      case 64: /*struct_member ::= const_def */ yytestcase(yyruleno==64);
      case 85: /*mixin_class_member ::= const_def */ yytestcase(yyruleno==85);
#line 187 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy128 = yymsp[0].minor.yy139; /*X-overwrites-A*/ }
#line 2915 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 9: /* external_declaration ::= include_def */
#line 188 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy128 = nullptr; }
#line 2920 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 10: /* opt_expr ::= */
#line 199 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	yymsp[1].minor.yy282 = NULL;
}
#line 2927 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 11: /* include_def ::= INCLUDE string_constant */
{  yy_destructor(yypParser,51,&yymsp[-1].minor);
#line 206 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	AddInclude(yymsp[0].minor.yy423);
}
#line 2935 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 12: /* class_definition ::= class_head class_body */
#line 220 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	yymsp[-1].minor.yy392->Body = yymsp[0].minor.yy128;
	yymsp[-1].minor.yy392 = yymsp[-1].minor.yy392; /*X-overwrites-A*/
}
#line 2944 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 13: /* class_head ::= EXTEND CLASS IDENTIFIER */
{  yy_destructor(yypParser,52,&yymsp[-2].minor);
#line 226 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(Class,head,yymsp[-1].minor.yy0);
	head->NodeName = yymsp[0].minor.yy0.Name();
	head->ParentName = nullptr;
	head->Flags =  ZCC_Extension;
	head->Replaces = nullptr;
	head->Version = {0, 0};
	head->Type = nullptr;
	head->Symbol = nullptr;
	yymsp[-2].minor.yy392 = head;
}
#line 2960 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 14: /* class_head ::= CLASS IDENTIFIER class_ancestry class_flags */
#line 240 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(Class,head,yymsp[-3].minor.yy0);
	head->NodeName = yymsp[-2].minor.yy0.Name();
	head->ParentName = yymsp[-1].minor.yy493;
	head->Flags = yymsp[0].minor.yy515.Flags;
	head->Replaces = yymsp[0].minor.yy515.Replaces;
	head->Sealed = yymsp[0].minor.yy515.Sealed;
	head->Version = yymsp[0].minor.yy515.Version;
	head->Type = nullptr;
	head->Symbol = nullptr;
	yylhsminor.yy392 = head;
}
#line 2977 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[-3].minor.yy392 = yylhsminor.yy392;
        break;
      case 15: /* class_ancestry ::= */
      case 180: /*class_restrictor ::= */ yytestcase(yyruleno==180);
#line 254 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[1].minor.yy493 = NULL; }
#line 2984 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 16: /* class_ancestry ::= COLON dottable_id */
{  yy_destructor(yypParser,14,&yymsp[-1].minor);
#line 255 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-1].minor.yy493 = yymsp[0].minor.yy493; /*yymsp[-1].minor.yy493-overwrites-yymsp[0].minor.yy493*/ }
#line 2990 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 17: /* class_flags ::= */
#line 258 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[1].minor.yy515.Flags = 0; yymsp[1].minor.yy515.Replaces = NULL; yymsp[1].minor.yy515.Version = {0,0}; yymsp[1].minor.yy515.Sealed = NULL; }
#line 2996 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 18: /* class_flags ::= class_flags ABSTRACT */
#line 259 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy515.Flags = yymsp[-1].minor.yy515.Flags | ZCC_Abstract; yylhsminor.yy515.Replaces = yymsp[-1].minor.yy515.Replaces; yylhsminor.yy515.Version = yymsp[-1].minor.yy515.Version; yylhsminor.yy515.Sealed = yymsp[-1].minor.yy515.Sealed; }
#line 3001 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,55,&yymsp[0].minor);
  yymsp[-1].minor.yy515 = yylhsminor.yy515;
        break;
      case 19: /* class_flags ::= class_flags FINAL */
#line 260 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy515.Flags = yymsp[-1].minor.yy515.Flags | ZCC_Final; yylhsminor.yy515.Replaces = yymsp[-1].minor.yy515.Replaces; yylhsminor.yy515.Version = yymsp[-1].minor.yy515.Version; yylhsminor.yy515.Sealed =  yymsp[-1].minor.yy515.Sealed;}
#line 3008 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,56,&yymsp[0].minor);
  yymsp[-1].minor.yy515 = yylhsminor.yy515;
        break;
      case 20: /* class_flags ::= class_flags NATIVE */
#line 261 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy515.Flags = yymsp[-1].minor.yy515.Flags | ZCC_Native; yylhsminor.yy515.Replaces = yymsp[-1].minor.yy515.Replaces; yylhsminor.yy515.Version = yymsp[-1].minor.yy515.Version; yylhsminor.yy515.Sealed =  yymsp[-1].minor.yy515.Sealed; }
#line 3015 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,57,&yymsp[0].minor);
  yymsp[-1].minor.yy515 = yylhsminor.yy515;
        break;
      case 21: /* class_flags ::= class_flags UI */
#line 262 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy515.Flags = yymsp[-1].minor.yy515.Flags | ZCC_UIFlag; yylhsminor.yy515.Replaces = yymsp[-1].minor.yy515.Replaces; yylhsminor.yy515.Version = yymsp[-1].minor.yy515.Version; yylhsminor.yy515.Sealed =  yymsp[-1].minor.yy515.Sealed; }
#line 3022 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,58,&yymsp[0].minor);
  yymsp[-1].minor.yy515 = yylhsminor.yy515;
        break;
      case 22: /* class_flags ::= class_flags PLAY */
#line 263 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy515.Flags = yymsp[-1].minor.yy515.Flags | ZCC_Play; yylhsminor.yy515.Replaces = yymsp[-1].minor.yy515.Replaces; yylhsminor.yy515.Version = yymsp[-1].minor.yy515.Version; yylhsminor.yy515.Sealed =  yymsp[-1].minor.yy515.Sealed; }
#line 3029 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,59,&yymsp[0].minor);
  yymsp[-1].minor.yy515 = yylhsminor.yy515;
        break;
      case 23: /* class_flags ::= class_flags REPLACES dottable_id */
#line 264 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy515.Flags = yymsp[-2].minor.yy515.Flags; yylhsminor.yy515.Replaces = yymsp[0].minor.yy493; yylhsminor.yy515.Version = yymsp[-2].minor.yy515.Version; yylhsminor.yy515.Sealed = yymsp[-2].minor.yy515.Sealed; }
#line 3036 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,60,&yymsp[-1].minor);
  yymsp[-2].minor.yy515 = yylhsminor.yy515;
        break;
      case 24: /* class_flags ::= class_flags VERSION LPAREN STRCONST RPAREN */
#line 265 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy515.Flags = yymsp[-4].minor.yy515.Flags | ZCC_Version; yylhsminor.yy515.Replaces = yymsp[-4].minor.yy515.Replaces; yylhsminor.yy515.Version = yymsp[-1].minor.yy0.String->GetChars(); yylhsminor.yy515.Sealed =  yymsp[-4].minor.yy515.Sealed; }
#line 3043 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,61,&yymsp[-3].minor);
  yy_destructor(yypParser,45,&yymsp[-2].minor);
  yy_destructor(yypParser,63,&yymsp[0].minor);
  yymsp[-4].minor.yy515 = yylhsminor.yy515;
        break;
      case 25: /* class_flags ::= class_flags SEALED LPAREN states_opt RPAREN */
#line 266 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy515.Flags = yymsp[-4].minor.yy515.Flags | ZCC_Sealed; yylhsminor.yy515.Replaces = yymsp[-4].minor.yy515.Replaces; yylhsminor.yy515.Version = yymsp[-4].minor.yy515.Version; yylhsminor.yy515.Sealed = yymsp[-1].minor.yy493; }
#line 3052 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,64,&yymsp[-3].minor);
  yy_destructor(yypParser,45,&yymsp[-2].minor);
  yy_destructor(yypParser,63,&yymsp[0].minor);
  yymsp[-4].minor.yy515 = yylhsminor.yy515;
        break;
      case 26: /* dottable_id ::= IDENTIFIER */
      case 49: /*identifier_list ::= IDENTIFIER */ yytestcase(yyruleno==49);
      case 92: /*states_opt ::= IDENTIFIER */ yytestcase(yyruleno==92);
#line 274 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(Identifier,id,yymsp[0].minor.yy0);
	id->Id = yymsp[0].minor.yy0.Name();
	yylhsminor.yy493 = id;
}
#line 3067 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy493 = yylhsminor.yy493;
        break;
      case 27: /* dottable_id ::= ACTION */
#line 281 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(Identifier,id,yymsp[0].minor.yy0);
	id->Id = NAME_Action;
	yylhsminor.yy493 = id;
}
#line 3077 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy493 = yylhsminor.yy493;
        break;
      case 28: /* dottable_id ::= dottable_id DOT IDENTIFIER */
#line 287 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(Identifier,id2,yymsp[-2].minor.yy493);
	id2->Id = yymsp[0].minor.yy0.Name();
	AppendTreeNodeSibling(yymsp[-2].minor.yy493, id2);
	yymsp[-2].minor.yy493 = yymsp[-2].minor.yy493; /*X-overwrites-A*/
}
#line 3088 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,44,&yymsp[-1].minor);
        break;
      case 29: /* dottable_id ::= dottable_id DOT DEFAULT */
#line 294 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(Identifier,id2,yymsp[-2].minor.yy493);
	id2->Id = NAME_Default;
	AppendTreeNodeSibling(yymsp[-2].minor.yy493, id2);
	yymsp[-2].minor.yy493 = yymsp[-2].minor.yy493; /*X-overwrites-A*/
}
#line 3099 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,44,&yymsp[-1].minor);
  yy_destructor(yypParser,66,&yymsp[0].minor);
        break;
      case 30: /* dottable_id ::= dottable_id DOT COLOR */
#line 305 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(Identifier,id2,yymsp[-2].minor.yy493);
	id2->Id = NAME_Color;
	AppendTreeNodeSibling(yymsp[-2].minor.yy493, id2);
	yymsp[-2].minor.yy493 = yymsp[-2].minor.yy493; /*X-overwrites-A*/
}
#line 3111 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,44,&yymsp[-1].minor);
  yy_destructor(yypParser,67,&yymsp[0].minor);
        break;
      case 31: /* class_body ::= SEMICOLON class_innards EOF */
{  yy_destructor(yypParser,49,&yymsp[-2].minor);
#line 322 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-2].minor.yy128 = yymsp[-1].minor.yy128; /*yymsp[-2].minor.yy128-overwrites-yymsp[-1].minor.yy128*/ }
#line 3119 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,48,&yymsp[0].minor);
}
        break;
      case 32: /* class_body ::= LBRACE class_innards RBRACE */
{  yy_destructor(yypParser,68,&yymsp[-2].minor);
#line 323 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-2].minor.yy128 = yymsp[-1].minor.yy128; /*yymsp[-2].minor.yy128-overwrites-yymsp[-1].minor.yy128*/ }
#line 3127 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,69,&yymsp[0].minor);
}
        break;
      case 35: /* class_member ::= declarator */
      case 62: /*struct_member ::= declarator */ yytestcase(yyruleno==62);
      case 80: /*mixin_class_member ::= declarator */ yytestcase(yyruleno==80);
#line 336 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy128 = yymsp[0].minor.yy327; /*X-overwrites-A*/ }
#line 3136 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 36: /* class_member ::= mixin_statement */
#line 337 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy128 = yymsp[0].minor.yy351; /*X-overwrites-A*/ }
#line 3141 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 39: /* class_member ::= states_def */
      case 83: /*mixin_class_member ::= states_def */ yytestcase(yyruleno==83);
#line 340 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy128 = yymsp[0].minor.yy82; /*X-overwrites-A*/ }
#line 3147 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 40: /* class_member ::= default_def */
      case 84: /*mixin_class_member ::= default_def */ yytestcase(yyruleno==84);
#line 341 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy128 = yymsp[0].minor.yy251; /*X-overwrites-A*/ }
#line 3153 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 42: /* class_member ::= property_def */
      case 86: /*mixin_class_member ::= property_def */ yytestcase(yyruleno==86);
#line 343 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy128 = yymsp[0].minor.yy240; /*X-overwrites-A*/ }
#line 3159 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 43: /* class_member ::= flag_def */
      case 66: /*struct_member ::= flag_def */ yytestcase(yyruleno==66);
      case 87: /*mixin_class_member ::= flag_def */ yytestcase(yyruleno==87);
#line 344 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy128 = yymsp[0].minor.yy343; /*X-overwrites-A*/ }
#line 3166 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 44: /* class_member ::= staticarray_statement */
      case 65: /*struct_member ::= staticarray_statement */ yytestcase(yyruleno==65);
      case 88: /*mixin_class_member ::= staticarray_statement */ yytestcase(yyruleno==88);
#line 345 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy128 = yymsp[0].minor.yy286; /*X-overwrites-A*/ }
#line 3173 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 45: /* mixin_statement ::= MIXIN IDENTIFIER SEMICOLON */
#line 350 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(MixinStmt,stmt,yymsp[-2].minor.yy0);
	stmt->MixinName = yymsp[-1].minor.yy0.Name();
	yylhsminor.yy351 = stmt;
}
#line 3182 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,49,&yymsp[0].minor);
  yymsp[-2].minor.yy351 = yylhsminor.yy351;
        break;
      case 46: /* property_def ::= PROPERTY IDENTIFIER COLON identifier_list SEMICOLON */
#line 365 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(Property,def,yymsp[-4].minor.yy0);
	def->NodeName = yymsp[-3].minor.yy0.Name();
	def->Body = yymsp[-1].minor.yy493;
	yylhsminor.yy240 = def;
}
#line 3194 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,14,&yymsp[-2].minor);
  yy_destructor(yypParser,49,&yymsp[0].minor);
  yymsp[-4].minor.yy240 = yylhsminor.yy240;
        break;
      case 47: /* flag_def ::= FLAGDEF IDENTIFIER COLON IDENTIFIER COMMA INTCONST SEMICOLON */
#line 373 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(FlagDef,def,yymsp[-6].minor.yy0);
	def->NodeName = yymsp[-5].minor.yy0.Name();
	def->RefName = yymsp[-3].minor.yy0.Name();
	def->BitValue = yymsp[-1].minor.yy0.Int;
	yylhsminor.yy343 = def;
}
#line 3208 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,14,&yymsp[-4].minor);
  yy_destructor(yypParser,50,&yymsp[-2].minor);
  yy_destructor(yypParser,49,&yymsp[0].minor);
  yymsp[-6].minor.yy343 = yylhsminor.yy343;
        break;
      case 48: /* flag_def ::= FLAGDEF INTERNAL IDENTIFIER COLON IDENTIFIER COMMA INTCONST SEMICOLON */
#line 382 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(FlagDef,def,yymsp[-7].minor.yy0);
	def->NodeName = yymsp[-5].minor.yy0.Name();
	def->RefName = yymsp[-3].minor.yy0.Name();
	def->BitValue = yymsp[-1].minor.yy0.Int | 0x10000;
	yylhsminor.yy343 = def;
}
#line 3223 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,74,&yymsp[-6].minor);
  yy_destructor(yypParser,14,&yymsp[-4].minor);
  yy_destructor(yypParser,50,&yymsp[-2].minor);
  yy_destructor(yypParser,49,&yymsp[0].minor);
  yymsp[-7].minor.yy343 = yylhsminor.yy343;
        break;
      case 50: /* identifier_list ::= states_opt COMMA IDENTIFIER */
      case 93: /*states_opt ::= states_opt COMMA IDENTIFIER */ yytestcase(yyruleno==93);
#line 399 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(Identifier,id,yymsp[0].minor.yy0);
	id->Id = yymsp[0].minor.yy0.Name();
	yymsp[-2].minor.yy493 = yymsp[-2].minor.yy493; /*X-overwrites-A*/
	AppendTreeNodeSibling(yymsp[-2].minor.yy493, id);
}
#line 3239 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,50,&yymsp[-1].minor);
        break;
      case 51: /* struct_def ::= STRUCT IDENTIFIER struct_flags LBRACE opt_struct_body RBRACE opt_semicolon */
#line 407 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(Struct,def,yymsp[-6].minor.yy0);
	def->NodeName = yymsp[-5].minor.yy0.Name();
	def->Body = yymsp[-2].minor.yy128;
	def->Type = nullptr;
	def->Symbol = nullptr;
	def->Version = yymsp[-4].minor.yy515.Version;
	def->Flags = yymsp[-4].minor.yy515.Flags;
	yylhsminor.yy211 = def;
}
#line 3254 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,68,&yymsp[-3].minor);
  yy_destructor(yypParser,69,&yymsp[-1].minor);
  yymsp[-6].minor.yy211 = yylhsminor.yy211;
        break;
      case 52: /* struct_def ::= EXTEND STRUCT IDENTIFIER LBRACE opt_struct_body RBRACE opt_semicolon */
{  yy_destructor(yypParser,52,&yymsp[-6].minor);
#line 419 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(Struct,def,yymsp[-5].minor.yy0);
	def->NodeName = yymsp[-4].minor.yy0.Name();
	def->Body = yymsp[-2].minor.yy128;
	def->Type = nullptr;
	def->Symbol = nullptr;
	def->Flags = ZCC_Extension;
	yymsp[-6].minor.yy211 = def;
}
#line 3271 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,68,&yymsp[-3].minor);
  yy_destructor(yypParser,69,&yymsp[-1].minor);
}
        break;
      case 53: /* struct_flags ::= */
#line 430 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[1].minor.yy515.Flags = 0; yymsp[1].minor.yy515.Version = {0, 0}; }
#line 3279 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 54: /* struct_flags ::= struct_flags UI */
#line 431 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy515.Flags = yymsp[-1].minor.yy515.Flags | ZCC_UIFlag; }
#line 3284 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,58,&yymsp[0].minor);
  yymsp[-1].minor.yy515 = yylhsminor.yy515;
        break;
      case 55: /* struct_flags ::= struct_flags PLAY */
#line 432 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy515.Flags = yymsp[-1].minor.yy515.Flags | ZCC_Play; }
#line 3291 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,59,&yymsp[0].minor);
  yymsp[-1].minor.yy515 = yylhsminor.yy515;
        break;
      case 56: /* struct_flags ::= struct_flags CLEARSCOPE */
#line 433 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy515.Flags = yymsp[-1].minor.yy515.Flags | ZCC_ClearScope; }
#line 3298 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,76,&yymsp[0].minor);
  yymsp[-1].minor.yy515 = yylhsminor.yy515;
        break;
      case 57: /* struct_flags ::= struct_flags NATIVE */
#line 434 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy515.Flags = yymsp[-1].minor.yy515.Flags | ZCC_Native; }
#line 3305 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,57,&yymsp[0].minor);
  yymsp[-1].minor.yy515 = yylhsminor.yy515;
        break;
      case 58: /* struct_flags ::= struct_flags VERSION LPAREN STRCONST RPAREN */
#line 435 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy515.Flags = yymsp[-4].minor.yy515.Flags | ZCC_Version; yylhsminor.yy515.Version = yymsp[-1].minor.yy0.String->GetChars(); }
#line 3312 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,61,&yymsp[-3].minor);
  yy_destructor(yypParser,45,&yymsp[-2].minor);
  yy_destructor(yypParser,63,&yymsp[0].minor);
  yymsp[-4].minor.yy515 = yylhsminor.yy515;
        break;
      case 61: /* struct_body ::= struct_body struct_member */
#line 443 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-1].minor.yy128 = yymsp[-1].minor.yy128; /*X-overwrites-A*/ AppendTreeNodeSibling(yymsp[-1].minor.yy128, yymsp[0].minor.yy128); }
#line 3321 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 67: /* const_def ::= CONST IDENTIFIER EQ expr SEMICOLON */
#line 454 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ConstantDef,def,yymsp[-4].minor.yy0);
	def->NodeName = yymsp[-3].minor.yy0.Name();
	def->Value = yymsp[-1].minor.yy282;
	def->Symbol = NULL;
	yylhsminor.yy139 = def;
}
#line 3332 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,1,&yymsp[-2].minor);
  yy_destructor(yypParser,49,&yymsp[0].minor);
  yymsp[-4].minor.yy139 = yylhsminor.yy139;
        break;
      case 68: /* enum_def ::= ENUM IDENTIFIER enum_type LBRACE opt_enum_list RBRACE opt_semicolon */
#line 471 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(Enum,def,yymsp[-6].minor.yy0);
	def->NodeName = yymsp[-5].minor.yy0.Name();
	def->EnumType = (EZCCBuiltinType)yymsp[-4].minor.yy0.Int;
	def->Elements = yymsp[-2].minor.yy139;
	def->Symbol = nullptr;

	// If the first element does not have an explicit value, make it 0.
	if (yymsp[-2].minor.yy139 != NULL)
	{
		ZCC_ConstantDef *node = yymsp[-2].minor.yy139, *prev = node;

		if (node->Value == NULL)
		{
			NEW_INTCONST_NODE(zero, TypeSInt32, 0, yymsp[-2].minor.yy139);
			node->Value = zero;
		}
		for (node = static_cast<ZCC_ConstantDef *>(node->SiblingNext);
			node != yymsp[-2].minor.yy139;
			prev = node, node = static_cast<ZCC_ConstantDef *>(node->SiblingNext))
		{
			assert(node->NodeType == AST_ConstantDef);
			// Leave explicit values alone.
			if (node->Value != NULL)
			{
				continue;
			}
			// Compute implicit values by adding one to the preceding value.
			assert(prev->Value != NULL);
			// If the preceding node is a constant, then we can do this now.
			if (prev->Value->Operation == PEX_ConstValue && prev->Value->Type->isInt())
			{
				NEW_INTCONST_NODE(cval, prev->Value->Type, static_cast<ZCC_ExprConstant *>(prev->Value)->IntVal + 1, node);
				node->Value = cval;
			}
			// Otherwise, create a new addition expression to add 1.
			else
			{
				NEW_INTCONST_NODE(one, TypeSInt32, 1, yymsp[-6].minor.yy0);
				NEW_AST_NODE(ExprID, label, node);
				label->Operation = PEX_ID;
				label->Identifier = prev->NodeName;
				label->Type = NULL;

				BINARY_EXPR(label, one, PEX_Add);
				node->Value = expr2;
			}
		}
		// Add a new terminating node, to indicate that the ConstantDefs for this enum are done.
		NEW_AST_NODE(EnumTerminator,term,yymsp[-1].minor.yy0);
		AppendTreeNodeSibling(yymsp[-2].minor.yy139, term);
	}
	if (yymsp[-2].minor.yy139 != NULL)
	{
		AppendTreeNodeSibling(def, yymsp[-2].minor.yy139);
	}
	yylhsminor.yy239 = def;
}
#line 3397 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,68,&yymsp[-3].minor);
  yymsp[-6].minor.yy239 = yylhsminor.yy239;
        break;
      case 69: /* enum_type ::= */
#line 530 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[1].minor.yy0.Int = ZCC_IntAuto; yymsp[1].minor.yy0.SourceLoc = stat->sc->GetMessageLine(); }
#line 3404 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 70: /* enum_type ::= COLON int_type */
{  yy_destructor(yypParser,14,&yymsp[-1].minor);
#line 531 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-1].minor.yy0 = yymsp[0].minor.yy0; /*yymsp[-1].minor.yy0-overwrites-yymsp[0].minor.yy0*/ }
#line 3410 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 71: /* enum_list ::= error */
#line 533 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy139 = NULL; }
#line 3416 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 72: /* enum_list ::= enum_list COMMA enumerator */
#line 535 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-2].minor.yy139 = yymsp[-2].minor.yy139; /*X-overwrites-A*/ AppendTreeNodeSibling(yymsp[-2].minor.yy139, yymsp[0].minor.yy139); }
#line 3421 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,50,&yymsp[-1].minor);
        break;
      case 73: /* opt_enum_list ::= */
#line 537 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[1].minor.yy139 = NULL; }
#line 3427 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 74: /* enumerator ::= IDENTIFIER */
#line 541 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ConstantDef,node,yymsp[0].minor.yy0);
	node->NodeName = yymsp[0].minor.yy0.Name();
	node->Value = NULL;
	node->Symbol = NULL;
	yylhsminor.yy139 = node;
}
#line 3438 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy139 = yylhsminor.yy139;
        break;
      case 75: /* enumerator ::= IDENTIFIER EQ expr */
#line 549 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ConstantDef,node,yymsp[-2].minor.yy0);
	node->NodeName = yymsp[-2].minor.yy0.Name();
	node->Value = yymsp[0].minor.yy282;
	node->Symbol = NULL;
	yylhsminor.yy139 = node;
}
#line 3450 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,1,&yymsp[-1].minor);
  yymsp[-2].minor.yy139 = yylhsminor.yy139;
        break;
      case 76: /* mixin_definition ::= mixin_class_definition */
#line 563 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy498 = yymsp[0].minor.yy498; /*X-overwrites-A*/ }
#line 3457 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 77: /* mixin_class_definition ::= MIXIN CLASS IDENTIFIER LBRACE mixin_class_body RBRACE */
#line 570 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(MixinDef,def,yymsp[-5].minor.yy0);
	def->Body = yymsp[-1].minor.yy128;
	def->NodeName = yymsp[-3].minor.yy0.Name();
	def->MixinType = ZCC_Mixin_Class;
	def->Symbol = nullptr;
	yylhsminor.yy498 = def;
}
#line 3469 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,53,&yymsp[-4].minor);
  yy_destructor(yypParser,68,&yymsp[-2].minor);
  yy_destructor(yypParser,69,&yymsp[0].minor);
  yymsp[-5].minor.yy498 = yylhsminor.yy498;
        break;
      case 89: /* states_def ::= STATES states_opts scanner_mode LBRACE states_body RBRACE */
#line 619 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(States,def,yymsp[-5].minor.yy0);
	def->Flags = yymsp[-4].minor.yy493;
	def->Body = yymsp[-1].minor.yy74;
	yylhsminor.yy82 = def;
}
#line 3483 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,68,&yymsp[-2].minor);
  yy_destructor(yypParser,69,&yymsp[0].minor);
  yymsp[-5].minor.yy82 = yylhsminor.yy82;
        break;
      case 90: /* states_opts ::= */
#line 626 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[1].minor.yy493 = nullptr; }
#line 3491 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 91: /* states_opts ::= LPAREN states_opt RPAREN */
{  yy_destructor(yypParser,45,&yymsp[-2].minor);
#line 627 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-2].minor.yy493 = yymsp[-1].minor.yy493; /*yymsp[-2].minor.yy493-overwrites-yymsp[-1].minor.yy493*/ }
#line 3497 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,63,&yymsp[0].minor);
}
        break;
      case 94: /* scanner_mode ::= */
#line 656 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ stat->sc->SetStateMode(true); }
#line 3504 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 95: /* states_body ::= */
#line 658 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[1].minor.yy74 = NULL; }
#line 3509 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 96: /* states_body ::= error */
#line 659 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy74 = NULL; }
#line 3514 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 97: /* states_body ::= states_body state_line */
      case 98: /*states_body ::= states_body state_label */ yytestcase(yyruleno==98);
      case 99: /*states_body ::= states_body state_flow */ yytestcase(yyruleno==99);
#line 660 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ SAFE_APPEND(yymsp[-1].minor.yy74,yymsp[0].minor.yy74); }
#line 3521 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 100: /* state_label ::= NWS COLON */
#line 665 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(StateLabel, label, yymsp[-1].minor.yy0);
	label->Label = yymsp[-1].minor.yy0.Name();
	yylhsminor.yy74 = label;
}
#line 3530 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,14,&yymsp[0].minor);
  yymsp[-1].minor.yy74 = yylhsminor.yy74;
        break;
      case 101: /* state_flow_type ::= STOP */
#line 673 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ NEW_AST_NODE(StateStop, flow, yymsp[0].minor.yy0); yylhsminor.yy74 = flow; }
#line 3537 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy74 = yylhsminor.yy74;
        break;
      case 102: /* state_flow_type ::= WAIT */
#line 674 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ NEW_AST_NODE(StateWait, flow, yymsp[0].minor.yy0); yylhsminor.yy74 = flow; }
#line 3543 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy74 = yylhsminor.yy74;
        break;
      case 103: /* state_flow_type ::= FAIL */
#line 675 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ NEW_AST_NODE(StateFail, flow, yymsp[0].minor.yy0); yylhsminor.yy74 = flow; }
#line 3549 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy74 = yylhsminor.yy74;
        break;
      case 104: /* state_flow_type ::= LOOP */
#line 676 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ NEW_AST_NODE(StateLoop, flow, yymsp[0].minor.yy0); yylhsminor.yy74 = flow; }
#line 3555 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy74 = yylhsminor.yy74;
        break;
      case 105: /* state_flow_type ::= GOTO dottable_id state_goto_offset */
#line 678 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(StateGoto, flow, yymsp[-2].minor.yy0);
	flow->Label = yymsp[-1].minor.yy493;
	flow->Offset = yymsp[0].minor.yy282;
	flow->Qualifier = nullptr;
	yylhsminor.yy74 = flow;
}
#line 3567 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[-2].minor.yy74 = yylhsminor.yy74;
        break;
      case 106: /* state_flow_type ::= GOTO IDENTIFIER SCOPE dottable_id state_goto_offset */
#line 687 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(StateGoto, flow, yymsp[-4].minor.yy0);
	flow->Label = yymsp[-1].minor.yy493;
	flow->Offset = yymsp[0].minor.yy282;

	NEW_AST_NODE(Identifier,id,yymsp[-3].minor.yy0);
	id->Id = yymsp[-3].minor.yy0.Name();
	flow->Qualifier =id;
	yylhsminor.yy74 = flow;
}
#line 3582 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,47,&yymsp[-2].minor);
  yymsp[-4].minor.yy74 = yylhsminor.yy74;
        break;
      case 107: /* state_flow_type ::= GOTO SUPER SCOPE dottable_id state_goto_offset */
#line 699 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(StateGoto, flow, yymsp[-4].minor.yy0);
	flow->Label = yymsp[-1].minor.yy493;
	flow->Offset = yymsp[0].minor.yy282;

	NEW_AST_NODE(Identifier,id,yymsp[-3].minor.yy0);
	id->Id = NAME_Super;
	flow->Qualifier =id;
	yylhsminor.yy74 = flow;
}
#line 3598 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,47,&yymsp[-2].minor);
  yymsp[-4].minor.yy74 = yylhsminor.yy74;
        break;
      case 108: /* state_goto_offset ::= */
#line 710 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[1].minor.yy282 = NULL; }
#line 3605 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 109: /* state_goto_offset ::= ADD expr */
{  yy_destructor(yypParser,34,&yymsp[-1].minor);
#line 711 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-1].minor.yy282 = yymsp[0].minor.yy282; /*yymsp[-1].minor.yy282-overwrites-yymsp[0].minor.yy282*/ }
#line 3611 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 110: /* state_line ::= NWS NWS expr state_opts state_action */
#line 714 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(StateLine, line, yymsp[-4].minor.yy0);
	line->Sprite = stat->Strings.Alloc(FName(yymsp[-4].minor.yy0.Name()).GetChars());
	line->Frames = stat->Strings.Alloc(FName(yymsp[-3].minor.yy0.Name()).GetChars());
	line->Duration = yymsp[-2].minor.yy282;
	line->bBright = yymsp[-1].minor.yy360.Bright;
	line->bFast = yymsp[-1].minor.yy360.Fast;
	line->bSlow = yymsp[-1].minor.yy360.Slow;
	line->bNoDelay = yymsp[-1].minor.yy360.NoDelay;
	line->bCanRaise = yymsp[-1].minor.yy360.CanRaise;
	line->Offset = yymsp[-1].minor.yy360.Offset;
	line->Lights = yymsp[-1].minor.yy360.Lights;
	line->Action = yymsp[0].minor.yy128;
	yylhsminor.yy74 = line;
}
#line 3631 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[-4].minor.yy74 = yylhsminor.yy74;
        break;
      case 111: /* state_opts ::= */
#line 730 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ StateOpts opts; opts.Zero(); yymsp[1].minor.yy360 = opts; }
#line 3637 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 112: /* state_opts ::= state_opts BRIGHT */
#line 731 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-1].minor.yy360.Bright = true; yymsp[-1].minor.yy360 = yymsp[-1].minor.yy360; /*X-overwrites-A*/ }
#line 3642 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,87,&yymsp[0].minor);
        break;
      case 113: /* state_opts ::= state_opts FAST */
#line 732 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-1].minor.yy360.Fast = true; yymsp[-1].minor.yy360 = yymsp[-1].minor.yy360; /*X-overwrites-A*/ }
#line 3648 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,88,&yymsp[0].minor);
        break;
      case 114: /* state_opts ::= state_opts SLOW */
#line 733 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-1].minor.yy360.Slow = true; yymsp[-1].minor.yy360 = yymsp[-1].minor.yy360; /*X-overwrites-A*/ }
#line 3654 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,89,&yymsp[0].minor);
        break;
      case 115: /* state_opts ::= state_opts NODELAY */
#line 734 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-1].minor.yy360.NoDelay = true; yymsp[-1].minor.yy360 = yymsp[-1].minor.yy360; /*X-overwrites-A*/ }
#line 3660 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,90,&yymsp[0].minor);
        break;
      case 116: /* state_opts ::= state_opts CANRAISE */
#line 735 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-1].minor.yy360.CanRaise = true; yymsp[-1].minor.yy360 = yymsp[-1].minor.yy360; /*X-overwrites-A*/ }
#line 3666 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,91,&yymsp[0].minor);
        break;
      case 117: /* state_opts ::= state_opts OFFSET LPAREN expr COMMA expr RPAREN */
#line 736 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-6].minor.yy360.Offset = yymsp[-3].minor.yy282; AppendTreeNodeSibling(yymsp[-3].minor.yy282, yymsp[-1].minor.yy282); yymsp[-6].minor.yy360 = yymsp[-6].minor.yy360; /*X-overwrites-A*/ }
#line 3672 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,92,&yymsp[-5].minor);
  yy_destructor(yypParser,45,&yymsp[-4].minor);
  yy_destructor(yypParser,50,&yymsp[-2].minor);
  yy_destructor(yypParser,63,&yymsp[0].minor);
        break;
      case 118: /* state_opts ::= state_opts LIGHT LPAREN light_list RPAREN */
#line 737 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-4].minor.yy360 = yymsp[-4].minor.yy360; /*X-overwrites-A*/ yymsp[-4].minor.yy360.Lights = yymsp[-1].minor.yy423; }
#line 3681 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,93,&yymsp[-3].minor);
  yy_destructor(yypParser,45,&yymsp[-2].minor);
  yy_destructor(yypParser,63,&yymsp[0].minor);
        break;
      case 119: /* light_list ::= STRCONST */
      case 303: /*string_constant ::= STRCONST */ yytestcase(yyruleno==303);
#line 742 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ExprConstant, strconst, yymsp[0].minor.yy0);
	strconst->Operation = PEX_ConstValue;
	strconst->Type = TypeString;
	strconst->StringVal = yymsp[0].minor.yy0.String;
	yylhsminor.yy423 = strconst;
}
#line 3696 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy423 = yylhsminor.yy423;
        break;
      case 120: /* light_list ::= light_list COMMA STRCONST */
#line 751 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ExprConstant, strconst, yymsp[0].minor.yy0);
	strconst->Operation = PEX_ConstValue;
	strconst->Type = TypeString;
	strconst->StringVal = yymsp[0].minor.yy0.String;
	AppendTreeNodeSibling(yymsp[-2].minor.yy423, strconst); 
	yymsp[-2].minor.yy423 = yymsp[-2].minor.yy423; /*X-overwrites-A*/
}
#line 3709 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,50,&yymsp[-1].minor);
        break;
      case 121: /* state_action ::= LBRACE statement_list scanner_mode RBRACE */
#line 762 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(CompoundStmt,stmt,yymsp[-3].minor.yy0);
	stmt->Content = yymsp[-2].minor.yy3;
	yylhsminor.yy128 = stmt;
}
#line 3719 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,69,&yymsp[0].minor);
  yymsp[-3].minor.yy128 = yylhsminor.yy128;
        break;
      case 122: /* state_action ::= LBRACE scanner_mode RBRACE */
{  yy_destructor(yypParser,68,&yymsp[-2].minor);
#line 768 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	yymsp[-2].minor.yy128 = NULL;
}
#line 3729 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,69,&yymsp[0].minor);
}
        break;
      case 123: /* state_action ::= LBRACE error scanner_mode RBRACE */
{  yy_destructor(yypParser,68,&yymsp[-3].minor);
#line 771 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-3].minor.yy128 = NULL; }
#line 3737 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,69,&yymsp[0].minor);
}
        break;
      case 124: /* state_action ::= state_call scanner_mode SEMICOLON */
#line 772 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-2].minor.yy128 = yymsp[-2].minor.yy221; /*X-overwrites-A*/ }
#line 3744 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,49,&yymsp[0].minor);
        break;
      case 125: /* state_call ::= */
#line 774 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[1].minor.yy221 = NULL; }
#line 3750 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 126: /* state_call ::= IDENTIFIER state_call_params */
#line 776 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ExprFuncCall, expr, yymsp[-1].minor.yy0);
	NEW_AST_NODE(ExprID, func, yymsp[-1].minor.yy0);

	func->Operation = PEX_ID;
	func->Identifier = yymsp[-1].minor.yy0.Name();
	expr->Operation = PEX_FuncCall;
	expr->Function = func;
	expr->Parameters = yymsp[0].minor.yy478;
	yylhsminor.yy221 = expr;
}
#line 3765 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[-1].minor.yy221 = yylhsminor.yy221;
        break;
      case 127: /* state_call_params ::= */
#line 788 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[1].minor.yy478 = NULL; }
#line 3771 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 128: /* state_call_params ::= LPAREN func_expr_list RPAREN */
{  yy_destructor(yypParser,45,&yymsp[-2].minor);
#line 789 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-2].minor.yy478 = yymsp[-1].minor.yy478; /*yymsp[-2].minor.yy478-overwrites-yymsp[-1].minor.yy478*/ }
#line 3777 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,63,&yymsp[0].minor);
}
        break;
      case 129: /* default_def ::= DEFAULT LBRACE RBRACE */
{  yy_destructor(yypParser,66,&yymsp[-2].minor);
#line 797 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(Default,stmt,yymsp[-1].minor.yy0);
	stmt->Content = NULL;
	yymsp[-2].minor.yy251 = stmt;
}
#line 3789 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,69,&yymsp[0].minor);
}
        break;
      case 130: /* default_def ::= DEFAULT LBRACE default_statement_list RBRACE */
{  yy_destructor(yypParser,66,&yymsp[-3].minor);
#line 803 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(Default,stmt,yymsp[-2].minor.yy0);
	stmt->Content = yymsp[-1].minor.yy3;
	yymsp[-3].minor.yy251 = stmt;
}
#line 3801 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,69,&yymsp[0].minor);
}
        break;
      case 131: /* default_def ::= DEFAULT LBRACE error RBRACE */
{  yy_destructor(yypParser,66,&yymsp[-3].minor);
#line 809 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(Default,stmt,yymsp[-2].minor.yy0);
	stmt->Content = NULL;
	yymsp[-3].minor.yy251 = stmt;
}
#line 3813 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,69,&yymsp[0].minor);
}
        break;
      case 132: /* default_statement_list ::= default_statement */
      case 330: /*statement_list ::= statement */ yytestcase(yyruleno==330);
#line 816 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	yymsp[0].minor.yy3 = yymsp[0].minor.yy3; /*X-overwrites-A*/
}
#line 3823 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 133: /* default_statement_list ::= default_statement_list default_statement */
      case 331: /*statement_list ::= statement_list statement */ yytestcase(yyruleno==331);
#line 820 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	SAFE_APPEND(yymsp[-1].minor.yy3,yymsp[0].minor.yy3);
}
#line 3831 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 134: /* default_statement ::= SEMICOLON */
      case 312: /*statement ::= SEMICOLON */ yytestcase(yyruleno==312);
{  yy_destructor(yypParser,49,&yymsp[0].minor);
#line 825 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy3 = NULL; }
#line 3838 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 135: /* default_statement ::= error SEMICOLON */
      case 319: /*statement ::= error SEMICOLON */ yytestcase(yyruleno==319);
#line 826 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-1].minor.yy3 = NULL; }
#line 3845 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,49,&yymsp[0].minor);
        break;
      case 136: /* default_statement ::= property_statement */
#line 828 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy3 = yymsp[0].minor.yy283; /*X-overwrites-A*/ }
#line 3851 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 137: /* default_statement ::= flag_statement */
#line 829 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy3 = yymsp[0].minor.yy500; /*X-overwrites-A*/ }
#line 3856 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 138: /* flag_statement ::= ADD dottable_id */
{  yy_destructor(yypParser,34,&yymsp[-1].minor);
#line 834 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(FlagStmt, type, yymsp[0].minor.yy493);
	type->set = true;
	type->name = yymsp[0].minor.yy493;
	yymsp[-1].minor.yy500 = type;
}
#line 3867 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 139: /* flag_statement ::= SUB dottable_id */
{  yy_destructor(yypParser,33,&yymsp[-1].minor);
#line 841 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(FlagStmt, type, yymsp[0].minor.yy493);
	type->set = false;
	type->name = yymsp[0].minor.yy493;
	yymsp[-1].minor.yy500 = type;
}
#line 3879 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 140: /* property_statement ::= dottable_id expr_list SEMICOLON */
#line 851 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(PropertyStmt,stmt,yymsp[-2].minor.yy493);
	stmt->Prop = yymsp[-2].minor.yy493;
	stmt->Values = yymsp[-1].minor.yy282;
	yylhsminor.yy283 = stmt;
}
#line 3890 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,49,&yymsp[0].minor);
  yymsp[-2].minor.yy283 = yylhsminor.yy283;
        break;
      case 141: /* property_statement ::= dottable_id SEMICOLON */
#line 859 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(PropertyStmt,stmt,yymsp[-1].minor.yy493);
	stmt->Prop = yymsp[-1].minor.yy493;
	stmt->Values = nullptr;
	yylhsminor.yy283 = stmt;
}
#line 3902 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,49,&yymsp[0].minor);
  yymsp[-1].minor.yy283 = yylhsminor.yy283;
        break;
      case 142: /* int_type ::= SBYTE */
#line 871 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_SInt8; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 3909 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 143: /* int_type ::= BYTE */
#line 872 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_UInt8; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 3915 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 144: /* int_type ::= SHORT */
#line 873 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_SInt16; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 3921 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 145: /* int_type ::= USHORT */
#line 874 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_UInt16; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 3927 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 146: /* int_type ::= INT */
#line 875 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_SInt32; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 3933 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 147: /* int_type ::= UINT */
#line 876 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_UInt32; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 3939 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 148: /* type_name1 ::= BOOL */
#line 878 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Bool; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 3945 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 149: /* type_name1 ::= FLOAT */
#line 880 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_FloatAuto; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 3951 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 150: /* type_name1 ::= DOUBLE */
#line 881 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Float64; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 3957 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 151: /* type_name1 ::= VECTOR2 */
#line 883 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Vector2; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 3963 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 152: /* type_name1 ::= VECTOR3 */
#line 884 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Vector3; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 3969 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 153: /* type_name1 ::= VECTOR4 */
#line 885 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Vector4; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 3975 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 154: /* type_name1 ::= NAME */
#line 886 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Name; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 3981 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 155: /* type_name1 ::= SOUND */
#line 887 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Sound; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 3987 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 156: /* type_name1 ::= STATE */
#line 888 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_State; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 3993 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 157: /* type_name1 ::= COLOR */
#line 889 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Color; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 3999 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 158: /* type_name1 ::= LET */
#line 890 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Let; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 4005 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 159: /* type_name ::= type_name1 */
#line 893 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(BasicType, type, yymsp[0].minor.yy0);
	type->Type = (EZCCBuiltinType)yymsp[0].minor.yy0.Int;
	type->UserType = NULL;
	type->isconst = false;
	yylhsminor.yy430 = type;
}
#line 4017 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy430 = yylhsminor.yy430;
        break;
      case 160: /* type_name ::= IDENTIFIER */
#line 901 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(BasicType, type, yymsp[0].minor.yy0);
	NEW_AST_NODE(Identifier, id, yymsp[0].minor.yy0);
	type->Type = ZCC_UserType;
	type->UserType = id;
	type->isconst = false;
	id->Id = yymsp[0].minor.yy0.Name();
	yylhsminor.yy430 = type;
}
#line 4031 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy430 = yylhsminor.yy430;
        break;
      case 161: /* type_name ::= ATSIGN IDENTIFIER */
{  yy_destructor(yypParser,110,&yymsp[-1].minor);
#line 912 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(BasicType, type, yymsp[0].minor.yy0);
	NEW_AST_NODE(Identifier, id, yymsp[0].minor.yy0);
	type->Type = ZCC_NativeType;
	type->UserType = id;
	type->isconst = false;
	id->Id = yymsp[0].minor.yy0.Name();
	yymsp[-1].minor.yy430 = type;
}
#line 4046 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 162: /* type_name ::= READONLY LT IDENTIFIER GT */
{  yy_destructor(yypParser,111,&yymsp[-3].minor);
#line 923 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(BasicType, type, yymsp[-1].minor.yy0);
	NEW_AST_NODE(Identifier, id, yymsp[-1].minor.yy0);
	type->Type = ZCC_UserType;
	type->UserType = id;
	type->isconst = true;
	id->Id = yymsp[-1].minor.yy0.Name();
	yymsp[-3].minor.yy430 = type;
}
#line 4061 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,20,&yymsp[-2].minor);
  yy_destructor(yypParser,21,&yymsp[0].minor);
}
        break;
      case 163: /* type_name ::= READONLY LT ATSIGN IDENTIFIER GT */
{  yy_destructor(yypParser,111,&yymsp[-4].minor);
#line 934 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(BasicType, type, yymsp[-1].minor.yy0);
	NEW_AST_NODE(Identifier, id, yymsp[-1].minor.yy0);
	type->Type = ZCC_NativeType;
	type->UserType = id;
	type->isconst = true;
	id->Id = yymsp[-1].minor.yy0.Name();
	yymsp[-4].minor.yy430 = type;
}
#line 4078 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,20,&yymsp[-3].minor);
  yy_destructor(yypParser,110,&yymsp[-2].minor);
  yy_destructor(yypParser,21,&yymsp[0].minor);
}
        break;
      case 164: /* type_name ::= DOT dottable_id */
{  yy_destructor(yypParser,44,&yymsp[-1].minor);
#line 945 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(BasicType, type, yymsp[0].minor.yy493);
	type->Type = ZCC_UserType;
	type->UserType = yymsp[0].minor.yy493;
	type->isconst = false;
	yymsp[-1].minor.yy430 = type;
}
#line 4094 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 165: /* aggregate_type ::= MAP LT type_or_array COMMA type_or_array GT */
#line 969 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(MapType,map,yymsp[-5].minor.yy0);
	map->KeyType = yymsp[-3].minor.yy196;
	map->ValueType = yymsp[-1].minor.yy196;
	yylhsminor.yy196 = map;
}
#line 4105 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,20,&yymsp[-4].minor);
  yy_destructor(yypParser,50,&yymsp[-2].minor);
  yy_destructor(yypParser,21,&yymsp[0].minor);
  yymsp[-5].minor.yy196 = yylhsminor.yy196;
        break;
      case 166: /* aggregate_type ::= MAPITERATOR LT type_or_array COMMA type_or_array GT */
#line 977 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(MapIteratorType,map_it,yymsp[-5].minor.yy0);
	map_it->KeyType = yymsp[-3].minor.yy196;
	map_it->ValueType = yymsp[-1].minor.yy196;
	yylhsminor.yy196 = map_it;
}
#line 4119 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,20,&yymsp[-4].minor);
  yy_destructor(yypParser,50,&yymsp[-2].minor);
  yy_destructor(yypParser,21,&yymsp[0].minor);
  yymsp[-5].minor.yy196 = yylhsminor.yy196;
        break;
      case 167: /* aggregate_type ::= ARRAY LT type_or_array GT */
#line 985 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(DynArrayType,arr,yymsp[-3].minor.yy0);
	arr->ElementType = yymsp[-1].minor.yy196;
	yylhsminor.yy196 = arr;
}
#line 4132 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,20,&yymsp[-2].minor);
  yy_destructor(yypParser,21,&yymsp[0].minor);
  yymsp[-3].minor.yy196 = yylhsminor.yy196;
        break;
      case 168: /* aggregate_type ::= func_ptr_type */
#line 991 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy196 = yymsp[0].minor.yy418; /*X-overwrites-A*/ }
#line 4140 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 169: /* fn_ptr_flag ::= UI */
{  yy_destructor(yypParser,58,&yymsp[0].minor);
#line 999 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy0.Int = ZCC_UIFlag; }
#line 4146 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 170: /* fn_ptr_flag ::= PLAY */
{  yy_destructor(yypParser,59,&yymsp[0].minor);
#line 1000 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy0.Int = ZCC_Play; }
#line 4153 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 171: /* fn_ptr_flag ::= CLEARSCOPE */
{  yy_destructor(yypParser,76,&yymsp[0].minor);
#line 1001 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy0.Int = ZCC_ClearScope; }
#line 4160 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 172: /* func_ptr_type ::= FNTYPE LT fn_ptr_flag type_list_or_void LPAREN func_ptr_params RPAREN GT */
#line 1006 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(FuncPtrType,fn_ptr,yymsp[-7].minor.yy0);
	fn_ptr->RetType = yymsp[-4].minor.yy196;
	fn_ptr->Params = yymsp[-2].minor.yy513;
	fn_ptr->Scope = yymsp[-5].minor.yy0.Int;
	yylhsminor.yy418 = fn_ptr;
}
#line 4172 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,20,&yymsp[-6].minor);
  yy_destructor(yypParser,45,&yymsp[-3].minor);
  yy_destructor(yypParser,63,&yymsp[-1].minor);
  yy_destructor(yypParser,21,&yymsp[0].minor);
  yymsp[-7].minor.yy418 = yylhsminor.yy418;
        break;
      case 173: /* func_ptr_type ::= FNTYPE LT VOID GT */
#line 1015 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(FuncPtrType,fn_ptr,yymsp[-3].minor.yy0);
	fn_ptr->RetType = nullptr;
	fn_ptr->Params = nullptr;
	fn_ptr->Scope = -1;
	yylhsminor.yy418 = fn_ptr;
}
#line 4188 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,20,&yymsp[-2].minor);
  yy_destructor(yypParser,116,&yymsp[-1].minor);
  yy_destructor(yypParser,21,&yymsp[0].minor);
  yymsp[-3].minor.yy418 = yylhsminor.yy418;
        break;
      case 174: /* func_ptr_params ::= */
#line 1023 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[1].minor.yy513 = NULL; }
#line 4197 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 175: /* func_ptr_params ::= VOID */
{  yy_destructor(yypParser,116,&yymsp[0].minor);
#line 1024 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy513 = NULL; }
#line 4203 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 176: /* func_ptr_param_list ::= func_ptr_param_list COMMA func_ptr_param */
#line 1038 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-2].minor.yy513 = yymsp[-2].minor.yy513; /*X-overwrites-A*/ AppendTreeNodeSibling(yymsp[-2].minor.yy513, yymsp[0].minor.yy513); }
#line 4209 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,50,&yymsp[-1].minor);
        break;
      case 177: /* func_ptr_param ::= func_param_flags type */
#line 1041 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(FuncPtrParamDecl,parm,yymsp[-1].minor.yy0.SourceLoc ? yymsp[-1].minor.yy0.SourceLoc : yymsp[0].minor.yy196->SourceLoc);
	parm->Type = yymsp[0].minor.yy196;
	parm->Flags = yymsp[-1].minor.yy0.Int;
	yylhsminor.yy513 = parm;
}
#line 4220 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[-1].minor.yy513 = yylhsminor.yy513;
        break;
      case 178: /* func_ptr_param ::= func_param_flags type AND */
#line 1049 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(FuncPtrParamDecl,parm,yymsp[-2].minor.yy0.SourceLoc ? yymsp[-2].minor.yy0.SourceLoc : yymsp[-1].minor.yy196->SourceLoc);
	parm->Type = yymsp[-1].minor.yy196;
	parm->Flags = yymsp[-2].minor.yy0.Int | ZCC_Out;
	yylhsminor.yy513 = parm;
}
#line 4231 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,29,&yymsp[0].minor);
  yymsp[-2].minor.yy513 = yylhsminor.yy513;
        break;
      case 179: /* aggregate_type ::= CLASS class_restrictor */
#line 1057 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ClassType,cls,yymsp[-1].minor.yy0);
	cls->Restriction = yymsp[0].minor.yy493;
	yylhsminor.yy196 = cls;
}
#line 4242 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[-1].minor.yy196 = yylhsminor.yy196;
        break;
      case 181: /* class_restrictor ::= LT dottable_id GT */
{  yy_destructor(yypParser,20,&yymsp[-2].minor);
#line 1063 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-2].minor.yy493 = yymsp[-1].minor.yy493; /*yymsp[-2].minor.yy493-overwrites-yymsp[-1].minor.yy493*/ }
#line 4249 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,21,&yymsp[0].minor);
}
        break;
      case 182: /* type ::= type_name */
#line 1065 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy196 = yymsp[0].minor.yy430; /*X-overwrites-A*/ yymsp[0].minor.yy196->ArraySize = NULL; }
#line 4256 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 183: /* type ::= aggregate_type */
#line 1066 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy196 = yymsp[0].minor.yy196; /*X-overwrites-A*/ yymsp[0].minor.yy196->ArraySize = NULL; }
#line 4261 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 184: /* type_or_array ::= type array_size */
#line 1069 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-1].minor.yy196 = yymsp[-1].minor.yy196; /*X-overwrites-A*/ yymsp[-1].minor.yy196->ArraySize = yymsp[0].minor.yy282; }
#line 4266 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 185: /* type_list ::= type_list COMMA type_or_array */
#line 1072 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-2].minor.yy196 = yymsp[-2].minor.yy196; /*X-overwrites-A*/ AppendTreeNodeSibling(yymsp[-2].minor.yy196, yymsp[0].minor.yy196); }
#line 4271 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,50,&yymsp[-1].minor);
        break;
      case 186: /* type_list_or_void ::= VOID */
{  yy_destructor(yypParser,116,&yymsp[0].minor);
#line 1074 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy196 = NULL; }
#line 4278 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 187: /* array_size_expr ::= LBRACKET opt_expr RBRACKET */
#line 1078 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	if (yymsp[-1].minor.yy282 == NULL)
	{
		NEW_AST_NODE(Expression,nil,yymsp[-2].minor.yy0.SourceLoc);
		nil->Operation = PEX_Nil;
		nil->Type = NULL;
		yylhsminor.yy282 = nil;
	}
	else
	{
		yylhsminor.yy282 = yymsp[-1].minor.yy282;
	}
}
#line 4296 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,122,&yymsp[0].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 188: /* array_size ::= array_size array_size_expr */
#line 1093 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	AppendTreeNodeSibling(yymsp[-1].minor.yy282, yymsp[0].minor.yy282);
	yymsp[-1].minor.yy282 = yymsp[-1].minor.yy282; /*X-overwrites-A*/
}
#line 4306 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 189: /* declarator ::= decl_flags type_list_or_void variables_or_function */
#line 1102 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	if (yymsp[0].minor.yy364.FuncName == NAME_None && yymsp[0].minor.yy364.VarNames == NULL)
	{ // An error. yymsp[-2].minor.yy415 message was already printed.
		yylhsminor.yy327 = NULL;
	}
	else if (yymsp[0].minor.yy364.FuncName != NAME_None)
	{ // yymsp[-2].minor.yy415 function
		NEW_AST_NODE(FuncDeclarator, decl, yymsp[-2].minor.yy415 == nullptr? yymsp[0].minor.yy364.SourceLoc : yymsp[-2].minor.yy415->SourceLoc);
		decl->Type = yymsp[-1].minor.yy196;
		decl->Params = yymsp[0].minor.yy364.FuncParams;
		decl->Name = yymsp[0].minor.yy364.FuncName;
		decl->UseFlags = yymsp[-2].minor.yy415 == nullptr? nullptr : yymsp[-2].minor.yy415->Id;
		decl->Flags = (yymsp[-2].minor.yy415 == nullptr? 0 : yymsp[-2].minor.yy415->Flags) | yymsp[0].minor.yy364.FuncFlags;
		decl->DeprecationMessage = yymsp[-2].minor.yy415 == nullptr ? nullptr : yymsp[-2].minor.yy415->DeprecationMessage;
		if (yymsp[-2].minor.yy415 == nullptr) decl->Version = {0,0,0};
		else decl->Version = yymsp[-2].minor.yy415->Version;

		decl->Body = yymsp[0].minor.yy364.FuncBody;
		yylhsminor.yy327 = decl;
	}
	else if (yymsp[-1].minor.yy196 != NULL && yymsp[-1].minor.yy196->SiblingNext == yymsp[-1].minor.yy196)
	{ // yymsp[-2].minor.yy415 variable
		NEW_AST_NODE(VarDeclarator, decl, yymsp[-2].minor.yy415 == nullptr? yymsp[-1].minor.yy196->SourceLoc : yymsp[-2].minor.yy415->SourceLoc);
		decl->Type = yymsp[-1].minor.yy196;
		decl->Names = yymsp[0].minor.yy364.VarNames;
		if (yymsp[-2].minor.yy415 == nullptr)
		{
			decl->Flags = 0;
			decl->Version = {0,0,0};
			decl->DeprecationMessage = nullptr;
		}
		else
		{
			decl->Flags = yymsp[-2].minor.yy415->Flags;
			decl->Version = yymsp[-2].minor.yy415->Version;
			decl->DeprecationMessage = yymsp[-2].minor.yy415->DeprecationMessage;
		}
		yylhsminor.yy327 = decl;
	}
	else
	{ // An invalid
		if (yymsp[-1].minor.yy196 == NULL)
		{
			stat->sc->ScriptMessage("Variables may not be of type void.\n");
		}
		else
		{
			stat->sc->ScriptMessage("Variables may be of only one type.\n");
		}
		yylhsminor.yy327 = NULL;
	}
}
#line 4362 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[-2].minor.yy327 = yylhsminor.yy327;
        break;
      case 190: /* variables_or_function ::= IDENTIFIER LPAREN func_params RPAREN func_const opt_func_body */
#line 1157 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	VarOrFun fun;

	fun.VarNames = NULL;
	fun.FuncParams = yymsp[-3].minor.yy439;
	fun.FuncFlags = yymsp[-1].minor.yy0.Int;
	fun.FuncName = yymsp[-5].minor.yy0.Name();
	fun.FuncBody = yymsp[0].minor.yy475;
	fun.SourceLoc = yymsp[-5].minor.yy0.SourceLoc;
	yylhsminor.yy364 = fun;
}
#line 4378 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,45,&yymsp[-4].minor);
  yy_destructor(yypParser,63,&yymsp[-2].minor);
  yymsp[-5].minor.yy364 = yylhsminor.yy364;
        break;
      case 191: /* variables_or_function ::= variable_list SEMICOLON */
#line 1169 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	VarOrFun var;

	var.VarNames = yymsp[-1].minor.yy492;
	var.FuncParams = NULL;
	var.FuncFlags = 0;
	var.FuncName = NAME_None;
	var.FuncBody = NULL;
	var.SourceLoc = yymsp[-1].minor.yy492->SourceLoc;
	yylhsminor.yy364 = var;
}
#line 4396 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,49,&yymsp[0].minor);
  yymsp[-1].minor.yy364 = yylhsminor.yy364;
        break;
      case 192: /* variables_or_function ::= error SEMICOLON */
#line 1181 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	VarOrFun bad;
	bad.VarNames = NULL;
	bad.FuncParams = NULL;
	bad.FuncFlags = 0;
	bad.FuncName = NAME_None;
	bad.FuncBody = NULL;
	bad.SourceLoc = yymsp[0].minor.yy0.SourceLoc;
	yymsp[-1].minor.yy364 = bad;
}
#line 4412 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 193: /* variable_name ::= IDENTIFIER */
#line 1198 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(VarName,var,yymsp[0].minor.yy0);
	var->Name = ENamedName(yymsp[0].minor.yy0.Int);
	var->ArraySize = NULL;
	yylhsminor.yy492 = var;
}
#line 4422 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy492 = yylhsminor.yy492;
        break;
      case 194: /* variable_name ::= IDENTIFIER array_size */
#line 1205 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(VarName,var,yymsp[-1].minor.yy0);
	var->Name = ENamedName(yymsp[-1].minor.yy0.Int);
	var->ArraySize = yymsp[0].minor.yy282;
	yylhsminor.yy492 = var;
}
#line 4433 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[-1].minor.yy492 = yylhsminor.yy492;
        break;
      case 195: /* variable_list ::= variable_list COMMA variable_name */
#line 1214 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	AppendTreeNodeSibling(yymsp[-2].minor.yy492, yymsp[0].minor.yy492);
	yymsp[-2].minor.yy492 = yymsp[-2].minor.yy492; /*X-overwrites-A*/
}
#line 4442 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,50,&yymsp[-1].minor);
        break;
      case 196: /* decl_flags ::= */
#line 1220 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[1].minor.yy415 = NULL; }
#line 4448 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 197: /* decl_flags ::= decl_flags decl_flag */
#line 1222 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	if (yymsp[-1].minor.yy415 == nullptr)
	{
		NEW_AST_NODE(DeclFlags,nil_f,yymsp[0].minor.yy0);
		yylhsminor.yy415 = nil_f;
		yylhsminor.yy415->Id = nullptr;
		yylhsminor.yy415->Flags = yymsp[0].minor.yy0.Int;
		yylhsminor.yy415->Version = { 0, 0 };
		yylhsminor.yy415->DeprecationMessage = nullptr;
	}
	else
	{
		yylhsminor.yy415 = yymsp[-1].minor.yy415;
		yylhsminor.yy415->Flags |= yymsp[0].minor.yy0.Int;
	}
}
#line 4468 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[-1].minor.yy415 = yylhsminor.yy415;
        break;
      case 198: /* decl_flags ::= decl_flags ACTION states_opts */
#line 1241 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	if (yymsp[-2].minor.yy415 == nullptr)
	{
		NEW_AST_NODE(DeclFlags,nil_f,yymsp[-1].minor.yy0.SourceLoc);
		yylhsminor.yy415 = nil_f;
		yylhsminor.yy415->Flags = ZCC_Action;
		yylhsminor.yy415->Id = nullptr;
		yylhsminor.yy415->Version = { 0, 0 };
		yylhsminor.yy415->DeprecationMessage = nullptr;
	}
	else
	{
		yylhsminor.yy415 = yymsp[-2].minor.yy415;
		yylhsminor.yy415->Flags |= ZCC_Action;
	}
	yylhsminor.yy415->Id = yymsp[0].minor.yy493;
}
#line 4490 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[-2].minor.yy415 = yylhsminor.yy415;
        break;
      case 199: /* opt_deprecation_message ::= */
#line 1259 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[1].minor.yy0.String = nullptr; yymsp[1].minor.yy0.SourceLoc = stat->sc->GetMessageLine(); }
#line 4496 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 200: /* opt_deprecation_message ::= COMMA STRCONST */
{  yy_destructor(yypParser,50,&yymsp[-1].minor);
#line 1260 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-1].minor.yy0 = yymsp[0].minor.yy0; }
#line 4502 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 201: /* decl_flags ::= decl_flags DEPRECATED LPAREN STRCONST opt_deprecation_message RPAREN */
#line 1263 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	if (yymsp[-5].minor.yy415 == nullptr)
	{
		NEW_AST_NODE(DeclFlags,nil_f,yymsp[-4].minor.yy0.SourceLoc);
		yylhsminor.yy415 = nil_f;
		yylhsminor.yy415->Flags = ZCC_Deprecated;
		yylhsminor.yy415->Id = nullptr;
		yylhsminor.yy415->Version = { 0, 0 };
	}
	else
	{
		yylhsminor.yy415 = yymsp[-5].minor.yy415;
		yylhsminor.yy415->Flags |= ZCC_Deprecated;
	}
	yylhsminor.yy415->Version = yymsp[-2].minor.yy0.String->GetChars();
	yylhsminor.yy415->DeprecationMessage = yymsp[-1].minor.yy0.String;
}
#line 4524 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,45,&yymsp[-3].minor);
  yy_destructor(yypParser,63,&yymsp[0].minor);
  yymsp[-5].minor.yy415 = yylhsminor.yy415;
        break;
      case 202: /* decl_flags ::= decl_flags VERSION LPAREN STRCONST RPAREN */
#line 1282 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ 
	if (yymsp[-4].minor.yy415 == nullptr)
	{
		NEW_AST_NODE(DeclFlags,nil_f,yymsp[-3].minor.yy0.SourceLoc);
		yylhsminor.yy415 = nil_f;
		yylhsminor.yy415->Flags = ZCC_Version;
		yylhsminor.yy415->Id = nullptr;
		yylhsminor.yy415->Version = { 0, 0 };
		yylhsminor.yy415->DeprecationMessage = nullptr;
	}
	else
	{
		yylhsminor.yy415 = yymsp[-4].minor.yy415;
		yylhsminor.yy415->Flags |= ZCC_Version;
	}
	yylhsminor.yy415->Version = yymsp[-1].minor.yy0.String->GetChars();
}
#line 4548 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,45,&yymsp[-2].minor);
  yy_destructor(yypParser,63,&yymsp[0].minor);
  yymsp[-4].minor.yy415 = yylhsminor.yy415;
        break;
      case 203: /* decl_flag ::= NATIVE */
#line 1300 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Native; yylhsminor.yy0.SourceLoc =  yymsp[0].minor.yy0.SourceLoc; }
#line 4556 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 204: /* decl_flag ::= STATIC */
#line 1301 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Static; yylhsminor.yy0.SourceLoc =  yymsp[0].minor.yy0.SourceLoc; }
#line 4562 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 205: /* decl_flag ::= PRIVATE */
#line 1302 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Private; yylhsminor.yy0.SourceLoc =  yymsp[0].minor.yy0.SourceLoc; }
#line 4568 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 206: /* decl_flag ::= PROTECTED */
#line 1303 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Protected; yylhsminor.yy0.SourceLoc =  yymsp[0].minor.yy0.SourceLoc; }
#line 4574 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 207: /* decl_flag ::= LATENT */
#line 1304 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Latent; yylhsminor.yy0.SourceLoc =  yymsp[0].minor.yy0.SourceLoc; }
#line 4580 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 208: /* decl_flag ::= FINAL */
#line 1305 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Final; yylhsminor.yy0.SourceLoc =  yymsp[0].minor.yy0.SourceLoc; }
#line 4586 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 209: /* decl_flag ::= META */
#line 1306 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Meta; yylhsminor.yy0.SourceLoc =  yymsp[0].minor.yy0.SourceLoc; }
#line 4592 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 210: /* decl_flag ::= TRANSIENT */
#line 1307 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Transient; yylhsminor.yy0.SourceLoc =  yymsp[0].minor.yy0.SourceLoc; }
#line 4598 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 211: /* decl_flag ::= READONLY */
#line 1308 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_ReadOnly; yylhsminor.yy0.SourceLoc =  yymsp[0].minor.yy0.SourceLoc; }
#line 4604 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 212: /* decl_flag ::= INTERNAL */
#line 1309 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Internal; yylhsminor.yy0.SourceLoc =  yymsp[0].minor.yy0.SourceLoc; }
#line 4610 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 213: /* decl_flag ::= VIRTUAL */
#line 1310 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Virtual; yylhsminor.yy0.SourceLoc =  yymsp[0].minor.yy0.SourceLoc; }
#line 4616 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 214: /* decl_flag ::= OVERRIDE */
#line 1311 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Override; yylhsminor.yy0.SourceLoc =  yymsp[0].minor.yy0.SourceLoc; }
#line 4622 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 215: /* decl_flag ::= ABSTRACT */
#line 1312 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Abstract; yylhsminor.yy0.SourceLoc =  yymsp[0].minor.yy0.SourceLoc; }
#line 4628 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 216: /* decl_flag ::= VARARG */
#line 1313 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_VarArg; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 4634 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 217: /* decl_flag ::= UI */
#line 1314 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_UIFlag; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 4640 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 218: /* decl_flag ::= PLAY */
#line 1315 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_Play; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 4646 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 219: /* decl_flag ::= CLEARSCOPE */
#line 1316 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_ClearScope; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 4652 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 220: /* decl_flag ::= VIRTUALSCOPE */
#line 1317 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_VirtualScope; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 4658 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 221: /* func_const ::= */
#line 1319 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[1].minor.yy0.Int = 0; yymsp[1].minor.yy0.SourceLoc = stat->sc->GetMessageLine(); }
#line 4664 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 222: /* func_const ::= CONST */
#line 1320 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = ZCC_FuncConst; yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc; }
#line 4669 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 223: /* opt_func_body ::= SEMICOLON */
{  yy_destructor(yypParser,49,&yymsp[0].minor);
#line 1322 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy475 = NULL; }
#line 4676 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 224: /* func_params ::= */
#line 1329 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[1].minor.yy439 = NULL; }
#line 4682 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 225: /* func_params ::= VOID */
{  yy_destructor(yypParser,116,&yymsp[0].minor);
#line 1330 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy439 = NULL; }
#line 4688 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 226: /* func_params ::= func_param_list COMMA ELLIPSIS */
#line 1334 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(FuncParamDecl,parm,stat->sc->GetMessageLine());
	parm->Type = nullptr;
	parm->Name = NAME_None;
	parm->Flags = 0;
	parm->Default = nullptr;
	yymsp[-2].minor.yy439 = yymsp[-2].minor.yy439; /*X-overwrites-A*/ 
	AppendTreeNodeSibling(yymsp[-2].minor.yy439, parm);
}
#line 4702 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,50,&yymsp[-1].minor);
  yy_destructor(yypParser,134,&yymsp[0].minor);
        break;
      case 227: /* func_param_list ::= func_param_list COMMA func_param */
#line 1345 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-2].minor.yy439 = yymsp[-2].minor.yy439; /*X-overwrites-A*/ AppendTreeNodeSibling(yymsp[-2].minor.yy439, yymsp[0].minor.yy439); }
#line 4709 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,50,&yymsp[-1].minor);
        break;
      case 228: /* func_param ::= func_param_flags type IDENTIFIER */
#line 1348 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(FuncParamDecl,parm,yymsp[-2].minor.yy0.SourceLoc ? yymsp[-2].minor.yy0.SourceLoc : yymsp[-1].minor.yy196->SourceLoc);
	parm->Type = yymsp[-1].minor.yy196;
	parm->Name = yymsp[0].minor.yy0.Name();
	parm->Flags = yymsp[-2].minor.yy0.Int;
	parm->Default = nullptr;
	yylhsminor.yy439 = parm;
}
#line 4722 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[-2].minor.yy439 = yylhsminor.yy439;
        break;
      case 229: /* func_param ::= func_param_flags type IDENTIFIER EQ expr */
#line 1358 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(FuncParamDecl,parm,yymsp[-4].minor.yy0.SourceLoc ? yymsp[-4].minor.yy0.SourceLoc : yymsp[-3].minor.yy196->SourceLoc);
	parm->Type = yymsp[-3].minor.yy196;
	parm->Name = yymsp[-2].minor.yy0.Name();
	parm->Flags = yymsp[-4].minor.yy0.Int;
	parm->Default = yymsp[0].minor.yy282;
	yylhsminor.yy439 = parm;
}
#line 4735 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,1,&yymsp[-1].minor);
  yymsp[-4].minor.yy439 = yylhsminor.yy439;
        break;
      case 230: /* func_param ::= func_param_flags type AND IDENTIFIER */
#line 1368 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(FuncParamDecl,parm,yymsp[-3].minor.yy0.SourceLoc ? yymsp[-3].minor.yy0.SourceLoc : yymsp[-2].minor.yy196->SourceLoc);
	parm->Type = yymsp[-2].minor.yy196;
	parm->Name = yymsp[0].minor.yy0.Name();
	parm->Flags = yymsp[-3].minor.yy0.Int | ZCC_Out;
	parm->Default = nullptr;
	yylhsminor.yy439 = parm;
}
#line 4749 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,29,&yymsp[-1].minor);
  yymsp[-3].minor.yy439 = yylhsminor.yy439;
        break;
      case 231: /* func_param ::= func_param_flags type AND IDENTIFIER EQ expr */
#line 1378 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(FuncParamDecl,parm,yymsp[-5].minor.yy0.SourceLoc ? yymsp[-5].minor.yy0.SourceLoc : yymsp[-4].minor.yy196->SourceLoc);
	parm->Type = yymsp[-4].minor.yy196;
	parm->Name = yymsp[-2].minor.yy0.Name();
	parm->Flags = yymsp[-5].minor.yy0.Int | ZCC_Out;
	parm->Default = yymsp[0].minor.yy282;
	yylhsminor.yy439 = parm;
}
#line 4763 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,29,&yymsp[-3].minor);
  yy_destructor(yypParser,1,&yymsp[-1].minor);
  yymsp[-5].minor.yy439 = yylhsminor.yy439;
        break;
      case 232: /* func_param_flags ::= */
#line 1387 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[1].minor.yy0.Int = 0; yymsp[1].minor.yy0.SourceLoc = 0; }
#line 4771 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 233: /* func_param_flags ::= func_param_flags IN */
#line 1388 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = yymsp[-1].minor.yy0.Int | ZCC_In; yylhsminor.yy0.SourceLoc =  yymsp[0].minor.yy0.SourceLoc; }
#line 4776 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[-1].minor.yy0 = yylhsminor.yy0;
        break;
      case 234: /* func_param_flags ::= func_param_flags OUT */
#line 1389 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = yymsp[-1].minor.yy0.Int | ZCC_Out; yylhsminor.yy0.SourceLoc =  yymsp[0].minor.yy0.SourceLoc; }
#line 4782 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[-1].minor.yy0 = yylhsminor.yy0;
        break;
      case 235: /* func_param_flags ::= func_param_flags OPTIONAL */
#line 1390 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yylhsminor.yy0.Int = yymsp[-1].minor.yy0.Int | ZCC_Optional; yylhsminor.yy0.SourceLoc =  yymsp[0].minor.yy0.SourceLoc; }
#line 4788 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[-1].minor.yy0 = yylhsminor.yy0;
        break;
      case 236: /* primary ::= IDENTIFIER */
#line 1406 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ExprID, expr, yymsp[0].minor.yy0);
	expr->Operation = PEX_ID;
	expr->Identifier = yymsp[0].minor.yy0.Name();
	expr->Type = NULL;
	yylhsminor.yy282 = expr;
}
#line 4800 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy282 = yylhsminor.yy282;
        break;
      case 237: /* primary ::= SUPER */
#line 1414 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(Expression, expr, yymsp[0].minor.yy0);
	expr->Operation = PEX_Super;
	expr->Type = NULL;
	yylhsminor.yy282 = expr;
}
#line 4811 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy282 = yylhsminor.yy282;
        break;
      case 238: /* primary ::= constant */
#line 1420 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy282 = yymsp[0].minor.yy423; /*X-overwrites-A*/ }
#line 4817 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 239: /* primary ::= LPAREN expr COMMA expr COMMA expr COMMA expr RPAREN */
{  yy_destructor(yypParser,45,&yymsp[-8].minor);
#line 1422 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(VectorValue, expr, yymsp[-7].minor.yy282);
	expr->Operation = PEX_Vector;
	expr->Type = TypeVector4;
	expr->X = yymsp[-7].minor.yy282;
	expr->Y = yymsp[-5].minor.yy282;
	expr->Z = yymsp[-3].minor.yy282;
	expr->W = yymsp[-1].minor.yy282;
	yymsp[-8].minor.yy282 = expr;
}
#line 4832 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,50,&yymsp[-6].minor);
  yy_destructor(yypParser,50,&yymsp[-4].minor);
  yy_destructor(yypParser,50,&yymsp[-2].minor);
  yy_destructor(yypParser,63,&yymsp[0].minor);
}
        break;
      case 240: /* primary ::= LPAREN expr COMMA expr COMMA expr RPAREN */
{  yy_destructor(yypParser,45,&yymsp[-6].minor);
#line 1433 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(VectorValue, expr, yymsp[-5].minor.yy282);
	expr->Operation = PEX_Vector;
	expr->Type = TypeVector3;
	expr->X = yymsp[-5].minor.yy282;
	expr->Y = yymsp[-3].minor.yy282;
	expr->Z = yymsp[-1].minor.yy282;
	expr->W = nullptr;
	yymsp[-6].minor.yy282 = expr;
}
#line 4852 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,50,&yymsp[-4].minor);
  yy_destructor(yypParser,50,&yymsp[-2].minor);
  yy_destructor(yypParser,63,&yymsp[0].minor);
}
        break;
      case 241: /* primary ::= LPAREN expr COMMA expr RPAREN */
{  yy_destructor(yypParser,45,&yymsp[-4].minor);
#line 1444 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(VectorValue, expr, yymsp[-3].minor.yy282);
	expr->Operation = PEX_Vector;
	expr->Type = TypeVector2;
	expr->X = yymsp[-3].minor.yy282;
	expr->Y = yymsp[-1].minor.yy282;
	expr->Z = nullptr;
	expr->W = nullptr;
	yymsp[-4].minor.yy282 = expr;
}
#line 4871 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,50,&yymsp[-2].minor);
  yy_destructor(yypParser,63,&yymsp[0].minor);
}
        break;
      case 242: /* primary ::= LPAREN expr RPAREN */
{  yy_destructor(yypParser,45,&yymsp[-2].minor);
#line 1455 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	yymsp[-2].minor.yy282 = yymsp[-1].minor.yy282; /*yymsp[-2].minor.yy282-overwrites-yymsp[-1].minor.yy282*/
}
#line 4882 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,63,&yymsp[0].minor);
}
        break;
      case 243: /* primary ::= primary LPAREN func_expr_list RPAREN */
#line 1460 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ExprFuncCall, expr, yymsp[-3].minor.yy282);
	expr->Operation = PEX_FuncCall;
	expr->Type = NULL;
	expr->Function = yymsp[-3].minor.yy282;
	expr->Parameters = yymsp[-1].minor.yy478;
	yylhsminor.yy282 = expr;
}
#line 4896 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,45,&yymsp[-2].minor);
  yy_destructor(yypParser,63,&yymsp[0].minor);
  yymsp[-3].minor.yy282 = yylhsminor.yy282;
        break;
      case 244: /* primary ::= LPAREN CLASS LT IDENTIFIER GT RPAREN LPAREN func_expr_list RPAREN */
{  yy_destructor(yypParser,45,&yymsp[-8].minor);
#line 1469 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ClassCast, expr, yymsp[-5].minor.yy0);
	expr->Operation = PEX_ClassCast;
	expr->ClassName = ENamedName(yymsp[-5].minor.yy0.Int);
	expr->Parameters = yymsp[-1].minor.yy478;
	yymsp[-8].minor.yy282 = expr;
}
#line 4911 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,53,&yymsp[-7].minor);
  yy_destructor(yypParser,20,&yymsp[-6].minor);
  yy_destructor(yypParser,21,&yymsp[-4].minor);
  yy_destructor(yypParser,63,&yymsp[-3].minor);
  yy_destructor(yypParser,45,&yymsp[-2].minor);
  yy_destructor(yypParser,63,&yymsp[0].minor);
}
        break;
      case 245: /* primary ::= LPAREN func_ptr_type RPAREN LPAREN expr RPAREN */
{  yy_destructor(yypParser,45,&yymsp[-5].minor);
#line 1478 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(FunctionPtrCast, expr, yymsp[-4].minor.yy418);
	expr->Operation = PEX_FunctionPtrCast;
	yymsp[-4].minor.yy418->ArraySize = NULL;
	expr->PtrType = yymsp[-4].minor.yy418;
	expr->Expr = yymsp[-1].minor.yy282;
	yymsp[-5].minor.yy282 = expr;
}
#line 4931 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,63,&yymsp[-3].minor);
  yy_destructor(yypParser,45,&yymsp[-2].minor);
  yy_destructor(yypParser,63,&yymsp[0].minor);
}
        break;
      case 246: /* primary ::= primary LBRACKET expr RBRACKET */
#line 1488 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ExprBinary, expr, yymsp[-1].minor.yy282);
	expr->Operation = PEX_ArrayAccess;
	expr->Type = NULL;
	expr->Left = yymsp[-3].minor.yy282;
	expr->Right = yymsp[-1].minor.yy282;
	yylhsminor.yy282 = expr;
}
#line 4947 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,46,&yymsp[-2].minor);
  yy_destructor(yypParser,122,&yymsp[0].minor);
  yymsp[-3].minor.yy282 = yylhsminor.yy282;
        break;
      case 247: /* primary ::= primary DOT IDENTIFIER */
#line 1498 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ExprMemberAccess, expr, yymsp[0].minor.yy0);
	expr->Operation = PEX_MemberAccess;
	expr->Type = NULL;
	expr->Left = yymsp[-2].minor.yy282;
	expr->Right = ENamedName(yymsp[0].minor.yy0.Int);
	yylhsminor.yy282 = expr;
}
#line 4962 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,44,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 248: /* primary ::= primary ADDADD */
#line 1507 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	UNARY_EXPR(yymsp[-1].minor.yy282,PEX_PostInc);
	yylhsminor.yy282 = expr1;
}
#line 4972 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,42,&yymsp[0].minor);
  yymsp[-1].minor.yy282 = yylhsminor.yy282;
        break;
      case 249: /* primary ::= primary SUBSUB */
#line 1512 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	UNARY_EXPR(yymsp[-1].minor.yy282,PEX_PostDec);
	yylhsminor.yy282 = expr1;
}
#line 4982 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,43,&yymsp[0].minor);
  yymsp[-1].minor.yy282 = yylhsminor.yy282;
        break;
      case 250: /* unary_expr ::= SUB unary_expr */
{  yy_destructor(yypParser,33,&yymsp[-1].minor);
#line 1528 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	ZCC_ExprConstant *con = static_cast<ZCC_ExprConstant *>(yymsp[0].minor.yy282);
	if (yymsp[0].minor.yy282->Operation == PEX_ConstValue && (con->Type->isInt() || con->Type->isFloat()))
	{ // For constants, manipulate the child node directly, and don't create a new node.
		if (con->Type->isInt())
		{
			con->IntVal = -con->IntVal;
		}
		else
		{
			con->DoubleVal = -con->DoubleVal;
		}
		yymsp[-1].minor.yy282 = yymsp[0].minor.yy282;
	}
	else
	{ // For everything else, create a new node and do the negation later.
		UNARY_EXPR(yymsp[0].minor.yy282,PEX_Negate);
		yymsp[-1].minor.yy282 = expr1;
	}
}
#line 5009 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 251: /* unary_expr ::= ADD unary_expr */
{  yy_destructor(yypParser,34,&yymsp[-1].minor);
#line 1549 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	// Even though this is really a no-op, we still need to make a node for
	// it so we can type check that it is being applied to something numeric.
	// But we can do that right now for constant numerals.
	ZCC_ExprConstant *con = static_cast<ZCC_ExprConstant *>(yymsp[0].minor.yy282);
	if (yymsp[0].minor.yy282->Operation != PEX_ConstValue || (!con->Type->isInt() && !con->Type->isFloat()))
	{
		UNARY_EXPR(yymsp[0].minor.yy282,PEX_AntiNegate);
		yymsp[-1].minor.yy282 = expr1;
	}
	else
	{
		yymsp[-1].minor.yy282 = yymsp[0].minor.yy282;
	}
}
#line 5030 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 252: /* unary_expr ::= SUBSUB unary_expr */
{  yy_destructor(yypParser,43,&yymsp[-1].minor);
#line 1565 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	UNARY_EXPR(yymsp[0].minor.yy282,PEX_PreDec);
	yymsp[-1].minor.yy282 = expr1;
}
#line 5040 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 253: /* unary_expr ::= ADDADD unary_expr */
{  yy_destructor(yypParser,42,&yymsp[-1].minor);
#line 1570 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	UNARY_EXPR(yymsp[0].minor.yy282,PEX_PreInc);
	yymsp[-1].minor.yy282 = expr1;
}
#line 5050 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 254: /* unary_expr ::= TILDE unary_expr */
{  yy_destructor(yypParser,138,&yymsp[-1].minor);
#line 1575 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	UNARY_EXPR(yymsp[0].minor.yy282,PEX_BitNot);
	yymsp[-1].minor.yy282 = expr1;
}
#line 5060 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 255: /* unary_expr ::= BANG unary_expr */
{  yy_destructor(yypParser,139,&yymsp[-1].minor);
#line 1580 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	UNARY_EXPR(yymsp[0].minor.yy282,PEX_BoolNot);
	yymsp[-1].minor.yy282 = expr1;
}
#line 5070 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 256: /* unary_expr ::= SIZEOF unary_expr */
{  yy_destructor(yypParser,140,&yymsp[-1].minor);
#line 1585 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	UNARY_EXPR(yymsp[0].minor.yy282,PEX_SizeOf);
	yymsp[-1].minor.yy282 = expr1;
}
#line 5080 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 257: /* unary_expr ::= ALIGNOF unary_expr */
{  yy_destructor(yypParser,141,&yymsp[-1].minor);
#line 1590 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	UNARY_EXPR(yymsp[0].minor.yy282,PEX_AlignOf);
	yymsp[-1].minor.yy282 = expr1;
}
#line 5090 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 258: /* expr ::= expr ADD expr */
#line 1603 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_Add);
	yylhsminor.yy282 = expr2;
}
#line 5099 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,34,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 259: /* expr ::= expr SUB expr */
#line 1608 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_Sub);
	yylhsminor.yy282 = expr2;
}
#line 5109 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,33,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 260: /* expr ::= expr MUL expr */
#line 1613 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_Mul);
	yylhsminor.yy282 = expr2;
}
#line 5119 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,35,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 261: /* expr ::= expr DIV expr */
#line 1618 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_Div);
	yylhsminor.yy282 = expr2;
}
#line 5129 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,36,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 262: /* expr ::= expr MOD expr */
#line 1623 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_Mod);
	yylhsminor.yy282 = expr2;
}
#line 5139 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,37,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 263: /* expr ::= expr POW expr */
#line 1628 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_Pow);
	yylhsminor.yy282 = expr2;
}
#line 5149 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,40,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 264: /* expr ::= expr CROSSPROD expr */
#line 1633 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_CrossProduct);
	yylhsminor.yy282 = expr2;
}
#line 5159 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,38,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 265: /* expr ::= expr DOTPROD expr */
#line 1638 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_DotProduct);
	yylhsminor.yy282 = expr2;
}
#line 5169 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,39,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 266: /* expr ::= expr LSH expr */
#line 1643 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_LeftShift);
	yylhsminor.yy282 = expr2;
}
#line 5179 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,30,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 267: /* expr ::= expr RSH expr */
#line 1648 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_RightShift);
	yylhsminor.yy282 = expr2;
}
#line 5189 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,31,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 268: /* expr ::= expr URSH expr */
#line 1653 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_URightShift);
	yylhsminor.yy282 = expr2;
}
#line 5199 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,32,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 269: /* expr ::= expr DOTDOT expr */
#line 1658 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_Concat);
	yylhsminor.yy282 = expr2;
}
#line 5209 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,26,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 270: /* expr ::= expr LT expr */
#line 1664 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_LT);
	yylhsminor.yy282 = expr2;
}
#line 5219 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,20,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 271: /* expr ::= expr GT expr */
#line 1669 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_GT);
	yylhsminor.yy282 = expr2;
}
#line 5229 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,21,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 272: /* expr ::= expr LTEQ expr */
#line 1674 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_LTEQ);
	yylhsminor.yy282 = expr2;
}
#line 5239 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,22,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 273: /* expr ::= expr GTEQ expr */
#line 1679 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_GTEQ);
	yylhsminor.yy282 = expr2;
}
#line 5249 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,23,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 274: /* expr ::= expr LTGTEQ expr */
#line 1684 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_LTGTEQ);
	yylhsminor.yy282 = expr2;
}
#line 5259 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,24,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 275: /* expr ::= expr IS expr */
#line 1689 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_Is);
	yylhsminor.yy282 = expr2;
}
#line 5269 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,25,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 276: /* expr ::= expr EQEQ expr */
#line 1695 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_EQEQ);
	yylhsminor.yy282 = expr2;
}
#line 5279 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,17,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 277: /* expr ::= expr NEQ expr */
#line 1700 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_NEQ);
	yylhsminor.yy282 = expr2;
}
#line 5289 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,18,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 278: /* expr ::= expr APPROXEQ expr */
#line 1705 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_APREQ);
	yylhsminor.yy282 = expr2;
}
#line 5299 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,19,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 279: /* expr ::= expr AND expr */
#line 1711 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_BitAnd);
	yylhsminor.yy282 = expr2;
}
#line 5309 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,29,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 280: /* expr ::= expr XOR expr */
#line 1716 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_BitXor);
	yylhsminor.yy282 = expr2;
}
#line 5319 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,28,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 281: /* expr ::= expr OR expr */
#line 1721 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_BitOr);
	yylhsminor.yy282 = expr2;
}
#line 5329 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,27,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 282: /* expr ::= expr ANDAND expr */
#line 1726 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_BoolAnd);
	yylhsminor.yy282 = expr2;
}
#line 5339 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,16,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 283: /* expr ::= expr OROR expr */
#line 1731 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_BoolOr);
	yylhsminor.yy282 = expr2;
}
#line 5349 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,15,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 284: /* expr ::= expr EQ expr */
#line 1736 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_Assign);
	yylhsminor.yy282 = expr2;
}
#line 5359 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,1,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 285: /* expr ::= expr ADDEQ expr */
#line 1741 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_AddAssign);
	yylhsminor.yy282 = expr2;
}
#line 5369 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,5,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 286: /* expr ::= expr SUBEQ expr */
#line 1746 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_SubAssign);
	yylhsminor.yy282 = expr2;
}
#line 5379 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,6,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 287: /* expr ::= expr MULEQ expr */
#line 1751 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_MulAssign);
	yylhsminor.yy282 = expr2;
}
#line 5389 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,2,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 288: /* expr ::= expr DIVEQ expr */
#line 1756 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_DivAssign);
	yylhsminor.yy282 = expr2;
}
#line 5399 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,3,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 289: /* expr ::= expr MODEQ expr */
#line 1761 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_ModAssign);
	yylhsminor.yy282 = expr2;
}
#line 5409 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,4,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 290: /* expr ::= expr LSHEQ expr */
#line 1766 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_LshAssign);
	yylhsminor.yy282 = expr2;
}
#line 5419 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,7,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 291: /* expr ::= expr RSHEQ expr */
#line 1771 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_RshAssign);
	yylhsminor.yy282 = expr2;
}
#line 5429 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,8,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 292: /* expr ::= expr URSHEQ expr */
#line 1776 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_URshAssign);
	yylhsminor.yy282 = expr2;
}
#line 5439 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,12,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 293: /* expr ::= expr ANDEQ expr */
#line 1781 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_AndAssign);
	yylhsminor.yy282 = expr2;
}
#line 5449 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,9,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 294: /* expr ::= expr OREQ expr */
#line 1786 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_OrAssign);
	yylhsminor.yy282 = expr2;
}
#line 5459 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,10,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 295: /* expr ::= expr XOREQ expr */
#line 1791 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_XorAssign);
	yylhsminor.yy282 = expr2;
}
#line 5469 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,11,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 296: /* expr ::= expr SCOPE expr */
#line 1798 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	BINARY_EXPR(yymsp[-2].minor.yy282,yymsp[0].minor.yy282,PEX_Scope);
	yylhsminor.yy282 = expr2;
}
#line 5479 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,47,&yymsp[-1].minor);
  yymsp[-2].minor.yy282 = yylhsminor.yy282;
        break;
      case 297: /* expr ::= expr QUESTION expr COLON expr */
#line 1806 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ExprTrinary, expr, yymsp[-4].minor.yy282);
	expr->Operation = PEX_Trinary;
	expr->Type = NULL;
	expr->Test = yymsp[-4].minor.yy282;
	expr->Left = yymsp[-2].minor.yy282;
	expr->Right = yymsp[0].minor.yy282;
	yylhsminor.yy282 = expr;
}
#line 5494 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,13,&yymsp[-3].minor);
  yy_destructor(yypParser,14,&yymsp[-1].minor);
  yymsp[-4].minor.yy282 = yylhsminor.yy282;
        break;
      case 298: /* expr_list ::= expr_list COMMA expr */
#line 1822 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	yymsp[-2].minor.yy282 = yymsp[-2].minor.yy282;	/*X-overwrites-A*/
	AppendTreeNodeSibling(yymsp[-2].minor.yy282, yymsp[0].minor.yy282);
}
#line 5505 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,50,&yymsp[-1].minor);
        break;
      case 299: /* func_expr_list ::= func_expr_list COMMA func_expr_item */
#line 1838 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	// Omitted parameters still need to appear as nodes in the list.
	if (yymsp[-2].minor.yy478 == NULL)
	{
		NEW_AST_NODE(FuncParm,nil_a,yymsp[-1].minor.yy0);
		nil_a->Value = NULL;
		nil_a->Label = NAME_None;
		yymsp[-2].minor.yy478 = nil_a;
	}
	if (yymsp[0].minor.yy478 == NULL)
	{
		NEW_AST_NODE(FuncParm,nil_b,yymsp[-1].minor.yy0);
		nil_b->Value = NULL;
		nil_b->Label = NAME_None;
		yymsp[0].minor.yy478 = nil_b;
	}
	yymsp[-2].minor.yy478 = yymsp[-2].minor.yy478;	/*X-overwrites-A*/
	AppendTreeNodeSibling(yymsp[-2].minor.yy478, yymsp[0].minor.yy478);
}
#line 5529 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 300: /* func_expr_item ::= */
#line 1859 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	yymsp[1].minor.yy478 = NULL;
}
#line 5536 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 301: /* named_expr ::= IDENTIFIER COLON expr */
#line 1865 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(FuncParm, parm, yymsp[-2].minor.yy0);
	parm->Value = yymsp[0].minor.yy282;
	parm->Label = ENamedName(yymsp[-2].minor.yy0.Int);
	yylhsminor.yy478 = parm;
}
#line 5546 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,14,&yymsp[-1].minor);
  yymsp[-2].minor.yy478 = yylhsminor.yy478;
        break;
      case 302: /* named_expr ::= expr */
#line 1872 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(FuncParm, parm, yymsp[0].minor.yy282);
	parm->Value = yymsp[0].minor.yy282;
	parm->Label = NAME_None;
	yylhsminor.yy478 = parm;
}
#line 5558 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy478 = yylhsminor.yy478;
        break;
      case 304: /* string_constant ::= string_constant STRCONST */
#line 1893 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ExprConstant, strconst, yymsp[-1].minor.yy423);
	strconst->Operation = PEX_ConstValue;
	strconst->Type = TypeString;
	strconst->StringVal = stat->Strings.Alloc(*(yymsp[-1].minor.yy423->StringVal) + *(yymsp[0].minor.yy0.String));
	yylhsminor.yy423 = strconst;
}
#line 5570 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[-1].minor.yy423 = yylhsminor.yy423;
        break;
      case 305: /* constant ::= INTCONST */
#line 1903 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_INTCONST_NODE(intconst, TypeSInt32, yymsp[0].minor.yy0.Int, yymsp[0].minor.yy0);
	yylhsminor.yy423 = intconst;
}
#line 5579 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy423 = yylhsminor.yy423;
        break;
      case 306: /* constant ::= UINTCONST */
#line 1908 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_INTCONST_NODE(intconst, TypeUInt32, yymsp[0].minor.yy0.Int, yymsp[0].minor.yy0);
	yylhsminor.yy423 = intconst;
}
#line 5588 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy423 = yylhsminor.yy423;
        break;
      case 307: /* constant ::= FLOATCONST */
#line 1913 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ExprConstant, floatconst, yymsp[0].minor.yy0);
	floatconst->Operation = PEX_ConstValue;
	floatconst->Type = TypeFloat64;
	floatconst->DoubleVal = yymsp[0].minor.yy0.Float;
	yylhsminor.yy423 = floatconst;
}
#line 5600 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy423 = yylhsminor.yy423;
        break;
      case 308: /* constant ::= NAMECONST */
#line 1921 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ExprConstant, floatconst, yymsp[0].minor.yy0);
	floatconst->Operation = PEX_ConstValue;
	floatconst->Type = TypeName;
	floatconst->IntVal = yymsp[0].minor.yy0.Int;
	yylhsminor.yy423 = floatconst;
}
#line 5612 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy423 = yylhsminor.yy423;
        break;
      case 309: /* constant ::= FALSE */
#line 1929 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_INTCONST_NODE(boolconst, TypeBool, false, yymsp[0].minor.yy0);
	yylhsminor.yy423 = boolconst;
}
#line 5621 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy423 = yylhsminor.yy423;
        break;
      case 310: /* constant ::= TRUE */
#line 1934 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_INTCONST_NODE(boolconst, TypeBool, true, yymsp[0].minor.yy0);
	yylhsminor.yy423 = boolconst;
}
#line 5630 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy423 = yylhsminor.yy423;
        break;
      case 311: /* constant ::= NULLPTR */
#line 1939 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ExprConstant, nullptrconst, yymsp[0].minor.yy0);
	nullptrconst->Operation = PEX_ConstValue;
	nullptrconst->Type = TypeNullPtr;
	nullptrconst->StringVal = nullptr;
	yylhsminor.yy423 = nullptrconst;
}
#line 5642 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy423 = yylhsminor.yy423;
        break;
      case 313: /* statement ::= labeled_statement */
#line 1953 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy3 = yymsp[0].minor.yy254; /*X-overwrites-A*/ }
#line 5648 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 314: /* statement ::= compound_statement */
#line 1954 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy3 = yymsp[0].minor.yy475; /*X-overwrites-A*/ }
#line 5653 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 315: /* statement ::= expression_statement SEMICOLON */
#line 1955 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-1].minor.yy3 = yymsp[-1].minor.yy330; /*X-overwrites-A*/ }
#line 5658 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,49,&yymsp[0].minor);
        break;
      case 316: /* statement ::= assign_statement SEMICOLON */
#line 1963 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-1].minor.yy3 = yymsp[-1].minor.yy284; /*X-overwrites-A*/ }
#line 5664 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,49,&yymsp[0].minor);
        break;
      case 317: /* statement ::= assign_decl_statement SEMICOLON */
#line 1964 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-1].minor.yy3 = yymsp[-1].minor.yy7; /*X-overwrites-A*/ }
#line 5670 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,49,&yymsp[0].minor);
        break;
      case 318: /* statement ::= local_var SEMICOLON */
#line 1965 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-1].minor.yy3 = yymsp[-1].minor.yy454; /*X-overwrites-A*/ }
#line 5676 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,49,&yymsp[0].minor);
        break;
      case 320: /* statement ::= staticarray_statement */
#line 1967 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy3 = yymsp[0].minor.yy286; /*X-overwrites-A*/ }
#line 5682 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 321: /* staticarray_statement ::= STATICCONST type IDENTIFIER LBRACKET RBRACKET EQ LBRACE expr_list RBRACE SEMICOLON */
{  yy_destructor(yypParser,148,&yymsp[-9].minor);
#line 1974 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(StaticArrayStatement, stmt, yymsp[-8].minor.yy196);
	stmt->Type = yymsp[-8].minor.yy196;
	stmt->Id = ENamedName(yymsp[-7].minor.yy0.Int);
	stmt->Values = yymsp[-2].minor.yy282;
	yymsp[-9].minor.yy286 = stmt;
}
#line 5694 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,46,&yymsp[-6].minor);
  yy_destructor(yypParser,122,&yymsp[-5].minor);
  yy_destructor(yypParser,1,&yymsp[-4].minor);
  yy_destructor(yypParser,68,&yymsp[-3].minor);
  yy_destructor(yypParser,69,&yymsp[-1].minor);
  yy_destructor(yypParser,49,&yymsp[0].minor);
}
        break;
      case 322: /* staticarray_statement ::= STATICCONST type LBRACKET RBRACKET IDENTIFIER EQ LBRACE expr_list RBRACE SEMICOLON */
{  yy_destructor(yypParser,148,&yymsp[-9].minor);
#line 1983 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(StaticArrayStatement, stmt, yymsp[-8].minor.yy196);
	stmt->Type = yymsp[-8].minor.yy196;
	stmt->Id = ENamedName(yymsp[-5].minor.yy0.Int);
	stmt->Values = yymsp[-2].minor.yy282;
	yymsp[-9].minor.yy286 = stmt;
}
#line 5713 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,46,&yymsp[-7].minor);
  yy_destructor(yypParser,122,&yymsp[-6].minor);
  yy_destructor(yypParser,1,&yymsp[-4].minor);
  yy_destructor(yypParser,68,&yymsp[-3].minor);
  yy_destructor(yypParser,69,&yymsp[-1].minor);
  yy_destructor(yypParser,49,&yymsp[0].minor);
}
        break;
      case 323: /* jump_statement ::= CONTINUE SEMICOLON */
#line 1996 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ContinueStmt, stmt, yymsp[-1].minor.yy0);
	yylhsminor.yy3 = stmt;
}
#line 5728 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,49,&yymsp[0].minor);
  yymsp[-1].minor.yy3 = yylhsminor.yy3;
        break;
      case 324: /* jump_statement ::= BREAK SEMICOLON */
#line 2001 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(BreakStmt, stmt, yymsp[-1].minor.yy0);
	yylhsminor.yy3 = stmt;
}
#line 5738 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,49,&yymsp[0].minor);
  yymsp[-1].minor.yy3 = yylhsminor.yy3;
        break;
      case 325: /* jump_statement ::= RETURN SEMICOLON */
#line 2006 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ReturnStmt, stmt, yymsp[-1].minor.yy0);
	stmt->Values = NULL;
	yylhsminor.yy3 = stmt;
}
#line 5749 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,49,&yymsp[0].minor);
  yymsp[-1].minor.yy3 = yylhsminor.yy3;
        break;
      case 326: /* jump_statement ::= RETURN expr_list SEMICOLON */
#line 2012 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ReturnStmt, stmt, yymsp[-2].minor.yy0);
	stmt->Values = yymsp[-1].minor.yy282;
	yylhsminor.yy3 = stmt;
}
#line 5760 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,49,&yymsp[0].minor);
  yymsp[-2].minor.yy3 = yylhsminor.yy3;
        break;
      case 327: /* compound_statement ::= LBRACE RBRACE */
#line 2024 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(CompoundStmt,stmt,yymsp[-1].minor.yy0);
	stmt->Content = NULL;
	yylhsminor.yy475 = stmt;
}
#line 5771 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,69,&yymsp[0].minor);
  yymsp[-1].minor.yy475 = yylhsminor.yy475;
        break;
      case 328: /* compound_statement ::= LBRACE statement_list RBRACE */
#line 2030 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(CompoundStmt,stmt,yymsp[-2].minor.yy0);
	stmt->Content = yymsp[-1].minor.yy3;
	yylhsminor.yy475 = stmt;
}
#line 5782 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,69,&yymsp[0].minor);
  yymsp[-2].minor.yy475 = yylhsminor.yy475;
        break;
      case 329: /* compound_statement ::= LBRACE error RBRACE */
#line 2036 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(CompoundStmt,stmt,yymsp[-2].minor.yy0);
	stmt->Content = NULL;
	yylhsminor.yy475 = stmt;
}
#line 5793 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,69,&yymsp[0].minor);
  yymsp[-2].minor.yy475 = yylhsminor.yy475;
        break;
      case 332: /* expression_statement ::= expr */
#line 2056 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ExpressionStmt, stmt, yymsp[0].minor.yy282);
	stmt->Expression = yymsp[0].minor.yy282;
	yylhsminor.yy330 = stmt;
}
#line 5804 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy330 = yylhsminor.yy330;
        break;
      case 333: /* iteration_statement ::= while_or_until LPAREN expr RPAREN statement */
#line 2068 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(IterationStmt, iter, yymsp[-4].minor.yy0);
	if (yymsp[-4].minor.yy0.Int == ZCC_UNTIL)
	{ // Negate the loop condition
		UNARY_EXPR(yymsp[-2].minor.yy282,PEX_BoolNot);
		iter->LoopCondition = expr1;
	}
	else
	{
		iter->LoopCondition = yymsp[-2].minor.yy282;
	}
	iter->LoopStatement = yymsp[0].minor.yy3;
	iter->LoopBumper = NULL;
	iter->CheckAt = ZCC_IterationStmt::Start;
	yylhsminor.yy3 = iter;
}
#line 5825 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,45,&yymsp[-3].minor);
  yy_destructor(yypParser,63,&yymsp[-1].minor);
  yymsp[-4].minor.yy3 = yylhsminor.yy3;
        break;
      case 334: /* iteration_statement ::= DO statement while_or_until LPAREN expr RPAREN */
#line 2086 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(IterationStmt, iter, yymsp[-5].minor.yy0);
	if (yymsp[-3].minor.yy0.Int == ZCC_UNTIL)
	{ // Negate the loop condition
		UNARY_EXPR(yymsp[-1].minor.yy282,PEX_BoolNot);
		iter->LoopCondition = expr1;
	}
	else
	{
		iter->LoopCondition = yymsp[-1].minor.yy282;
	}
	iter->LoopStatement = yymsp[-4].minor.yy3;
	iter->LoopBumper = NULL;
	iter->CheckAt = ZCC_IterationStmt::End;
	yylhsminor.yy3 = iter;
}
#line 5848 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,45,&yymsp[-2].minor);
  yy_destructor(yypParser,63,&yymsp[0].minor);
  yymsp[-5].minor.yy3 = yylhsminor.yy3;
        break;
      case 335: /* iteration_statement ::= FOR LPAREN for_init SEMICOLON opt_expr SEMICOLON for_bump RPAREN statement */
#line 2104 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(IterationStmt, iter, yymsp[-8].minor.yy0);
	iter->LoopCondition = yymsp[-4].minor.yy282;
	iter->LoopStatement = yymsp[0].minor.yy3;
	iter->LoopBumper = yymsp[-2].minor.yy3;
	iter->CheckAt = ZCC_IterationStmt::Start;
	// The initialization expression appears outside the loop
	// for_init may be NULL if there is no initialization.
	SAFE_APPEND(yymsp[-6].minor.yy3, iter);
	// And the whole thing gets wrapped inside a compound statement in case the loop
	// initializer defined any variables.
	NEW_AST_NODE(CompoundStmt, wrap, yymsp[-8].minor.yy0);
	wrap->Content = yymsp[-6].minor.yy3;
	yylhsminor.yy3 = wrap;
}
#line 5870 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,45,&yymsp[-7].minor);
  yy_destructor(yypParser,49,&yymsp[-5].minor);
  yy_destructor(yypParser,49,&yymsp[-3].minor);
  yy_destructor(yypParser,63,&yymsp[-1].minor);
  yymsp[-8].minor.yy3 = yylhsminor.yy3;
        break;
      case 336: /* array_iteration_statement ::= FOREACH LPAREN variable_name COLON expr RPAREN statement */
#line 2123 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ArrayIterationStmt, iter, yymsp[-6].minor.yy0);
	iter->ItName = yymsp[-4].minor.yy492;
	iter->ItArray = yymsp[-2].minor.yy282;
	iter->LoopStatement = yymsp[0].minor.yy3;
	yylhsminor.yy3 = iter;
}
#line 5886 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,45,&yymsp[-5].minor);
  yy_destructor(yypParser,14,&yymsp[-3].minor);
  yy_destructor(yypParser,63,&yymsp[-1].minor);
  yymsp[-6].minor.yy3 = yylhsminor.yy3;
        break;
      case 337: /* two_arg_iteration_statement ::= FOREACH LPAREN variable_name COMMA variable_name COLON expr RPAREN statement */
#line 2134 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(TwoArgIterationStmt, iter, yymsp[-8].minor.yy0);
	iter->ItKey = yymsp[-6].minor.yy492;
	iter->ItValue = yymsp[-4].minor.yy492;
	iter->ItMap = yymsp[-2].minor.yy282;
	iter->LoopStatement = yymsp[0].minor.yy3;
	yylhsminor.yy3 = iter;
}
#line 5902 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,45,&yymsp[-7].minor);
  yy_destructor(yypParser,50,&yymsp[-5].minor);
  yy_destructor(yypParser,14,&yymsp[-3].minor);
  yy_destructor(yypParser,63,&yymsp[-1].minor);
  yymsp[-8].minor.yy3 = yylhsminor.yy3;
        break;
      case 338: /* three_arg_iteration_statement ::= FOREACH LPAREN variable_name COMMA variable_name COMMA variable_name COLON expr RPAREN statement */
#line 2146 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(ThreeArgIterationStmt, iter, yymsp[-10].minor.yy0);
	iter->ItVar = yymsp[-8].minor.yy492;
	iter->ItPos = yymsp[-6].minor.yy492;
	iter->ItFlags = yymsp[-4].minor.yy492;
	iter->ItBlock = yymsp[-2].minor.yy282;
	iter->LoopStatement = yymsp[0].minor.yy3;
	yylhsminor.yy3 = iter;
}
#line 5920 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,45,&yymsp[-9].minor);
  yy_destructor(yypParser,50,&yymsp[-7].minor);
  yy_destructor(yypParser,50,&yymsp[-5].minor);
  yy_destructor(yypParser,14,&yymsp[-3].minor);
  yy_destructor(yypParser,63,&yymsp[-1].minor);
  yymsp[-10].minor.yy3 = yylhsminor.yy3;
        break;
      case 339: /* typed_iteration_statement ::= FOREACH LPAREN variable_name variable_name COLON expr RPAREN statement */
#line 2159 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(TypedIterationStmt, iter, yymsp[-7].minor.yy0);
	iter->ItType = yymsp[-5].minor.yy492;
	iter->ItVar = yymsp[-4].minor.yy492;
	iter->ItExpr = yymsp[-2].minor.yy282;
	iter->LoopStatement = yymsp[0].minor.yy3;
	yylhsminor.yy3 = iter;
}
#line 5938 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,45,&yymsp[-6].minor);
  yy_destructor(yypParser,14,&yymsp[-3].minor);
  yy_destructor(yypParser,63,&yymsp[-1].minor);
  yymsp[-7].minor.yy3 = yylhsminor.yy3;
        break;
      case 340: /* while_or_until ::= WHILE */
#line 2169 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	yylhsminor.yy0.Int = ZCC_WHILE;
	yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc;
}
#line 5950 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 341: /* while_or_until ::= UNTIL */
#line 2174 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	yylhsminor.yy0.Int = ZCC_UNTIL;
	yylhsminor.yy0.SourceLoc = yymsp[0].minor.yy0.SourceLoc;
}
#line 5959 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy0 = yylhsminor.yy0;
        break;
      case 342: /* for_init ::= local_var */
#line 2180 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy3 = yymsp[0].minor.yy454 /*X-overwrites-A*/; }
#line 5965 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 343: /* for_init ::= for_bump */
#line 2181 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy3 = yymsp[0].minor.yy3 /*X-overwrites-A*/; }
#line 5970 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 344: /* for_bump ::= */
#line 2184 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[1].minor.yy3 = NULL; }
#line 5975 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 345: /* for_bump ::= expression_statement */
#line 2185 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[0].minor.yy3 = yymsp[0].minor.yy330; /*X-overwrites-A*/ }
#line 5980 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 346: /* for_bump ::= for_bump COMMA expression_statement */
#line 2186 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ yymsp[-2].minor.yy3 = yymsp[-2].minor.yy3; /*X-overwrites-A*/ AppendTreeNodeSibling(yymsp[-2].minor.yy3, yymsp[0].minor.yy330); }
#line 5985 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,50,&yymsp[-1].minor);
        break;
      case 347: /* selection_statement ::= if_front */
#line 2200 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	yymsp[0].minor.yy3 = yymsp[0].minor.yy45; /*X-overwrites-A*/
}
#line 5993 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
        break;
      case 348: /* selection_statement ::= if_front ELSE statement */
#line 2204 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	yymsp[-2].minor.yy45->FalsePath = yymsp[0].minor.yy3;
	yymsp[-2].minor.yy3 = yymsp[-2].minor.yy45; /*X-overwrites-A*/
}
#line 6001 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,158,&yymsp[-1].minor);
        break;
      case 349: /* if_front ::= IF LPAREN expr RPAREN statement */
#line 2210 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(IfStmt,stmt,yymsp[-4].minor.yy0);
	stmt->Condition = yymsp[-2].minor.yy282;
	stmt->TruePath = yymsp[0].minor.yy3;
	stmt->FalsePath = NULL;
	yylhsminor.yy45 = stmt;
}
#line 6013 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,45,&yymsp[-3].minor);
  yy_destructor(yypParser,63,&yymsp[-1].minor);
  yymsp[-4].minor.yy45 = yylhsminor.yy45;
        break;
      case 350: /* selection_statement ::= SWITCH LPAREN expr RPAREN statement */
#line 2221 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(SwitchStmt,stmt,yymsp[-4].minor.yy0);
	stmt->Condition = yymsp[-2].minor.yy282;
	stmt->Content = yymsp[0].minor.yy3;
	yylhsminor.yy3 = stmt;
}
#line 6026 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,45,&yymsp[-3].minor);
  yy_destructor(yypParser,63,&yymsp[-1].minor);
  yymsp[-4].minor.yy3 = yylhsminor.yy3;
        break;
      case 351: /* labeled_statement ::= CASE expr COLON */
#line 2233 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(CaseStmt,stmt,yymsp[-2].minor.yy0);
	stmt->Condition = yymsp[-1].minor.yy282;
	yylhsminor.yy254 = stmt;
}
#line 6038 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,14,&yymsp[0].minor);
  yymsp[-2].minor.yy254 = yylhsminor.yy254;
        break;
      case 352: /* labeled_statement ::= DEFAULT COLON */
#line 2239 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(CaseStmt,stmt,yymsp[-1].minor.yy0);
	stmt->Condition = NULL;
	yylhsminor.yy254 = stmt;
}
#line 6049 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,14,&yymsp[0].minor);
  yymsp[-1].minor.yy254 = yylhsminor.yy254;
        break;
      case 353: /* assign_statement ::= LBRACKET expr_list RBRACKET EQ expr */
{  yy_destructor(yypParser,46,&yymsp[-4].minor);
#line 2252 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(AssignStmt,stmt,yymsp[-3].minor.yy282);
	stmt->AssignOp = ZCC_EQ;
	stmt->Dests = yymsp[-3].minor.yy282;
	stmt->Sources = yymsp[0].minor.yy282;
	yymsp[-4].minor.yy284 = stmt;
}
#line 6063 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,122,&yymsp[-2].minor);
  yy_destructor(yypParser,1,&yymsp[-1].minor);
}
        break;
      case 354: /* assign_decl_statement ::= LET LBRACKET identifier_list RBRACKET EQ expr */
{  yy_destructor(yypParser,109,&yymsp[-5].minor);
#line 2263 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(AssignDeclStmt,stmt,yymsp[-3].minor.yy493);
	stmt->AssignOp = ZCC_EQ;
	stmt->Dests = yymsp[-3].minor.yy493;
	stmt->Sources = yymsp[0].minor.yy282;
	yymsp[-5].minor.yy7 = stmt;
}
#line 6078 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,46,&yymsp[-4].minor);
  yy_destructor(yypParser,122,&yymsp[-2].minor);
  yy_destructor(yypParser,1,&yymsp[-1].minor);
}
        break;
      case 355: /* local_var ::= type variable_list_with_init */
#line 2276 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(LocalVarStmt,vardef,yymsp[-1].minor.yy196);
	vardef->Type = yymsp[-1].minor.yy196;
	vardef->Vars = yymsp[0].minor.yy55;
	yylhsminor.yy454 = vardef;
}
#line 6092 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[-1].minor.yy454 = yylhsminor.yy454;
        break;
      case 356: /* var_init ::= IDENTIFIER */
#line 2285 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ 
	NEW_AST_NODE(VarInit,var,yymsp[0].minor.yy0);
	var->Name = ENamedName(yymsp[0].minor.yy0.Int);
	var->ArraySize = NULL;
	var->Init = NULL;
	var->InitIsArray = false;
	yylhsminor.yy55 = var;
}
#line 6105 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[0].minor.yy55 = yylhsminor.yy55;
        break;
      case 357: /* var_init ::= IDENTIFIER array_size */
#line 2295 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	NEW_AST_NODE(VarInit,var,yymsp[-1].minor.yy0);
	var->Name = ENamedName(yymsp[-1].minor.yy0.Int);
	var->ArraySize = yymsp[0].minor.yy282;
	var->Init = NULL;
	var->InitIsArray = false;
	yylhsminor.yy55 = var;
}
#line 6118 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yymsp[-1].minor.yy55 = yylhsminor.yy55;
        break;
      case 358: /* var_init ::= IDENTIFIER EQ expr */
#line 2305 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ 
	NEW_AST_NODE(VarInit,var,yymsp[-2].minor.yy0);
	var->Name = ENamedName(yymsp[-2].minor.yy0.Int);
	var->ArraySize = NULL;
	var->Init = yymsp[0].minor.yy282;
	var->InitIsArray = false;
	yylhsminor.yy55 = var;
}
#line 6131 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,1,&yymsp[-1].minor);
  yymsp[-2].minor.yy55 = yylhsminor.yy55;
        break;
      case 359: /* var_init ::= IDENTIFIER EQ LBRACE expr_list RBRACE */
#line 2315 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ 
	NEW_AST_NODE(VarInit,var,yymsp[-4].minor.yy0);
	var->Name = ENamedName(yymsp[-4].minor.yy0.Int);
	var->ArraySize = NULL;
	var->Init = yymsp[-1].minor.yy282;
	var->InitIsArray = true;
	yylhsminor.yy55 = var;
}
#line 6145 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,1,&yymsp[-3].minor);
  yy_destructor(yypParser,68,&yymsp[-2].minor);
  yy_destructor(yypParser,69,&yymsp[0].minor);
  yymsp[-4].minor.yy55 = yylhsminor.yy55;
        break;
      case 360: /* var_init ::= IDENTIFIER array_size EQ LBRACE expr_list RBRACE */
#line 2325 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ 
	NEW_AST_NODE(VarInit,var,yymsp[-5].minor.yy0);
	var->Name = ENamedName(yymsp[-5].minor.yy0.Int);
	var->ArraySize = yymsp[-4].minor.yy282;
	var->Init = yymsp[-1].minor.yy282;
	var->InitIsArray = true;
	yylhsminor.yy55 = var;
}
#line 6161 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,1,&yymsp[-3].minor);
  yy_destructor(yypParser,68,&yymsp[-2].minor);
  yy_destructor(yypParser,69,&yymsp[0].minor);
  yymsp[-5].minor.yy55 = yylhsminor.yy55;
        break;
      case 361: /* var_init ::= IDENTIFIER EQ LBRACE error RBRACE */
{  yy_destructor(yypParser,54,&yymsp[-4].minor);
#line 2335 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{ 
	yymsp[-4].minor.yy55 = NULL;
}
#line 6173 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,1,&yymsp[-3].minor);
  yy_destructor(yypParser,68,&yymsp[-2].minor);
  yy_destructor(yypParser,69,&yymsp[0].minor);
}
        break;
      case 362: /* variable_list_with_init ::= variable_list_with_init COMMA var_init */
#line 2343 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
	AppendTreeNodeSibling(yymsp[-2].minor.yy55, yymsp[0].minor.yy55);
	yymsp[-2].minor.yy55 = yymsp[-2].minor.yy55; /*X-overwrites-A*/
}
#line 6185 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,50,&yymsp[-1].minor);
        break;
      case 363: /* translation_unit ::= translation_unit EOF */
#line 179 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
}
#line 6192 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,48,&yymsp[0].minor);
        break;
      case 365: /* opt_semicolon ::= SEMICOLON */
{  yy_destructor(yypParser,49,&yymsp[0].minor);
#line 192 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
}
#line 6200 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 367: /* opt_comma ::= COMMA */
{  yy_destructor(yypParser,50,&yymsp[0].minor);
#line 195 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
}
#line 6208 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
}
        break;
      case 373: /* state_flow ::= state_flow_type scanner_mode SEMICOLON */
#line 671 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
}
#line 6215 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,49,&yymsp[0].minor);
        break;
      case 385: /* primary ::= LPAREN error RPAREN */
{  yy_destructor(yypParser,45,&yymsp[-2].minor);
#line 1458 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
{
}
#line 6223 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
  yy_destructor(yypParser,63,&yymsp[0].minor);
}
        break;
      default:
      /* (364) opt_semicolon ::= */ yytestcase(yyruleno==364);
      /* (366) opt_comma ::= */ yytestcase(yyruleno==366);
      /* (368) opt_expr ::= expr */ yytestcase(yyruleno==368);
      /* (369) opt_struct_body ::= struct_body */ yytestcase(yyruleno==369);
      /* (370) struct_body ::= struct_member */ yytestcase(yyruleno==370);
      /* (371) enum_list ::= enumerator */ yytestcase(yyruleno==371);
      /* (372) opt_enum_list ::= enum_list opt_comma */ yytestcase(yyruleno==372);
      /* (374) type_name1 ::= int_type */ yytestcase(yyruleno==374);
      /* (375) func_ptr_params ::= func_ptr_param_list */ yytestcase(yyruleno==375);
      /* (376) func_ptr_param_list ::= func_ptr_param */ yytestcase(yyruleno==376);
      /* (377) type_or_array ::= type */ yytestcase(yyruleno==377);
      /* (378) type_list ::= type_or_array */ yytestcase(yyruleno==378);
      /* (379) type_list_or_void ::= type_list */ yytestcase(yyruleno==379);
      /* (380) array_size ::= array_size_expr */ yytestcase(yyruleno==380);
      /* (381) variable_list ::= variable_name */ yytestcase(yyruleno==381);
      /* (382) opt_func_body ::= function_body */ yytestcase(yyruleno==382);
      /* (383) func_params ::= func_param_list */ yytestcase(yyruleno==383);
      /* (384) func_param_list ::= func_param */ yytestcase(yyruleno==384);
      /* (386) unary_expr ::= primary */ yytestcase(yyruleno==386);
      /* (387) expr ::= unary_expr */ yytestcase(yyruleno==387);
      /* (388) expr_list ::= expr */ yytestcase(yyruleno==388);
      /* (389) func_expr_list ::= func_expr_item */ yytestcase(yyruleno==389);
      /* (390) func_expr_item ::= named_expr */ yytestcase(yyruleno==390);
      /* (391) constant ::= string_constant */ yytestcase(yyruleno==391);
      /* (392) function_body ::= compound_statement */ yytestcase(yyruleno==392);
      /* (393) statement ::= selection_statement */ yytestcase(yyruleno==393);
      /* (394) statement ::= iteration_statement */ yytestcase(yyruleno==394);
      /* (395) statement ::= array_iteration_statement */ yytestcase(yyruleno==395);
      /* (396) statement ::= two_arg_iteration_statement */ yytestcase(yyruleno==396);
      /* (397) statement ::= three_arg_iteration_statement */ yytestcase(yyruleno==397);
      /* (398) statement ::= typed_iteration_statement */ yytestcase(yyruleno==398);
      /* (399) statement ::= jump_statement */ yytestcase(yyruleno==399);
      /* (400) variable_list_with_init ::= var_init */ yytestcase(yyruleno==400);
       break;
/********** End reduce actions ************************************************/
  };
  assert( yyruleno<sizeof(yyRuleInfo)/sizeof(yyRuleInfo[0]) );
  yygoto = yyRuleInfo[yyruleno].lhs;
  yysize = yyRuleInfo[yyruleno].nrhs;
  yyact = yy_find_reduce_action(yymsp[-yysize].stateno,(YYCODETYPE)yygoto);
  if( yyact <= YY_MAX_SHIFTREDUCE ){
    if( yyact>YY_MAX_SHIFT ){
      yyact += YY_MIN_REDUCE - YY_MIN_SHIFTREDUCE;
    }
    yymsp -= yysize-1;
    yypParser->yytos = yymsp;
    yymsp->stateno = (YYACTIONTYPE)yyact;
    yymsp->major = (YYCODETYPE)yygoto;
    yyTraceShift(yypParser, yyact);
  }else{
    assert( yyact == YY_ACCEPT_ACTION );
    yypParser->yytos -= yysize;
    yy_accept(yypParser);
  }
}

/*
** The following code executes when the parse fails
*/
#ifndef YYNOERRORRECOVERY
static void yy_parse_failed(
  yyParser *yypParser           /* The parser */
){
  ZCCParseARG_FETCH;
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sFail!\n",yyTracePrompt);
    fflush(yyTraceFILE);
  }
#endif
  while( yypParser->yytos>yypParser->yystack ) yy_pop_parser_stack(yypParser);
  /* Here code is inserted which will be executed whenever the
  ** parser fails */
/************ Begin %parse_failure code ***************************************/
#line 149 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
 /**failed = true;*/ 
#line 6304 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
/************ End %parse_failure code *****************************************/
  ZCCParseARG_STORE; /* Suppress warning about unused %extra_argument variable */
}
#endif /* YYNOERRORRECOVERY */

/*
** The following code executes when a syntax error first occurs.
*/
static void yy_syntax_error(
  yyParser *yypParser,           /* The parser */
  int yymajor,                   /* The major type of the error token */
  ZCCParseTOKENTYPE yyminor         /* The minor type of the error token */
){
  ZCCParseARG_FETCH;
#define TOKEN yyminor
/************ Begin %syntax_error code ****************************************/
#line 122 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"

	FString unexpected, expecting;

	int i;
	int stateno = yypParser->yytos->stateno;

	unexpected << "Unexpected " << ZCCTokenName(yymajor);

	// Determine all the terminals that the parser would have accepted at this point
	// (see yy_find_shift_action). This list can get quite long. Is it worthwhile to
	// print it when not debugging the grammar, or would that be too confusing to
	// the average user?
	if (stateno < YY_SHIFT_MAX && (i = yy_shift_ofst[stateno])!=YY_SHIFT_USE_DFLT)
	{
		for (int j = 1; j < YYERRORSYMBOL; ++j)
		{
			int k = i + j;
			if (k >= 0 && k < YY_ACTTAB_COUNT && yy_lookahead[k] == j)
			{
				expecting << (expecting.IsEmpty() ? "Expecting " : " or ") << ZCCTokenName(j);
			}
		}
	}
	stat->sc->ScriptMessage("%s\n%s\n", unexpected.GetChars(), expecting.GetChars());
	FScriptPosition::ErrorCounter++;
#line 6347 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
/************ End %syntax_error code ******************************************/
  ZCCParseARG_STORE; /* Suppress warning about unused %extra_argument variable */
}

/*
** The following is executed when the parser accepts
*/
static void yy_accept(
  yyParser *yypParser           /* The parser */
){
  ZCCParseARG_FETCH;
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sAccept!\n",yyTracePrompt);
    fflush(yyTraceFILE);
  }
#endif
#ifndef YYNOERRORRECOVERY
  yypParser->yyerrcnt = -1;
#endif
#if 0
  assert( yypParser->yytos==yypParser->yystack );
#else
  while (yypParser->yytos>yypParser->yystack) yy_pop_parser_stack(yypParser);
#endif
  /* Here code is inserted which will be executed whenever the
  ** parser accepts */
/*********** Begin %parse_accept code *****************************************/
#line 148 "F:/download/browser/gzdoom-g4.14.0_origin/src/common/scripting/frontend/zcc-parse.lemon"
 DPrintf(DMSG_SPAMMY, "Input accepted\n"); 
#line 6378 "F:/download/browser/gzdoom-g4.14.0_origin/win_x64/src/zcc-parse.c"
/*********** End %parse_accept code *******************************************/
  ZCCParseARG_STORE; /* Suppress warning about unused %extra_argument variable */
}

/* The main parser program.
** The first argument is a pointer to a structure obtained from
** "ZCCParseAlloc" which describes the current state of the parser.
** The second argument is the major token number.  The third is
** the minor token.  The fourth optional argument is whatever the
** user wants (and specified in the grammar) and is available for
** use by the action routines.
**
** Inputs:
** <ul>
** <li> A pointer to the parser (an opaque structure.)
** <li> The major token number.
** <li> The minor token number.
** <li> An option argument of a grammar-specified type.
** </ul>
**
** Outputs:
** None.
*/
void ZCCParse(
  void *yyp,                   /* The parser */
  int yymajor,                 /* The major token code number */
  ZCCParseTOKENTYPE yyminor       /* The value for the token */
  ZCCParseARG_PDECL               /* Optional %extra_argument parameter */
){
  YYMINORTYPE yyminorunion;
  unsigned int yyact;   /* The parser action. */
#if !defined(YYERRORSYMBOL) && !defined(YYNOERRORRECOVERY)
  int yyendofinput;     /* True if we are at the end of input */
#endif
#ifdef YYERRORSYMBOL
  int yyerrorhit = 0;   /* True if yymajor has invoked an error */
#endif
  yyParser *yypParser;  /* The parser */

  yypParser = (yyParser*)yyp;
  assert( yypParser->yytos!=0 );
#if !defined(YYERRORSYMBOL) && !defined(YYNOERRORRECOVERY)
  yyendofinput = (yymajor==0);
#endif
  ZCCParseARG_STORE;

#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sInput '%s'\n",yyTracePrompt,yyTokenName[yymajor]);
    fflush(yyTraceFILE);
  }
#endif

  do{
    yyact = yy_find_shift_action(yypParser,(YYCODETYPE)yymajor);
    if( yyact <= YY_MAX_SHIFTREDUCE ){
      yy_shift(yypParser,yyact,yymajor,yyminor);
#ifndef YYNOERRORRECOVERY
      yypParser->yyerrcnt--;
#endif
      yymajor = YYNOCODE;
    }else if( yyact <= YY_MAX_REDUCE ){
      yy_reduce(yypParser,yyact-YY_MIN_REDUCE);
    }else{
      assert( yyact == YY_ERROR_ACTION );
      yyminorunion.yy0 = yyminor;
#ifdef YYERRORSYMBOL
      int yymx;
#endif
#ifndef NDEBUG
      if( yyTraceFILE ){
        fprintf(yyTraceFILE,"%sSyntax Error!\n",yyTracePrompt);
        fflush(yyTraceFILE);
      }
#endif
#ifdef YYERRORSYMBOL
      /* A syntax error has occurred.
      ** The response to an error depends upon whether or not the
      ** grammar defines an error token "ERROR".  
      **
      ** This is what we do if the grammar does define ERROR:
      **
      **  * Call the %syntax_error function.
      **
      **  * Begin popping the stack until we enter a state where
      **    it is legal to shift the error symbol, then shift
      **    the error symbol.
      **
      **  * Set the error count to three.
      **
      **  * Begin accepting and shifting new tokens.  No new error
      **    processing will occur until three tokens have been
      **    shifted successfully.
      **
      */
      if( yypParser->yyerrcnt<0 ){
        yy_syntax_error(yypParser,yymajor,yyminor);
      }
      yymx = yypParser->yytos->major;
      if( yymx==YYERRORSYMBOL || yyerrorhit ){
#ifndef NDEBUG
        if( yyTraceFILE ){
          fprintf(yyTraceFILE,"%sDiscard input token %s\n",
             yyTracePrompt,yyTokenName[yymajor]);
          fflush(yyTraceFILE);
        }
#endif
        yy_destructor(yypParser, (YYCODETYPE)yymajor, &yyminorunion);
        yymajor = YYNOCODE;
      }else{
        while( yypParser->yytos >= yypParser->yystack
            && yymx != YYERRORSYMBOL
            && (yyact = yy_find_reduce_action(
                        yypParser->yytos->stateno,
                        YYERRORSYMBOL)) >= YY_MIN_REDUCE
        ){
          yy_pop_parser_stack(yypParser);
        }
        if( yypParser->yytos < yypParser->yystack || yymajor==0 ){
          yy_destructor(yypParser,(YYCODETYPE)yymajor,&yyminorunion);
          yy_parse_failed(yypParser);
#ifndef YYNOERRORRECOVERY
          yypParser->yyerrcnt = -1;
#endif
          yymajor = YYNOCODE;
        }else if( yymx!=YYERRORSYMBOL ){
          yy_shift(yypParser,yyact,YYERRORSYMBOL,yyminor);
        }
      }
      yypParser->yyerrcnt = 3;
      yyerrorhit = 1;
#elif defined(YYNOERRORRECOVERY)
      /* If the YYNOERRORRECOVERY macro is defined, then do not attempt to
      ** do any kind of error recovery.  Instead, simply invoke the syntax
      ** error routine and continue going as if nothing had happened.
      **
      ** Applications can set this macro (for example inside %include) if
      ** they intend to abandon the parse upon the first syntax error seen.
      */
      yy_syntax_error(yypParser,yymajor, yyminor);
      yy_destructor(yypParser,(YYCODETYPE)yymajor,&yyminorunion);
      yymajor = YYNOCODE;
      
#else  /* YYERRORSYMBOL is not defined */
      /* This is what we do if the grammar does not define ERROR:
      **
      **  * Report an error message, and throw away the input token.
      **
      **  * If the input token is $, then fail the parse.
      **
      ** As before, subsequent error messages are suppressed until
      ** three input tokens have been successfully shifted.
      */
      if( yypParser->yyerrcnt<=0 ){
        yy_syntax_error(yypParser,yymajor, yyminor);
      }
      yypParser->yyerrcnt = 3;
      yy_destructor(yypParser,(YYCODETYPE)yymajor,&yyminorunion);
      if( yyendofinput ){
        yy_parse_failed(yypParser);
#ifndef YYNOERRORRECOVERY
        yypParser->yyerrcnt = -1;
#endif
      }
      yymajor = YYNOCODE;
#endif
    }
  }while( yymajor!=YYNOCODE && yypParser->yytos>yypParser->yystack );
#ifndef NDEBUG
  if( yyTraceFILE ){
    yyStackEntry *i;
    char cDiv = '[';
    fprintf(yyTraceFILE,"%sReturn. Stack=",yyTracePrompt);
    for(i=&yypParser->yystack[1]; i<=yypParser->yytos; i++){
      fprintf(yyTraceFILE,"%c%s", cDiv, yyTokenName[i->major]);
      cDiv = ' ';
    }
    fprintf(yyTraceFILE,"]\n");
    fflush(yyTraceFILE);
  }
#endif
  return;
}
